//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.annotation;

import org.eclipse.jface.text.Position;

/**
 * 
 * @author Yoonki Song
 */
public final class ExceptionAnnotation extends FixBugsAnnotation {

	/** */
	private boolean fIsCaught;

	/** */
	private boolean fIsMultiCaught;

	/**
	 * Constructor.
	 * 
	 * @param type
	 *            the type.
	 * @param position
	 *            the position.
	 */
	public ExceptionAnnotation(final String type, final Position position) {
		super(type, position);
		fIsCaught = false;
		fIsMultiCaught = false;
	}

	public ExceptionAnnotation(final String type) {
		this(type, null);
	}

	/**
	 * Checks if the annotation is in catch clause.
	 * 
	 * @return true if the annotation is in catch clause.
	 */
	public boolean isCaught() {
		return fIsCaught;
	}

	/**
	 * Sets ...
	 * 
	 * @param isCaught
	 *            inCatch
	 */
	public void setCaught(final boolean isCaught) {
		fIsCaught = isCaught;
	}

	/**
	 * Sets ...
	 * 
	 * @param isMultiCaught
	 *            inMultiCatch
	 */
	public void setMultiCaught(final boolean isMultiCaught) {
		fIsMultiCaught = isMultiCaught;
	}

	/**
	 * Checks if the annotation is in multi-catch clause.
	 * 
	 * @return true if the annotation is in multi-catch clause.
	 */
	public boolean isMultiCaught() {
		return fIsMultiCaught;
	}

	/**
	 * Checks if the annotation is in throws clause.
	 * 
	 * @return true if the annotation is in throws clause.
	 */
	public boolean isThrown() {
		return !fIsCaught;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append(super.toString());
		builder.append(", IsCaught: " + fIsCaught);
		builder.append(", InMultiCaught: " + fIsMultiCaught);
		return builder.toString();
	}
}
