//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.annotation;

import java.util.Comparator;

import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.source.Annotation;

/**
 * 
 * @author Yoonki Song
 */
public class FixBugsAnnotation extends Annotation implements Comparator<FixBugsAnnotation> {

	/** */
	public static final String ADDED = "edu.ncsu.csc.emh.fixbugs.added"; //$NON-NLS-1$

	/** */
	public static final String MODIFIED = "edu.ncsu.csc.emh.fixbugs.modified"; //$NON-NLS-1$

	/** */
	public static final String DELETED = "edu.ncsu.csc.emh.fixbugs.deleted"; //$NON-NLS-1$

	/** */
	public static final String HIDDEN = "edu.ncsu.csc.emh.fixbugs.hidden"; //$NON-NLS-1$

	/** */
	public static final String RELATED = "edu.ncsu.csc.emh.fixbugs.related"; //$NON-NLS-1$

	/** */
	public static final String MISSING_ARGUMENT = "edu.ncsu.csc.emh.fixbugs.missingArgument"; //$NON-NLS-1$

	/** */
	public static final String DRAGGABLE = "edu.ncsu.csc.emh.fixbugs.draggable"; //$NON-NLS-1$

	/** */
	public static final String SELECTED = "edu.ncsu.csc.emh.fixbugs.selected"; //$NON-NLS-1$

	/** */
	public static final String MULTI_CALLER = "edu.ncsu.csc.emh.fixbugs.multiCaller"; //$NON-NLS-1$

	/** */
	public static final String KIND1P = "edu.ncsu.csc.emh.fixbugs.kind1p"; //$NON-NLS-1$

	/** */
	public static final String KIND1S = "edu.ncsu.csc.emh.fixbugs.kind1s"; //$NON-NLS-1$

	/** */
	public static final String KIND2P = "edu.ncsu.csc.emh.fixbugs.kind2p"; //$NON-NLS-1$

	/** */
	public static final String KIND2S = "edu.ncsu.csc.emh.fixbugs.kind2s"; //$NON-NLS-1$

	/** */
	public static final String KIND3P = "edu.ncsu.csc.emh.fixbugs.kind3p"; //$NON-NLS-1$

	/** */
	public static final String KIND3S = "edu.ncsu.csc.emh.fixbugs.kind3s"; //$NON-NLS-1$

	/** */
	public static final String KIND4P = "edu.ncsu.csc.emh.fixbugs.kind4p"; //$NON-NLS-1$

	/** */
	public static final String KIND4S = "edu.ncsu.csc.emh.fixbugs.kind4s"; //$NON-NLS-1$

	/** */
	public static final String KIND5P = "edu.ncsu.csc.emh.fixbugs.kind5p"; //$NON-NLS-1$

	/** */
	public static final String KIND5S = "edu.ncsu.csc.emh.fixbugs.kind5s"; //$NON-NLS-1$

	/** */
	public static final String KIND6P = "edu.ncsu.csc.emh.fixbugs.kind6p"; //$NON-NLS-1$

	/** */
	public static final String KIND6S = "edu.ncsu.csc.emh.fixbugs.kind6s"; //$NON-NLS-1$

	/** */
	public static final String KIND7P = "edu.ncsu.csc.emh.fixbugs.kind7p"; //$NON-NLS-1$

	/** */
	public static final String KIND7S = "edu.ncsu.csc.emh.fixbugs.kind7s"; //$NON-NLS-1$

	/** */
	public static final String KIND8P = "edu.ncsu.csc.emh.fixbugs.kind8p"; //$NON-NLS-1$

	/** */
	public static final String KIND8S = "edu.ncsu.csc.emh.fixbugs.kind8s"; //$NON-NLS-1$

	/** */
	public static final String KIND9P = "edu.ncsu.csc.emh.fixbugs.kind9p"; //$NON-NLS-1$

	/** */
	public static final String KIND9S = "edu.ncsu.csc.emh.fixbugs.kind9s"; //$NON-NLS-1$

	/** */
	public static final String KIND10P = "edu.ncsu.csc.emh.fixbugs.kind10p"; //$NON-NLS-1$

	/** */
	public static final String KIND10S = "edu.ncsu.csc.emh.fixbugs.kind10s"; //$NON-NLS-1$

	/** */
	public static final String KIND11P = "edu.ncsu.csc.emh.fixbugs.kind11p"; //$NON-NLS-1$

	/** */
	public static final String KIND11S = "edu.ncsu.csc.emh.fixbugs.kind11s"; //$NON-NLS-1$

	/** */
	public static final String NODE_TYPE_UNIDENTIFIERD = "Unidentified";

	/** */
	public static final String UNKNOWN_NAME = "Unknown";

	/** */
	private String fName;

	/** */
	private Position fPosition;

	/**
	 * Constructor.
	 * 
	 * @param type
	 *            the type.
	 * @param position
	 *            the position.
	 */
	public FixBugsAnnotation(final String type, final Position position) {
		setType(type);
		setText("");
		fName = "";
		fPosition = position;
	}

	public FixBugsAnnotation(final String type, final String text) {
		setType(type);
		setText(text);
		fName = "";
		fPosition = null;
	}

	public FixBugsAnnotation(final String type, final String text, final Position position) {
		setType(type);
		setText(text);
		fName = "";
		fPosition = position;
	}

	public FixBugsAnnotation(final String type, final String text, final String name, final Position position) {
		setType(type);
		setText(text);
		fName = name;
		fPosition = position;
	}

	public FixBugsAnnotation(final String type) {
		this(type, null, null);
	}

	/**
	 * Returns the position.
	 * 
	 * @return the position.
	 */
	public Position getPosition() {
		return fPosition;
	}

	/**
	 * Sets the position.
	 * 
	 * @param position
	 *            the position.
	 */
	public void setPosition(final Position position) {
		fPosition = position;
	}

	/**
	 * Returns the name.
	 * 
	 * @return the name.
	 */
	public String getName() {
		if ("".equals(fName)) {
			return UNKNOWN_NAME;
		}
		return fName;
	}

	/**
	 * Sets the node name.
	 * 
	 * @param name
	 *            the node name.
	 */
	public void setName(final String name) {
		fName = name;
	}

	public boolean isPreviewable() {
		if (ADDED.equals(getType()) || MODIFIED.equals(getType())) {
			return true;
		}
		if (getType().contains("edu.ncsu.csc.emh.fixbugs.kind")) {
			return true;
		}
		return false;
	}

	/**
	 * Checks if the annotation covers the given offset.
	 * 
	 * @param offset
	 *            the offset.
	 * @return true if the annotation covers the given offset; false otherwise.
	 */
	public boolean includes(final int offset) {
		return fPosition.includes(offset);
	}

	/**
	 * Checks if the annotation covers the given offset.
	 * 
	 * @param position
	 *            the position.
	 * @return
	 */
	public boolean includes(final Position position) {
		return fPosition.includes(position.getOffset());
	}

	@Override
	public int compare(final FixBugsAnnotation o1, final FixBugsAnnotation o2) {
		return o1.getPosition().getOffset() - o2.getPosition().getOffset();
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("Type: " + getType()).append(", ");
		builder.append("Text: " + getText()).append(", ");
		builder.append("Name: " + fName).append(", ");
		builder.append("isMarkedDeleted: " + isMarkedDeleted()).append(", ");
		builder.append("Position: " + fPosition.toString());
		return builder.toString();
	}
}
