/**
 *
 */
package edu.ncsu.csc.emh.fixbugs.core.resolution;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.internal.corext.dom.LinkedNodeFinder;
import org.eclipse.jface.text.Position;

/**
 * @author Yoonki
 */
@SuppressWarnings("restriction")
public abstract class FixBugsLinkedAssistant extends FixBugsAssistant implements ILinkedNodeSupport, ISeedNodeSupport {

	/**
	 * Constructor for FixBugsLinkedContext.
	 * 
	 * @param context
	 *            the invocation context.
	 */
	public FixBugsLinkedAssistant(final FixBugsContext context, final SearchHint hint) {
		super(context);
		fSearchHint = hint;
	}

	// -------------------------- seed node --------------------------//

	/** */
	private ASTNode fSeedNode = null;

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final ASTNode getSeedNode() {
		return fSeedNode;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void setSeedNode(final ASTNode seedNode) {
		fSeedNode = seedNode;
	}

	// -------------------------- search hint --------------------------//

	/** */
	private SearchHint fSearchHint = null;

	public void setSearchHint(final SearchHint searchHint) {
		fSearchHint = searchHint;
	}

	public SearchHint getSearchHint() {
		return fSearchHint;
	}

	public boolean matches(String qn) {
		return fSearchHint.matches(qn);
	}

	// -------------------------- linked nodes list -----------------------//

	/** */
	private List<ASTNode[]> fLinkedNodesList = null;

	/**
	 * Returns the list of the linked nodes.
	 * 
	 * @return the list of the linked nodes.
	 */
	@Override
	public final List<ASTNode[]> getLinkedNodesList() {
		return fLinkedNodesList;
	}

	/**
	 * Sets the list of the linked nodes.
	 * 
	 * @param linkedNodesList
	 *            the list of the linked nodes.
	 */
	@Override
	public final void setLinkedNodesList(final List<ASTNode[]> linkedNodesList) {
		fLinkedNodesList = linkedNodesList;
	}

	/**
	 * Gets the position of the first linked node.
	 * 
	 * @return the position of the first linked node.
	 */
	public final Position getFirstLinkedNodePosition() {
		ASTNode node = fLinkedNodesList.get(0)[0];
		return getNodePosition(node);
	}

	protected void addLinkedPositions(List<ITrackedNodePosition> trackedVariables, ASTNode targetNode,
			String fOriginalIdentifier, ASTRewrite rewrite) {
		VariableNameMatcher matcher = new VariableNameMatcher(fOriginalIdentifier);
		targetNode.accept(matcher);
		List<SimpleName> matchedNodes = matcher.getMatchedNodes();
		for (SimpleName name : matchedNodes) {
			trackedVariables.add(rewrite.track(name));
		}
	}

	class VariableNameMatcher extends ASTVisitor {
		String fIdentifier;
		List<SimpleName> fMatchedNodes = new ArrayList<SimpleName>();

		VariableNameMatcher(String identifier) {
			fIdentifier = identifier;
		}

		@Override
		public boolean visit(SimpleName node) {
			// System.out.println("\tSimpleName: " + node);
			if (fIdentifier.equals(node.getIdentifier())) {
				fMatchedNodes.add(node);
			}
			return super.visit(node);
		}

		public List<SimpleName> getMatchedNodes() {
			return fMatchedNodes;
		}

	};

	/**
	 * 
	 * @param name
	 * @return
	 */
	protected final SimpleName[] findLinkedNodesByNode(final SimpleName name) {
		return LinkedNodeFinder.findByNode(getContext().getASTRoot(), name);
	}

	/**
	 * 
	 * @param node
	 * @param identifier
	 * @return
	 */
	protected static boolean containsIdentifier(final ASTNode node, final String identifier) {
		if (node instanceof VariableDeclarationStatement) {
			List fragments = ((VariableDeclarationStatement) node).fragments();
			for (Object fragment : fragments) {
				VariableDeclarationFragment vdf = (VariableDeclarationFragment) fragment;
				if (identifier.equals(vdf.getName().getIdentifier())) {
					return true;
				}
			}
		}
		return false;
	}

}
