//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.resolution;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.ltk.core.refactoring.RefactoringStatus;

/**
 * 
 * @see RefactoringStatus
 * @author Yoonki
 */
public final class FixBugsStatus {

	/** */
	public static final int OK = 0;

	/** */
	public static final int INFO = 1;

	/** */
	public static final int WARNING = 2;

	/** */
	public static final int ERROR = 3;

	/** */
	public static final int FATAL = 4;

	/** */
	private final List fEntries;

	/** */
	private int fSeverity = OK;

	/** */
	public FixBugsStatus() {
		fEntries = new ArrayList(0);
	}

	/**
	 * Merges the receiver and the parameter statuses. The resulting list of
	 * entries in the receiver will contain entries from both. The resulting
	 * severity in the receiver will be the more severe of its current severity
	 * and the parameter's severity. Merging with <code>null</code> is allowed -
	 * it has no effect.
	 * 
	 * @param other
	 *            the refactoring status to merge with
	 */
	public void merge(final FixBugsStatus other) {
		if (other == null) {
			return;
		}
		fEntries.addAll(other.fEntries);
		fSeverity = Math.max(fSeverity, other.getSeverity());
	}

	/**
	 * Returns <code>true</code> if the current severity is <code>
	 * FATAL</code>.
	 * 
	 * @return <code>true</code> if the current severity is <code>
	 *  FATAL</code>; otherwise <code>false</code> is returned
	 */
	public boolean hasFatalError() {
		return fSeverity == FATAL;
	}

	/**
	 * Returns the severity.
	 * 
	 * @return the severity.
	 */
	public int getSeverity() {
		return fSeverity;
	}
}
