/**
 *
 */
package edu.ncsu.csc.emh.fixbugs.core.resolution;

import java.util.List;

import org.eclipse.jdt.core.dom.ASTNode;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;

/**
 * @author Yoonki
 */
public interface IAnnotationSupport {

	/**
	 * Collects <code>Statement</code>s or <code>Expression</code>s.
	 * <p>
	 * Filter the collected nodes and put them into a list of related
	 * <code>Statement</code>s. For example, if there are two nodes in a
	 * <code>Statement</code>, just collect it once.
	 * <p>
	 * NOTE org.eclipse.jdt.core.dom.VariableDeclarationStatement cannot be cast
	 * to java.lang.Comparable
	 */
	void initAnnotations();

	/**
	 * Finds an annotation that is covering the offset.
	 * 
	 * @param offset
	 *            the offset.
	 * @return the annotation found.
	 */
	FixBugsAnnotation findAnnotation(int offset);

	/**
	 * Returns the list of annotations.
	 * 
	 * @return the list of annotations.
	 */
	List<FixBugsAnnotation> getAnnotations();

	/**
	 * Sets the list of annotations.
	 * 
	 * @param annotations
	 *            the list of annotations.
	 */
	void setAnnotations(List<FixBugsAnnotation> annotations);

	/**
	 * Returns the array of annotation nodes.
	 * 
	 * @return the array of annotation nodes.
	 */
	ASTNode[] getAnnotationNodes();

	/**
	 * Sets the array of annotation nodes.
	 * 
	 * @param annotationNodes
	 *            the array of annotation nodes.
	 */
	void setAnnotationNodes(ASTNode[] annotationNodes);
}
