//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.resolution.badPractice.openStreamExceptionPath;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.internal.resources.mapping.ChangeDescription;
import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Assignment;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.BodyDeclaration;
import org.eclipse.jdt.core.dom.CatchClause;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ExpressionStatement;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.IfStatement;
import org.eclipse.jdt.core.dom.ImportDeclaration;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SimpleType;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.TryStatement;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.VariableDeclarationExpression;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite.ImportRewriteContext;
import org.eclipse.jdt.core.dom.rewrite.ListRewrite;
import org.eclipse.jdt.internal.corext.codemanipulation.ContextSensitiveImportRewriteContext;
import org.eclipse.jdt.internal.corext.codemanipulation.StubUtility;
import org.eclipse.jdt.internal.corext.dom.ASTNodes;
import org.eclipse.jdt.internal.corext.dom.Bindings;
import org.eclipse.jdt.internal.corext.dom.LinkedNodeFinder;
import org.eclipse.jdt.internal.corext.refactoring.util.TightSourceRangeComputer;
import org.eclipse.jdt.internal.corext.util.JavaModelUtil;
import org.eclipse.jdt.internal.ui.text.correction.ASTResolving;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Position;
import org.eclipse.text.edits.MultiTextEdit;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.dom.rewrite.RewriteUtil;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsLinkedAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.SearchHint;
import edu.ncsu.csc.emh.fixbugs.core.util.PrintUtil;
import edu.ncsu.csc.emh.fixbugs.ui.Messages;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.badPractice.openStreamExceptionPath.OpenStreamExceptionPathMode;

/**
 * @author Yoonki Song
 */
@SuppressWarnings("restriction")
public final class OpenStreamExceptionPathAssistant extends FixBugsLinkedAssistant {

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(OpenStreamExceptionPathAssistant.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	private ImportRewrite fImportRewrite;

	/** */
	private String fOriginalIdentifier = null;

	/**
	 * Constructor for OpenStreamExceptionPathAssistant.
	 * 
	 * @param context
	 *            the invocation context.
	 */
	public OpenStreamExceptionPathAssistant(final FixBugsContext context, final SearchHint hint) {
		super(context, hint);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init() throws BadLocationException, JavaModelException {
		super.init();
		initSeedNode(ASTNode.SIMPLE_NAME);
		initIdentifier();
		initLinkedNodesList();
		setAdjustedSelection(getFirstLinkedNodePosition());
		initAnnotations();
	}

	// ------------------ seed node --------------------//

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void initSeedNode(final int nodeType) {
		fLogger.info("OpenStreamExceptionPathAssistant#initSeedNode: ");
		ASTNode seedNode = findSeedNode(nodeType, getContext().getCoveringNode());
		setSeedNode(seedNode);
		fLogger.info(PrintUtil.toString("Seed Node Found", getSeedNode()));
	}

	/**
	 * Finds the seed node.
	 * 
	 * @param nodeType
	 *            the node type
	 * @param node
	 *            the clue AST node
	 * @return the AST node that is found.
	 */
	@Override
	public ASTNode findSeedNode(final int nodeType, final ASTNode node) {
		ASTNode seedNode = null;
		if (node instanceof Block) {
			List<Statement> statements = ((Block) node).statements();
			fLogger.info("statements.size(): " + statements.size());
			for (Statement statement : statements) {
				fLogger.info(ASTNodes.asString(statement) + ", " + statement.getStartPosition());
				// the nearest statement to the current position
				if (isSelectionOverlapped(statement)) {
					seedNode = findBestGuessedVariableName(statement);
					break;
				}
			}
		} else if (node instanceof VariableDeclarationStatement) {
			seedNode = findBestGuessedVariableName(node);
		} else if (node instanceof VariableDeclarationExpression) {
			seedNode = findBestGuessedVariableName(node);
		} else if (node instanceof ExpressionStatement) {
			seedNode = findBestGuessedVariableName(node);
		} else if (node instanceof SimpleName) {
			seedNode = node;
		}
		Assert.isNotNull(seedNode);
		return seedNode;
	}

	private ASTNode findBestGuessedVariableName(final ASTNode node) {
		ASTNode guessedVariable = null;
		if (node instanceof VariableDeclarationStatement) {
			// e.g. InputStream in = new FileInputStream("name.txt");
			Type type = ((VariableDeclarationStatement) node).getType();
			String qn = type.resolveBinding().getQualifiedName();
			if (matches(qn)) {
				// first element
				Object obj = ((VariableDeclarationStatement) node).fragments().get(0);
				if (obj instanceof VariableDeclarationFragment) {
					guessedVariable = ((VariableDeclarationFragment) obj).getName();
				}
			}
		} else if (node instanceof VariableDeclarationExpression) {
			// e.g. try (InputStream in = new FileInputStream("name.txt")) {...}
			Type type = ((VariableDeclarationExpression) node).getType();
			String qn = type.resolveBinding().getQualifiedName();
			if (matches(qn)) {
				// first element
				Object obj = ((VariableDeclarationExpression) node).fragments().get(0);
				if (obj instanceof VariableDeclarationFragment) {
					guessedVariable = ((VariableDeclarationFragment) obj).getName();
				}
			}
		} else if (node instanceof ExpressionStatement) {
			// e.g. in = new FileInputStream("name.txt");
			Expression expr = ((ExpressionStatement) node).getExpression();
			if (expr instanceof Assignment) {
				Expression lhs = ((Assignment) expr).getLeftHandSide();
				String qn = lhs.resolveTypeBinding().getQualifiedName();
				if (matches(qn)) {
					guessedVariable = lhs;
				}
			}
		}
		return guessedVariable;
	}

	// ------------------ identifier --------------------//

	private void initIdentifier() {
		fOriginalIdentifier = ((SimpleName) getSeedNode()).getIdentifier();
	}

	/**
	 * Returns the original identifier.
	 * 
	 * @return the original identifier.
	 */
	public String getOriginalIdentifier() {
		return fOriginalIdentifier;
	}

	// ------------------ annotations -------------------//

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void initAnnotations() {
		initAnnotationNodes();
		initAnnotationTypes();
	}

	private void initAnnotationNodes() {
		List<ASTNode[]> linkedNodesList = getLinkedNodesList();
		Assert.isNotNull(linkedNodesList); // pre-condition
		List<ASTNode> annotationNodes = new ArrayList<ASTNode>();
		for (ASTNode[] nodes : linkedNodesList) {
			for (ASTNode node : nodes) {
				// The related part can be a Statement or an Expression!!
				ASTNode parentNode = ASTResolving.findParentStatement(node);
				if (!annotationNodes.contains(parentNode)) {
					annotationNodes.add(parentNode);
				}
			}
		}
		setAnnotationNodes(annotationNodes.toArray(new ASTNode[annotationNodes.size()]));
	}

	private void initAnnotationTypes() {
		ASTNode[] annotationNodes = getAnnotationNodes();
		Assert.isNotNull(annotationNodes); // pre-condition
		List<FixBugsAnnotation> annotations = new ArrayList<FixBugsAnnotation>();
		for (ASTNode parentNode : annotationNodes) {
			Position p = getNodePosition(parentNode);
			annotations.add(new FixBugsAnnotation(FixBugsAnnotation.RELATED, p));
		}
		setAnnotations(annotations);
	}

	// ------------------ linked nodes list --------------------//

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void initLinkedNodesList() {
		fLogger.info("OpenStreamExceptionPathAssistant#initLinkedNodesList");
		SimpleName name = (SimpleName) getSeedNode();
		Assert.isNotNull(name, "The name may not be null"); // pre-condition

		// find all the same node
		ASTNode[] linkedNodes = findLinkedNodesByNode(name);

		// sort by the starting position
		Arrays.sort(linkedNodes, new Comparator<ASTNode>() {
			@Override
			public int compare(final ASTNode n1, final ASTNode n2) {
				return n1.getStartPosition() - n2.getStartPosition();
			}
		});

		List<ASTNode[]> linkedNodesList = new ArrayList<ASTNode[]>();
		linkedNodesList.add(linkedNodes);
		setLinkedNodesList(linkedNodesList);
	}

	/**
	 * Updates...
	 * 
	 * @param annotations
	 *            new annotations.
	 * @param position
	 *            new position.
	 */
	public void update(final List<FixBugsAnnotation> annotations) {
		fLogger.info("OpenStreamExceptionPathAssistant#update");
		long startTime = System.currentTimeMillis();
		// updateLinkedNodesList(varLinks);
		setAnnotations(annotations); // update
		fLogger.info("#update: " + (System.currentTimeMillis() - startTime) + " ms");
	}

	/**
	 * 
	 * @param newPosition
	 *            the new position
	 */
	public void updateLinkedNodesList(List<Position> varLinks) {
		fLogger.info("OpenStreamExceptionPathAssistant#updateLinkedNodesList");
		long startTime = System.currentTimeMillis();

		List<ASTNode[]> linkedNodesList = new ArrayList<ASTNode[]>();
		setLinkedNodesList(linkedNodesList);

		fLogger.info("#updateLinkedNodesList: " + (System.currentTimeMillis() - startTime) + " ms");
	}

	/**
	 * Applies the changes actually to current compilation unit.
	 * 
	 * @param trackedAnnotationMap
	 *            tracked annotation map.
	 * @param trackedVariables
	 *            tracked variables.
	 * @return the changed text
	 * 
	 * @throws BadLocationException
	 *             is thrown if one of the edits in the tree can't be executed.
	 *             The state of the document is undefined if this exception is
	 *             thrown.
	 * @throws CoreException
	 *             the exception is thrown if the rewrite fails.
	 */
	public String changeAST(final String suggestion, final Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap,
			final List<ITrackedNodePosition> trackedVariables) throws BadLocationException, CoreException {
		String changedText = null;
		if (OpenStreamExceptionPathMode.SPLIT_AND_MOVE_CLOSE_TO_FINALLY_BLOCK.equals(suggestion)) {
			changedText = changeCloseInFinallyBlock(trackedAnnotationMap, trackedVariables);
		} else {
			changedText = changeTryWithResources(trackedAnnotationMap, trackedVariables);
		}

		return changedText;
	}

	private String changeCloseInFinallyBlock(final Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap,
			final List<ITrackedNodePosition> trackedVariables) throws BadLocationException, CoreException {
		fLogger.info("OpenStreamExceptionPathAssistant#changeCloseInFinallyBlock");

		// get DOM/AST from a ICompilationUnit
		CompilationUnit root = getContext().getASTRoot();

		// creation of ASTRewrite
		ASTRewrite rewrite = ASTRewrite.create(root.getAST());

		// ----------------------------- Change -----------------------------//

		// get related nodes (statements or expressions)
		ASTNode[] annotationNodes = getAnnotationNodes();
		String varId = getOriginalIdentifier();
		// System.out.println("VarId: " + varId);

		// description of the change
		// XXX What if there is no close() method??
		Block newFinallyBlock = null;
		IfStatement newIfStatement = null; // will be used in finally block
		for (ASTNode node : annotationNodes) {
			if (needToDeclareStreamOutsideTryBlock(node, varId)) {
				declareStreamOutsideTryBlock(rewrite, node, varId, trackedAnnotationMap, trackedVariables);
			} else if (needToCreateFinallyBlock(node) && newFinallyBlock == null) {
				newFinallyBlock = createFinallyBlock(rewrite, node, trackedAnnotationMap, trackedVariables);
			} else if (needToCloseStreamInFinallyBlock(node) && newIfStatement == null) {
				newIfStatement = closeStreamInFinallyBlock(rewrite, node, newFinallyBlock, varId, trackedAnnotationMap,
						trackedVariables, root);
			} else {
				FixBugsAnnotation annotation = new FixBugsAnnotation(FixBugsAnnotation.RELATED, Messages.relatedAnnotation);
				trackedAnnotationMap.put(rewrite.track(node), annotation);
				addLinkedPositions(trackedVariables, node, fOriginalIdentifier, rewrite);
			}
		}
		fRewrite = rewrite; // (important for quickfix)
		return rewriteMethodDecl(rewrite);
	}

	private String changeTryWithResources(final Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap,
			final List<ITrackedNodePosition> trackedVariables) throws BadLocationException, JavaModelException {
		fLogger.info("OpenStreamExceptionPathAssistant#changeTryWithResources");
		// get DOM/AST from a ICompilationUnit
		CompilationUnit root = getContext().getASTRoot();

		// creation of ASTRewrite
		ASTRewrite rewrite = ASTRewrite.create(root.getAST());

		// ----------------------------- Change -----------------------------//

		// find a name node and its related nodes(statements or expressions)
		ASTNode[] annotationNodes = getAnnotationNodes();
		ASTNode[] linkedNodes = getLinkedNodesList().get(0);
		ASTNode firstLinkedNode = linkedNodes[0];

		TryStatement tryStatement = findCoveringTryStatement(annotationNodes);

		// description of the change
		if (canJoinVariable(root, firstLinkedNode)) {
			fLogger.info("Can Join");
			joinMoveStreamAndSkipOrRemove(rewrite, root, firstLinkedNode, tryStatement, annotationNodes, trackedAnnotationMap,
					trackedVariables);
		} else {
			fLogger.info("Cannot Join");
			moveStreamAndSkipOrRemove(rewrite, root, firstLinkedNode, tryStatement, annotationNodes, trackedAnnotationMap,
					trackedVariables);
		}
		fRewrite = rewrite; // (important for quickfix)
		return rewriteMethodDecl(rewrite);
	}

	private boolean needToDeclareStreamOutsideTryBlock(ASTNode node, final String varId) {
		fLogger.info("#needToDeclareStreamOutsideTryBlock");
		if (node instanceof VariableDeclarationStatement && containsIdentifier(node, varId)) {
			while ((node != null) && (!(node instanceof BodyDeclaration))) {
				node = node.getParent();
				if (node instanceof TryStatement) {
					fLogger.info("needToMoveOutOfTryBlock: YES!!!");
					return true;
				}
			}
		}

		return false;
	}

	// node must be VariableDeclarationStatement
	private void declareStreamOutsideTryBlock(final ASTRewrite rewrite, final ASTNode node, final String varId,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap,
			final List<ITrackedNodePosition> trackedVariables) {
		fLogger.info("#declareStreamOutsideTryBlock: " + ASTNodes.asString(node) + ", var: " + varId);

		//
		TryStatement tryStatement = ASTResolving.findParentTryStatement(node);
		AST ast = node.getAST();

		//
		VariableDeclarationStatement vds = (VariableDeclarationStatement) node;
		Type varType = (Type) ASTNode.copySubtree(ast, vds.getType());

		// create new VariableDeclarationStatement
		VariableDeclarationFragment newVdf = ast.newVariableDeclarationFragment();
		SimpleName variableName = ast.newSimpleName(varId);
		trackedVariables.add(rewrite.track(variableName)); // track!
		newVdf.setName(variableName);
		newVdf.setInitializer(ast.newNullLiteral());
		VariableDeclarationStatement newVds = ast.newVariableDeclarationStatement(newVdf);
		newVds.setType(varType);

		// rewrite new VariableDeclarationStatement
		Block parentTryBlock = (Block) tryStatement.getParent();
		ListRewrite listRewrite = rewrite.getListRewrite(parentTryBlock, Block.STATEMENTS_PROPERTY);
		listRewrite.insertBefore(newVds, tryStatement, null);

		FixBugsAnnotation ma1 = new FixBugsAnnotation(FixBugsAnnotation.MODIFIED, Messages.modifiedAnnotation);
		trackedAnnotationMap.put(rewrite.track(newVds), ma1);

		// create new ExpressionStatement
		VariableDeclarationFragment fragment = findFirstVariableDeclarationFragment(vds, varId);
		Expression placeholder = (Expression) rewrite.createMoveTarget(fragment.getInitializer());
		// System.out.println("\t- placeholder: " + placeholder);

		Statement newVarExprStatement;
		Assignment assignment = ast.newAssignment();
		assignment.setRightHandSide(placeholder);
		SimpleName sn2 = ast.newSimpleName(fragment.getName().getIdentifier());
		assignment.setLeftHandSide(sn2);
		trackedVariables.add(rewrite.track(sn2)); // track!

		newVarExprStatement = ast.newExpressionStatement(assignment);

		// rewrite new ExpressionStatement
		rewrite.replace(vds, newVarExprStatement, null);
		FixBugsAnnotation ma2 = new FixBugsAnnotation(FixBugsAnnotation.MODIFIED, Messages.modifiedAnnotation);
		trackedAnnotationMap.put(rewrite.track(newVarExprStatement), ma2);
	}

	/**
	 * 'node' is a <code>Statement</code>. So it's parent must be a
	 * <code>Block</code>.
	 * 
	 * @param node
	 * @return
	 */
	private boolean needToCreateFinallyBlock(final ASTNode node) {
		fLogger.info("OpenStreamExceptionPathAssistant#needToCreateFinallyBlock");
		TryStatement tryStatement = ASTResolving.findParentTryStatement(node);
		if (tryStatement == null /* || tryStatement.getFinally() != null */) {
			return false;
		}
		// Statement statement = ASTResolving.findParentStatement(node);
		// if (tryStatement != statement && tryStatement.getBody() != statement)
		// {
		// return false; // a node inside a catch or finally block
		// }

		return true;
	}

	private IfStatement closeStreamInFinallyBlock(final ASTRewrite rewrite, final ASTNode closeStatement,
			final Block finallyBlock, final String varIdentifier,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap,
			List<ITrackedNodePosition> trackedVariables, final CompilationUnit root) throws CoreException {
		Assert.isNotNull(finallyBlock);
		AST ast = finallyBlock.getAST();

		// if (in != null) {...}
		IfStatement ifStatement = ast.newIfStatement();
		InfixExpression infixExpression = ast.newInfixExpression();
		SimpleName sn1 = ast.newSimpleName(varIdentifier);
		trackedVariables.add(rewrite.track(sn1));
		infixExpression.setLeftOperand(sn1);
		infixExpression.setOperator(InfixExpression.Operator.NOT_EQUALS);
		infixExpression.setRightOperand(ast.newNullLiteral());

		Block thenBlock = ast.newBlock();
		// if (in != null) {
		// try {
		// ...
		// }
		// }
		TryStatement tryStatement = ast.newTryStatement();
		Block blockInInnerTry = ast.newBlock();
		ExpressionStatement expressionStatement = null;
		if (closeStatement == null) {
			MethodInvocation closeMI = null; // close()
			closeMI = ast.newMethodInvocation();
			SimpleName sn2 = ast.newSimpleName(varIdentifier);
			trackedVariables.add(rewrite.track(sn2));
			closeMI.setExpression(sn2);
			closeMI.setName(ast.newSimpleName("close")); //$NON-NLS-1$
			expressionStatement = ast.newExpressionStatement(closeMI);
		} else {
			// addImport(ast, rewrite, closeStatement, root);
			rewrite.remove(closeStatement, null); // remote the existing close()
			expressionStatement = (ExpressionStatement) ASTNode.copySubtree(ast, closeStatement); // copy
			MethodInvocation closeMI = (MethodInvocation) expressionStatement.getExpression();
			trackedVariables.add(rewrite.track(closeMI.getExpression()));
		}
		blockInInnerTry.statements().add(expressionStatement);

		// ////////////////////////////////////////
		// Catch exception
		BodyDeclaration decl = ASTResolving.findParentBodyDeclaration(closeStatement);
		if (decl == null) {
			fLogger.warning("Something wrong: d");
			return null;
		}

		ITypeBinding[] uncaughtExceptions = findUncaughtExceptions(root, closeStatement);
		if (uncaughtExceptions.length == 0) {
			fLogger.info("All exceptions are caught!!");
		} else {
			for (ITypeBinding tb : uncaughtExceptions) {
				fLogger.info("uncaught: " + tb.getQualifiedName());
			}
		}

		TryStatement surroundingTry = ASTResolving.findParentTryStatement(closeStatement);
		ICompilationUnit cu = getContext().getCompilationUnit();
		if (surroundingTry != null
				&& (ASTNodes.isParent(closeStatement, surroundingTry.getBody()) || closeStatement.getLocationInParent() == TryStatement.RESOURCES_PROPERTY)) {
			ImportRewrite importRewrite = StubUtility.createImportRewrite(root, true);
			ImportRewriteContext importRewriteContext = new ContextSensitiveImportRewriteContext(decl, importRewrite);

			ListRewrite clauseRewrite = rewrite.getListRewrite(surroundingTry, TryStatement.CATCH_CLAUSES_PROPERTY);
			for (int i = 0; i < uncaughtExceptions.length; i++) {
				ITypeBinding excBinding = uncaughtExceptions[i];
				String varName = StubUtility.getExceptionVariableName(cu.getJavaProject());
				SingleVariableDeclaration var = ast.newSingleVariableDeclaration();
				var.setName(ast.newSimpleName(varName));
				var.setType(importRewrite.addImport(excBinding, ast, importRewriteContext));
				CatchClause newClause = ast.newCatchClause();
				newClause.setException(var);
				String catchBody = StubUtility.getCatchBodyContent(cu, excBinding.getName(), varName, closeStatement,
						String.valueOf('\n'));
				if (catchBody != null) {
					ASTNode node = rewrite.createStringPlaceholder(catchBody, ASTNode.RETURN_STATEMENT);
					newClause.getBody().statements().add(node);
				}
				clauseRewrite.insertLast(newClause, null);

				String typeKey = "type" + i; //$NON-NLS-1$
				String nameKey = "name" + i; //$NON-NLS-1$
			}

			//
			if (JavaModelUtil.is17OrHigher(cu.getJavaProject())) {
				// YS do something
			}
		} else {
			fLogger.warning("Something wrong: x");
		}

		// ---------------------- Start ---------------------//
		if (decl instanceof MethodDeclaration) {
			MethodDeclaration methodDecl = (MethodDeclaration) decl;
			IMethodBinding binding = methodDecl.resolveBinding();
			boolean isApplicable = (binding != null);
			if (isApplicable) {
				IMethodBinding overriddenMethod = Bindings.findOverriddenMethod(binding, true);
				if (overriddenMethod != null) {
					isApplicable = overriddenMethod.getDeclaringClass().isFromSource();
					if (!isApplicable) { // bug 349051
						ITypeBinding[] exceptionTypes = overriddenMethod.getExceptionTypes();
						ArrayList<ITypeBinding> unhandledExceptions = new ArrayList<ITypeBinding>(uncaughtExceptions.length);
						for (int i = 0; i < uncaughtExceptions.length; i++) {
							ITypeBinding curr = uncaughtExceptions[i];
							if (isSubtype(curr, exceptionTypes)) {
								unhandledExceptions.add(curr);
							}
						}
						uncaughtExceptions = unhandledExceptions.toArray(new ITypeBinding[unhandledExceptions.size()]);
						isApplicable |= uncaughtExceptions.length > 0;
					}
				}
			}
			if (isApplicable) {
				ITypeBinding[] methodExceptions = binding.getExceptionTypes();
				ArrayList<ITypeBinding> unhandledExceptions = new ArrayList<ITypeBinding>(uncaughtExceptions.length);
				for (int i = 0; i < uncaughtExceptions.length; i++) {
					ITypeBinding curr = uncaughtExceptions[i];
					if (!isSubtype(curr, methodExceptions)) {
						unhandledExceptions.add(curr);
					}
				}
				uncaughtExceptions = unhandledExceptions.toArray(new ITypeBinding[unhandledExceptions.size()]);

				List<Name> exceptions = methodDecl.thrownExceptions();
				int nExistingExceptions = exceptions.size();
				ChangeDescription[] desc = new ChangeDescription[nExistingExceptions + uncaughtExceptions.length];
				for (int i = 0; i < exceptions.size(); i++) {
					Name elem = exceptions.get(i);
					if (isSubtype(elem.resolveTypeBinding(), uncaughtExceptions)) {
						// desc[i] = new RemoveDescription();
					}
				}
				for (int i = 0; i < uncaughtExceptions.length; i++) {
					// desc[i + nExistingExceptions] = new InsertDescription(uncaughtExceptions[i], ""); //$NON-NLS-1$
				}

				// String label =
				// CorrectionMessages.LocalCorrectionsSubProcessor_addthrows_description;
				// Image image =
				// JavaPluginImages.get(JavaPluginImages.IMG_OBJS_EXCEPTION);
				//
				// ChangeMethodSignatureProposal proposal = new
				// ChangeMethodSignatureProposal(label, cu, astRoot, binding,
				// null,
				// desc, 8, image);
				// for (int i = 0; i < uncaughtExceptions.length; i++) {
				// addExceptionTypeLinkProposals(proposal,
				// uncaughtExceptions[i],
				// proposal.getExceptionTypeGroupId(i + nExistingExceptions));
				// }
				// proposal.setCommandId(ADD_EXCEPTION_TO_THROWS_ID);
				// proposals.add(proposal);
			}
		}

		// ------------------------------ End -------------------------//

		// e.g., catch (IOException e)
		CatchClause catchClause = ast.newCatchClause();
		SingleVariableDeclaration svd = ast.newSingleVariableDeclaration();
		// YS do more precise analysis
		svd.setType(ast.newSimpleType(ast.newSimpleName("Exception"))); //$NON-NLS-1$
		svd.setVarargs(false);
		svd.setName(ast.newSimpleName("e")); //$NON-NLS-1$
		svd.setExtraDimensions(0);
		svd.setInitializer(null);
		catchClause.setException(svd);

		// e.g., e.printStackTrace();
		Block blockInCatch = ast.newBlock();
		MethodInvocation excPrintStackTraceMI = ast.newMethodInvocation();
		excPrintStackTraceMI.setExpression(ast.newSimpleName("e")); //$NON-NLS-1$
		excPrintStackTraceMI.setName(ast.newSimpleName("printStackTrace")); //$NON-NLS-1$
		blockInCatch.statements().add(ast.newExpressionStatement(excPrintStackTraceMI));

		catchClause.setBody(blockInCatch);

		tryStatement.setBody(blockInInnerTry);
		tryStatement.catchClauses().add(catchClause);
		tryStatement.setFinally(null);

		thenBlock.statements().add(tryStatement);

		ifStatement.setExpression(infixExpression);
		ifStatement.setThenStatement(thenBlock);
		ifStatement.setElseStatement(null);

		//
		TryStatement outerTryStatement = null;
		// finallyBlock.
		List innerStatements = finallyBlock.statements();
		ListRewrite listRewrite = rewrite.getListRewrite(finallyBlock, Block.STATEMENTS_PROPERTY);
		listRewrite.insertFirst(ifStatement, null);

		// YS add throws declaration

		return ifStatement;
	}

	private static boolean isSubtype(ITypeBinding curr, final ITypeBinding[] addedExceptions) {
		while (curr != null) {
			for (int i = 0; i < addedExceptions.length; i++) {
				if (curr == addedExceptions[i]) {
					return true;
				}
			}
			curr = curr.getSuperclass();
		}
		return false;
	}

	private ITypeBinding[] findUncaughtExceptions(final CompilationUnit root, final ASTNode closeStatement) {
		List<ITypeBinding> uncaught = new ArrayList<ITypeBinding>();
		List<ImportDeclaration> imports = root.imports();
		for (ImportDeclaration imp : imports) {
			fLogger.info("\t" + imp.getName().getFullyQualifiedName());
		}

		ITypeBinding[] needToCaught = findNeedToCatch(closeStatement);
		for (ITypeBinding tb : needToCaught) {
			if (!isCaught(tb, imports)) {
				uncaught.add(tb);
			}
		}

		return uncaught.toArray(new ITypeBinding[uncaught.size()]);
	}

	private boolean isCaught(final ITypeBinding tb, final List<ImportDeclaration> imports) {
		for (ImportDeclaration id : imports) {
			if (id.getName().getFullyQualifiedName().equals(tb.getQualifiedName())) {
				return true;
			}
		}
		return false;
	}

	private ITypeBinding[] findNeedToCatch(final ASTNode closeStatement) {
		if (closeStatement instanceof ExpressionStatement) {
			ExpressionStatement exprStmt = (ExpressionStatement) closeStatement;
			Expression expr = exprStmt.getExpression();
			if (expr instanceof MethodInvocation) {
				MethodInvocation mi = (MethodInvocation) expr;
				ITypeBinding tbb = mi.resolveTypeBinding();
				fLogger.info("TB (Return Type): " + tbb.getQualifiedName());
				IMethodBinding mb = mi.resolveMethodBinding();
				fLogger.info("mi: " + ASTNodes.asString(mi) + ", MB: " + mb);
				return mb.getExceptionTypes();
			}
		}
		return null;
	}

	/**
	 * @deprecated
	 */
	@Deprecated
	private void addImport(final AST ast, final ASTRewrite rewrite, final ASTNode closeStatement, final CompilationUnit root)
			throws CoreException {
		fLogger.info("addImport: ");
		List<ImportDeclaration> imports = root.imports();
		for (ImportDeclaration imp : imports) {
			fLogger.info("\t" + imp.getName().getFullyQualifiedName());
		}

		// YS Add imports. More precise exception recommendation
		if (closeStatement instanceof ExpressionStatement) {
			ExpressionStatement exprStmt = (ExpressionStatement) closeStatement;
			Expression expr = exprStmt.getExpression();
			if (expr instanceof MethodInvocation) {
				MethodInvocation mi = (MethodInvocation) expr;
				ITypeBinding tbb = mi.resolveTypeBinding();
				fLogger.info("TB (Return Type): " + tbb.getQualifiedName());
				IMethodBinding mb = mi.resolveMethodBinding();
				fLogger.info("mi: " + ASTNodes.asString(mi) + ", MB: " + mb);
				ITypeBinding[] tbs = mb.getExceptionTypes();
				ImportRewrite importRewrite = RewriteUtil.createImportRewrite(root, true);
				MultiTextEdit multiTextEdit = new MultiTextEdit();
				for (ITypeBinding tb : tbs) {
					ImportDeclaration id = ast.newImportDeclaration();
					id.setName(ast.newName(tb.getQualifiedName()));
					String typeName = importRewrite.addImport(tb);
					fLogger.info("TypeName: " + typeName);
					multiTextEdit.addChild(importRewrite.rewriteImports(null));
					// root.imports().add(id);
					// System.out.println("\t" + tb.getQualifiedName() +
					// ", text: " + text);
				}
			}
		}
		return;
	}

	// from QuickAssistProcessor
	private static void addExceptionToThrows(AST ast, MethodDeclaration methodDeclaration, ASTRewrite rewrite, SimpleType type2) {
		fLogger.info("addExceptionToThrows: ");
		ITypeBinding binding = type2.resolveBinding();
		if (binding == null || isNotYetThrown(binding, methodDeclaration.thrownExceptions())) {
			Name name = type2.getName();
			Name newName = (Name) ASTNode.copySubtree(ast, name);

			ListRewrite listRewriter = rewrite.getListRewrite(methodDeclaration, MethodDeclaration.THROWN_EXCEPTIONS_PROPERTY);
			listRewriter.insertLast(newName, null);
		}
	}

	private static boolean isNotYetThrown(ITypeBinding binding, final List<Name> thrownExceptions) {
		fLogger.info("isNotYetThrown: ");
		for (int i = 0; i < thrownExceptions.size(); i++) {
			Name name = thrownExceptions.get(i);
			ITypeBinding elem = (ITypeBinding) name.resolveBinding();
			if (elem != null) {
				if (Bindings.isSuperType(elem, binding)) { // existing exception
															// is base class of
															// new
					return false;
				}
			}
		}
		return true;
	}

	private ImportRewrite getImportRewrite(final CompilationUnit root) {
		if (fImportRewrite == null) {
			try {
				fImportRewrite = RewriteUtil.createImportRewrite(root, true);
			} catch (Exception e) {
				throw new IllegalStateException(e.getMessage());
			}
		}

		return fImportRewrite;
	}

	private boolean existsImport(final ASTNode node, final ITypeBinding tb) {
		CompilationUnit root1 = (CompilationUnit) node.getRoot();
		List<ImportDeclaration> imports = root1.imports();
		for (ImportDeclaration imp : imports) {
			if (tb.getQualifiedName().equals(imp.getName().getFullyQualifiedName())) {
				return true;
			}
		}
		return false;
	}

	private Block createFinallyBlock(final ASTRewrite rewrite, final ASTNode node,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap,
			final List<ITrackedNodePosition> trackedVariables) {
		fLogger.info("OpenStreamExceptionPathContext#createFinallyBlock");
		TryStatement tryStatement = ASTResolving.findParentTryStatement(node);
		AST ast = tryStatement.getAST();
		Block finallyBody = null;
		if (tryStatement.getFinally() != null) {
			// finally block exists already
			finallyBody = tryStatement.getFinally();
		} else {
			finallyBody = ast.newBlock();
		}
		rewrite.set(tryStatement, TryStatement.FINALLY_PROPERTY, finallyBody, null);

		FixBugsAnnotation aa = new FixBugsAnnotation(FixBugsAnnotation.ADDED, Messages.addedAnnotation);
		trackedAnnotationMap.put(rewrite.track(finallyBody), aa);

		// YS ...
		FixBugsAnnotation ra = new FixBugsAnnotation(FixBugsAnnotation.RELATED, Messages.relatedAnnotation);
		trackedAnnotationMap.put(rewrite.track(node), ra);
		addLinkedPositions(trackedVariables, node, fOriginalIdentifier, rewrite);

		return finallyBody;
	}

	private boolean needToCloseStreamInFinallyBlock(final ASTNode node) {
		ExpressionStatement exprStmt = null;
		if (node instanceof ExpressionStatement) {
			exprStmt = (ExpressionStatement) node;
			Expression expr = exprStmt.getExpression();
			if (expr instanceof MethodInvocation) {
				MethodInvocation mi = (MethodInvocation) expr;
				if ("close".equals(mi.getName().getIdentifier())) { //$NON-NLS-1$
					return true;
				}
			}
		}

		return false;
	}

	private boolean isCloseable(final ASTNode node) {
		ExpressionStatement exprStmt = null;
		if (node instanceof ExpressionStatement) {
			exprStmt = (ExpressionStatement) node;
			Expression expr = exprStmt.getExpression();
			if (expr instanceof MethodInvocation) {
				MethodInvocation mi = (MethodInvocation) expr;
				Expression varEx = mi.getExpression();
				// System.out.println("VarEx: " + varEx);
				ITypeBinding tb = varEx.resolveTypeBinding();
				ITypeBinding[] superTypes = Bindings.getAllSuperTypes(tb);
				for (ITypeBinding superType : superTypes) {
					// System.out.println("SuperType: " + superType);
					if (superType.getName().contains("java.io.Closeable")) {
						return true;
					}
				}
			}
		}

		return false;
	}

	private boolean isAutoCloseable(final ASTNode node) {
		ExpressionStatement exprStmt = null;
		if (node instanceof ExpressionStatement) {
			exprStmt = (ExpressionStatement) node;
			Expression expr = exprStmt.getExpression();
			if (expr instanceof MethodInvocation) {
				MethodInvocation mi = (MethodInvocation) expr;
				Expression varEx = mi.getExpression();
				// System.out.println("VarEx: " + varEx);
				ITypeBinding tb = varEx.resolveTypeBinding();
				ITypeBinding[] superTypes = Bindings.getAllSuperTypes(tb);
				for (ITypeBinding superType : superTypes) {
					// System.out.println("SuperType: " + superType);
					if (superType.getName().contains("java.lang.AutoCloseable")) { //$NON-NLS-1$
						return true;
					}
				}
			}
		}

		return false;
	}

	private boolean isStreamClose(final ASTNode node) {
		ExpressionStatement exprStmt = null;
		if (node instanceof ExpressionStatement) {
			exprStmt = (ExpressionStatement) node;
			Expression expr = exprStmt.getExpression();
			if (expr instanceof MethodInvocation) {
				MethodInvocation mi = (MethodInvocation) expr;
				if ("close".equals(mi.getName().getIdentifier())) { //$NON-NLS-1$
					// isAutoCloseable(node);
					return true;
				}
			}
		}

		return false;
	}

	/**
	 * Joins and Moves the I/O Stream, and then skip or remove it.
	 * 
	 * @param rewrite
	 * @param root
	 * @param firstLinkedNode
	 *            the first variable node.
	 * @param outerTryStatement
	 *            the try statement.
	 * @param nodes
	 *            related nodes.
	 * @param trackedAnnotationMap
	 *            tracked annotation map.
	 * @param trackedVariables
	 * @return
	 */
	private boolean joinMoveStreamAndSkipOrRemove(final ASTRewrite rewrite, final CompilationUnit root,
			final ASTNode firstLinkedNode, final TryStatement outerTryStatement, final ASTNode[] nodes,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap,
			final List<ITrackedNodePosition> trackedVariables) {
		ASTNode parent = firstLinkedNode.getParent();

		VariableDeclarationFragment fragment = null;
		boolean onFirstAccess = false;
		if (firstLinkedNode instanceof SimpleName
				&& firstLinkedNode.getLocationInParent() == Assignment.LEFT_HAND_SIDE_PROPERTY) {
			onFirstAccess = true;
			SimpleName name = (SimpleName) firstLinkedNode;
			IBinding binding = name.resolveBinding();
			if (!(binding instanceof IVariableBinding)) {
				return false;
			}
			ASTNode declaring = root.findDeclaringNode(binding);
			if (declaring instanceof VariableDeclarationFragment) {
				fragment = (VariableDeclarationFragment) declaring;
			} else {
				return false;
			}
		} else if (parent instanceof VariableDeclarationFragment) {
			fragment = (VariableDeclarationFragment) parent;
		} else {
			return false;
		}

		IVariableBinding binding = fragment.resolveBinding();
		Expression initializer = fragment.getInitializer();
		if ((initializer != null && initializer.getNodeType() != ASTNode.NULL_LITERAL) || binding == null || binding.isField()) {
			return false;
		}

		if (!(fragment.getParent() instanceof VariableDeclarationStatement)) {
			return false;
		}
		VariableDeclarationStatement statement = (VariableDeclarationStatement) fragment.getParent();

		SimpleName[] names = LinkedNodeFinder.findByBinding(statement.getParent(), binding);
		if (names.length <= 1 || names[0] != fragment.getName()) {
			return false;
		}
		SimpleName firstAccess = names[1];
		if (onFirstAccess) {
			if (firstAccess != firstLinkedNode) {
				return false;
			}
		} else {
			if (firstAccess.getLocationInParent() != Assignment.LEFT_HAND_SIDE_PROPERTY) {
				return false;
			}
		}
		Assignment assignment = (Assignment) firstAccess.getParent();
		if (assignment.getLocationInParent() != ExpressionStatement.EXPRESSION_PROPERTY) {
			return false;
		}
		ExpressionStatement assignParent = (ExpressionStatement) assignment.getParent();

		AST ast = statement.getAST();

		TightSourceRangeComputer sourceRangeComputer = new TightSourceRangeComputer();
		sourceRangeComputer.addTightSourceNode(assignParent);
		rewrite.setTargetSourceRangeComputer(sourceRangeComputer);

		Expression placeholder = (Expression) rewrite.createMoveTarget(assignment.getRightHandSide());

		VariableDeclarationFragment newFragment = ast.newVariableDeclarationFragment();
		SimpleName variableName = ast.newSimpleName(((SimpleName) firstLinkedNode).getIdentifier());
		newFragment.setName(variableName);
		newFragment.setInitializer(placeholder);
		VariableDeclarationExpression newExpression = ast.newVariableDeclarationExpression(newFragment);
		Type varType = (Type) ASTNode.copySubtree(ast, statement.getType());
		newExpression.setType(varType);

		ListRewrite listRewrite = rewrite.getListRewrite(outerTryStatement, TryStatement.RESOURCES_PROPERTY);
		listRewrite.insertFirst(newExpression, null);
		FixBugsAnnotation addedAfterAnnotation = new FixBugsAnnotation(FixBugsAnnotation.ADDED, Messages.addedAnnotation);
		trackedAnnotationMap.put(rewrite.track(newExpression), addedAfterAnnotation);

		trackedVariables.add(rewrite.track(variableName));

		// different part
		Statement firstStatement = ASTResolving.findParentStatement(firstLinkedNode);
		Statement secondStatement = ASTResolving.findParentStatement(firstAccess);

		for (ASTNode part : nodes) {
			if (part == firstStatement) {
				rewrite.remove(firstStatement, null);
			} else if (part == secondStatement) {
				rewrite.remove(secondStatement, null);
			} else if (isStreamClose(part)) {
				rewrite.remove(part, null);
			} else {
				FixBugsAnnotation ra = new FixBugsAnnotation(FixBugsAnnotation.RELATED, Messages.relatedAnnotation);
				trackedAnnotationMap.put(rewrite.track(part), ra);
				addLinkedPositions(trackedVariables, part, fOriginalIdentifier, rewrite);
			}
		}

		return true;
	}

	/**
	 * Moves the stream and skip or remove it.
	 * 
	 * @param rewrite
	 * @param root
	 * @param firstLinkedNode
	 *            the first linked node.
	 * @param tryStatement
	 *            the try statement.
	 * @param nodes
	 *            the related nodes.
	 * @param trackedAnnotationMap
	 * @param trackedVariables
	 */
	private void moveStreamAndSkipOrRemove(final ASTRewrite rewrite, final CompilationUnit root, final ASTNode firstLinkedNode,
			final TryStatement tryStatement, final ASTNode[] nodes,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap,
			final List<ITrackedNodePosition> trackedVariables) {
		// YS Implement this method
		Statement node = ASTResolving.findParentStatement(firstLinkedNode);

		if (node instanceof VariableDeclarationStatement) {
			VariableDeclarationStatement vds = (VariableDeclarationStatement) node;

			ListRewrite listRewrite = rewrite.getListRewrite(tryStatement, TryStatement.RESOURCES_PROPERTY);
			listRewrite.insertFirst(vds, null);
			FixBugsAnnotation ma = new FixBugsAnnotation(FixBugsAnnotation.MODIFIED, Messages.modifiedAnnotation);
			trackedAnnotationMap.put(rewrite.track(vds), ma);
			trackedVariables.add(rewrite.track(firstLinkedNode));

			for (ASTNode part : nodes) {
				if (node == part) {
					rewrite.remove(part, null);
				} else if (isStreamClose(part)) {
					rewrite.remove(part, null);
				} else {
					FixBugsAnnotation ra = new FixBugsAnnotation(FixBugsAnnotation.RELATED, Messages.relatedAnnotation);
					trackedAnnotationMap.put(rewrite.track(part), ra);
					addLinkedPositions(trackedVariables, part, fOriginalIdentifier, rewrite);
				}
			}
		}

		// ASTNode parent = node.getParent();
	}

	private boolean canJoinVariable(final CompilationUnit root, final ASTNode node) {
		ASTNode parent = node.getParent();

		VariableDeclarationFragment fragment = null;
		boolean onFirstAccess = false;
		if (node instanceof SimpleName && node.getLocationInParent() == Assignment.LEFT_HAND_SIDE_PROPERTY) {
			onFirstAccess = true;
			SimpleName name = (SimpleName) node;
			IBinding binding = name.resolveBinding();
			if (!(binding instanceof IVariableBinding)) {
				return false;
			}
			ASTNode declaring = root.findDeclaringNode(binding);
			if (declaring instanceof VariableDeclarationFragment) {
				fragment = (VariableDeclarationFragment) declaring;
			} else {
				return false;
			}
		} else if (parent instanceof VariableDeclarationFragment) {
			fragment = (VariableDeclarationFragment) parent;
		} else {
			return false;
		}

		IVariableBinding binding = fragment.resolveBinding();
		Expression initializer = fragment.getInitializer();
		if ((initializer != null && initializer.getNodeType() != ASTNode.NULL_LITERAL) || binding == null || binding.isField()) {
			return false;
		}

		if (!(fragment.getParent() instanceof VariableDeclarationStatement)) {
			return false;
		}
		VariableDeclarationStatement statement = (VariableDeclarationStatement) fragment.getParent();

		SimpleName[] names = LinkedNodeFinder.findByBinding(statement.getParent(), binding);
		if (names.length <= 1 || names[0] != fragment.getName()) {
			return false;
		}
		SimpleName firstAccess = names[1];
		if (onFirstAccess) {
			if (firstAccess != node) {
				return false;
			}
		} else {
			if (firstAccess.getLocationInParent() != Assignment.LEFT_HAND_SIDE_PROPERTY) {
				return false;
			}
		}
		Assignment assignment = (Assignment) firstAccess.getParent();
		if (assignment.getLocationInParent() != ExpressionStatement.EXPRESSION_PROPERTY) {
			return false;
		}

		return true;
	}

	private static TryStatement findCoveringTryStatement(final ASTNode[] nodes) {
		for (ASTNode node : nodes) {
			TryStatement tryStatement = ASTResolving.findParentTryStatement(node);
			if (tryStatement != null) {
				return tryStatement;
			}
		}
		return null;
	}

	/**
	 * Finds the first occurrence of the given variable identifier in fragments
	 * of a variable declaration statement.
	 * 
	 * @param vds
	 *            the statement.
	 * @param identifier
	 *            the identifier.
	 * @return new VariableDeclarationFragment
	 */
	private static VariableDeclarationFragment findFirstVariableDeclarationFragment(final VariableDeclarationStatement vds,
			final String identifier) {
		List fragments = vds.fragments();
		for (Object fragment : fragments) {
			VariableDeclarationFragment vdf = (VariableDeclarationFragment) fragment;
			if (identifier.equals(vdf.getName().getIdentifier())) {
				return vdf;
			}
		}
		return null;
	}

}
