//
// Copyright (c) 2007-2012 Yoonki Song.
// All rights reserved. This program and the accompanying materials
// are made available under the permission of me.
//
// Contributors:
// Yoonki Song - initial API and implementation
//
package edu.ncsu.csc.emh.fixbugs.core.resolution.correctness.missingArgument;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jface.text.Position;

/**
 * <code>VariableDropPositionFinder</code>.
 * 
 * @author Yoonki Song
 * 
 */
public final class VariableDropPositionFinder {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(VariableDropPositionFinder.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	private VariableDropPositionFinder() {
		// no instance
	}

	/**
	 * Finds drop positions for annotations being dragged.
	 * 
	 * @param assistant
	 *            the current context
	 * @param annotation
	 *            the source annotation
	 * @return the array of drop positions.
	 */
	public static VariableDropPosition[] perform(final MissingArgumentAssistant assistant, final Position position,
			final String name) {
		fLogger.info("VariableDropPositionFinder#perform: " + position + ", " + "name: " + name);
		List<VariableDropPosition> positions = new ArrayList<VariableDropPosition>();
		Position presentedDropPosition = assistant.getPresentedVariablePosition();
		VariableDropPosition dropPosition = null;
		if (presentedDropPosition == null) {
			Position p = assistant.getLastExistingArgumentPosition();
			dropPosition = new VariableDropPosition(p.getOffset() + p.getLength(), 1, name);
		} else {
			dropPosition = new VariableDropPosition(presentedDropPosition, name);
		}
		positions.add(dropPosition);

		// add itself
		VariableDropPosition dragPosition = new VariableDropPosition(position, name);
		positions.add(dragPosition);
		return positions.toArray(new VariableDropPosition[] {});
	}
}
