//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.BodyDeclaration;
import org.eclipse.jdt.core.dom.CatchClause;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SimpleType;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.TryStatement;
import org.eclipse.jdt.core.dom.UnionType;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite.ImportRewriteContext;
import org.eclipse.jdt.core.dom.rewrite.ListRewrite;
import org.eclipse.jdt.internal.corext.codemanipulation.ContextSensitiveImportRewriteContext;
import org.eclipse.jdt.internal.corext.codemanipulation.StubUtility;
import org.eclipse.jdt.internal.corext.dom.Bindings;
import org.eclipse.jdt.internal.corext.util.JavaModelUtil;
import org.eclipse.jdt.internal.ui.text.correction.ASTResolving;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Position;

import edu.ncsu.csc.emh.fixbugs.core.annotation.ExceptionAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.annotation.MethodCallAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.util.PrintUtil;
import edu.ncsu.csc.emh.fixbugs.ui.Messages;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.dodgyCode.catchException.CatchExceptionMode;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.dodgyCode.catchException.ExceptionColorRegistry;

/**
 * @author Yoonki
 */
@SuppressWarnings("restriction")
public final class CatchExceptionAssistant extends FixBugsAssistant {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(CatchExceptionAssistant.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** the annotation for an exception may thrown by multiple methods. */
	private String fMultiCallerAnnotation = null;

	/** pre-loaded exception annotation. */
	private List<String> fExceptionAnnotations = null;

	/** pre-loaded method annotation. */
	private List<String> fMethodAnnotations = null;

	/**  */
	private List<MethodsExceptionsPair> fOriginalCaughtExceptions = null;

	/** */
	private List<String> fOriginalThrownExceptions = null;

	/** */
	private List<ASTNode[]> fInitialAnnotationNodes;

	/**  */
	private ASTNode[] fMethodsThrowingExceptions;

	/** */
	private static final int TOTAL_ANNOTATIONS = 11;

	/**
	 * Constructor for CatchExceptionContext.
	 *
	 * @param context
	 *            the invocation context.
	 */
	public CatchExceptionAssistant(final FixBugsContext context) {
		super(context);

		// XXX what if the number of methods throwing exceptions exceed 10?
		fMultiCallerAnnotation = FixBugsAnnotation.MULTI_CALLER;

		fExceptionAnnotations = new ArrayList<String>();
		fExceptionAnnotations.add(FixBugsAnnotation.KIND1P);
		fExceptionAnnotations.add(FixBugsAnnotation.KIND2P);
		fExceptionAnnotations.add(FixBugsAnnotation.KIND3P);
		fExceptionAnnotations.add(FixBugsAnnotation.KIND4P);
		fExceptionAnnotations.add(FixBugsAnnotation.KIND5P);
		fExceptionAnnotations.add(FixBugsAnnotation.KIND6P);
		fExceptionAnnotations.add(FixBugsAnnotation.KIND7P);
		fExceptionAnnotations.add(FixBugsAnnotation.KIND8P);
		fExceptionAnnotations.add(FixBugsAnnotation.KIND9P);
		fExceptionAnnotations.add(FixBugsAnnotation.KIND10P);
		fExceptionAnnotations.add(FixBugsAnnotation.KIND11P);

		fMethodAnnotations = new ArrayList<String>();
		fMethodAnnotations.add(FixBugsAnnotation.KIND1S);
		fMethodAnnotations.add(FixBugsAnnotation.KIND2S);
		fMethodAnnotations.add(FixBugsAnnotation.KIND3S);
		fMethodAnnotations.add(FixBugsAnnotation.KIND4S);
		fMethodAnnotations.add(FixBugsAnnotation.KIND5S);
		fMethodAnnotations.add(FixBugsAnnotation.KIND6S);
		fMethodAnnotations.add(FixBugsAnnotation.KIND7S);
		fMethodAnnotations.add(FixBugsAnnotation.KIND8S);
		fMethodAnnotations.add(FixBugsAnnotation.KIND9S);
		fMethodAnnotations.add(FixBugsAnnotation.KIND10S);
		fMethodAnnotations.add(FixBugsAnnotation.KIND11S);
	}

	public List<MethodsExceptionsPair> getOriginalCaughtExceptions() {
		return fOriginalCaughtExceptions;
	}

	public List<String> getOriginalThrownExceptions() {
		return fOriginalThrownExceptions;
	}

	/**
	 * Finds the first annotation of method name in try statement.
	 *
	 * @return
	 */
	public FixBugsAnnotation findFirstAnnotationInTryStatement() {
		fLogger.info("#getFirstAnnotationInTryStatement");
		for (FixBugsAnnotation annotation : getAnnotations()) {
			if (annotation instanceof MethodCallAnnotation) {
				// it could be better by checking whether
				// the annotation is in try statement
				return annotation;
			}
		}

		// just in case, return the first annotations
		// it may be incorrect!!
		return getAnnotations().get(0);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init() throws BadLocationException, JavaModelException {
		super.init();
		initExceptionInfo();
		initAnnotations();
	}

	@Override
	public void initAnnotations() {
		fLogger.info("CatchExceptionContext#initAnnotations");
		initAnnotationNodes();
		initAnnotationTypes();
	}

	private void initAnnotationNodes() {
		Assert.isNotNull(fMethodsThrowingExceptions); // pre-condition
		List<ASTNode> annotationNodes = new ArrayList<ASTNode>();
		for (ASTNode node : fMethodsThrowingExceptions) {
			if (!annotationNodes.contains(node)) {
				annotationNodes.add(node); // do nothing!
			}
		}
		setAnnotationNodes(annotationNodes.toArray(new ASTNode[annotationNodes.size()]));
	}

	private void initAnnotationTypes() {
		ASTNode[] annotationNodes = getAnnotationNodes();
		Assert.isNotNull(annotationNodes); // pre-condition
		List<FixBugsAnnotation> annotations = new ArrayList<FixBugsAnnotation>();
		int index = 0;
		for (ASTNode annoNode : annotationNodes) {
			Position p = getNodePosition(annoNode);
			// XXX what if the number of methods throwing exceptions exceed 5?
			String type = fMethodAnnotations.get(index++ % TOTAL_ANNOTATIONS);
			annotations.add(new FixBugsAnnotation(type, p));
		}
		setAnnotations(annotations);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see edu.ncsu.csc.emh.fixbugs.ui.fix.FixBugsInvocationContext#
	 * initializeLinkedNodes()
	 */
	private void initExceptionInfo() {
		fLogger.info("CatchExceptionAssistant#initExceptionInfo");
		ASTNode coveringNode = getContext().getCoveringNode();
		Assert.isNotNull(coveringNode);

		fLogger.fine(PrintUtil.toString("coveringNode", coveringNode));
		if (coveringNode instanceof SimpleName) { // ASTNode of method name
			TryStatement tryStatement = ASTResolving.findParentTryStatement(coveringNode);
			coveringNode = tryStatement.getBody();
		}

		if (coveringNode instanceof Block) {
			List<Statement> statements = ((Block) coveringNode).statements();
			fLogger.info("statements.size(): " + statements.size());

			fMethodsThrowingExceptions = ExceptionCollector.perform(statements);

			// sort by the starting position
			Arrays.sort(fMethodsThrowingExceptions, new Comparator<ASTNode>() {
				@Override
				public int compare(final ASTNode n1, final ASTNode n2) {
					return n1.getStartPosition() - n2.getStartPosition();
				}
			});
		}

		// Find already thrown exceptions
		fOriginalThrownExceptions = findAlreadyThrownExceptions(coveringNode);

		//
		LinkedList<MethodsExceptionsPair> tempCaughtExceptions = new LinkedList<MethodsExceptionsPair>();
		List<ITypeBinding> visitedExceptionTypes = new ArrayList<ITypeBinding>();
		boolean subTypeFirst = false;
		Assert.isNotNull(fMethodsThrowingExceptions);
		for (ASTNode node : fMethodsThrowingExceptions) {
			List<String> exceptions = new ArrayList<String>();
			IMethodBinding mb = findMethodBinding(node);
			if (mb != null) {
				ITypeBinding[] tbs = mb.getExceptionTypes();
				fLogger.fine("\tMb: " + mb.getName());
				for (ITypeBinding tb : tbs) {
					fLogger.fine("\t\tEx: " + tb.getName());

					// check redundancy
					if (visitedExceptionTypes.contains(tb)) {
						// System.out.println("\t\t Redundant");
						//MethodExceptionsRegistry.getInstance().markAsMultiCaller(tb.getName());
						continue;
					}

					if (isSuperType(visitedExceptionTypes, tb)) {
						if (!isVisited(visitedExceptionTypes, tb)) {
							fLogger.warning("Check it out!! " + tb.getName());
							exceptions.add(tb.getName());
							visitedExceptionTypes.add(tb);
							subTypeFirst = true;
						}
					} else { // normal case
						exceptions.add(tb.getName());
						visitedExceptionTypes.add(tb);
					}
				}
				// if there are no exceptions to catch, ignore it.
				if (!exceptions.isEmpty()) {
					List<String> methods = new ArrayList<String>();
					methods.add(mb.getName());
					MethodsExceptionsPair newPair = new MethodsExceptionsPair(methods, exceptions);
					if (subTypeFirst) {
						MethodsExceptionsPair lastPair = tempCaughtExceptions.removeLast();
						tempCaughtExceptions.add(newPair);
						tempCaughtExceptions.add(lastPair);
					} else { // normal case
						tempCaughtExceptions.add(newPair);
					}

				}
			}
		}

		// Simplify
		fOriginalCaughtExceptions = simplifyExceptionPair(tempCaughtExceptions);
		// System.out.println("[Simplified]");
		// for (MethodsExceptionsPair pair : fOriginalCaughtExceptions) {
		// System.out.println(pair.toString());
		// }
	}

	/**
	 * Checks if the given type biding is a sub type of existing type bindings.
	 *
	 * @param visitedExceptionTypes
	 * @param type
	 * @return true if the given type biding is a sub type of existing type
	 *         bindings; false otherwise.
	 */
	private boolean isSuperType(final List<ITypeBinding> visitedExceptionTypes, final ITypeBinding type) {
		for (ITypeBinding possible : visitedExceptionTypes) {
			if (possible.getName().equals(type.getName())) {
				// if two type bindings are the same, ignore it.
				continue;
			}
			if (Bindings.isSuperType(possible, type)) {
				fLogger.fine("\t\tType (is-a): " + type.getName());
				return true;
			}
		}
		return false;
	}

	/**
	 * Checks if the given type biding is a sub type of existing type bindings.
	 *
	 * @param visitedExceptionTypes
	 * @param type
	 * @return true if the given type biding is a sub type of existing type
	 *         bindings; false otherwise.
	 */
	private boolean isVisited(final List<ITypeBinding> visitedExceptionTypes, final ITypeBinding type) {
		for (ITypeBinding possible : visitedExceptionTypes) {
			if (possible.getName().equals(type.getName())) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Checks if the given exception may thrown by multiple methods (including
	 * constructors) in try statement.
	 * <p>
	 * Input: e1
	 *
	 * <pre>
	 * [[m1], [e1]], [[m2], [e2, e3]]
	 * </pre>
	 *
	 * Output: false
	 *
	 * <p>
	 * Input: e1
	 *
	 * <pre>
	 * [[m1, m2], [e1]], [[m2], [e2, e3]]
	 * </pre>
	 *
	 * Output: true
	 *
	 * @param exceptionName
	 * @return
	 */
	private boolean hasMultiCallers(final String exceptionName) {
		for (MethodsExceptionsPair pair : fOriginalCaughtExceptions) {
			if (pair.containsException(exceptionName)) {
				if (pair.getMethods().size() > 1) {
					return true;
				}
			}
		}
		return false;
	}

	public static List<MethodsExceptionsPair> simplifyExceptionPair(final List<MethodsExceptionsPair> caughtExceptions) {
		List<MethodsExceptionsPair> newCaughtExceptions = new ArrayList<MethodsExceptionsPair>();
		List<MethodsExceptionsPair> visitedPairList = new ArrayList<MethodsExceptionsPair>();
		for (MethodsExceptionsPair sourcePair : caughtExceptions) {
			if (visitedPairList.contains(sourcePair)) {
				continue;
			}
			String sourceExListAsString = sourcePair.getExceptionsAsString();
			List<String> newMethodList = new ArrayList<String>();
			newMethodList.addAll(sourcePair.getMethods());
			for (MethodsExceptionsPair targetPair : caughtExceptions) {
				if (sourcePair == targetPair) {
					continue;
				}
				// in case, different!!
				String targetExListAsString = targetPair.getExceptionsAsString();
				if (sourceExListAsString.equals(targetExListAsString)) {
					newMethodList.addAll(targetPair.getMethods());
					visitedPairList.add(targetPair);
				}
			}
			MethodsExceptionsPair newPair = new MethodsExceptionsPair(newMethodList, sourcePair.getExceptions());
			newCaughtExceptions.add(newPair);
		}

		return newCaughtExceptions;
	}

	/**
	 * updates annotations.
	 *
	 * @param exceptionAnnotations
	 *            list of Exception Annotations.
	 * @param methodAnnotations
	 *            list of Method Annotations.
	 */
	public void update(final List<FixBugsAnnotation> exceptionAnnotations, final List<FixBugsAnnotation> methodAnnotations) {
		long startTime = System.currentTimeMillis();
		final List<FixBugsAnnotation> annotations = new ArrayList<FixBugsAnnotation>();
		annotations.addAll(methodAnnotations);
		annotations.addAll(exceptionAnnotations);
		for (FixBugsAnnotation exception : exceptionAnnotations) {
			FixBugsAnnotation draggable = new FixBugsAnnotation(FixBugsAnnotation.DRAGGABLE, exception.getPosition()); // initial
			annotations.add(draggable);
		}
		setAnnotations(annotations); // update
		fLogger.info("#update() Time Taken: " + (System.currentTimeMillis() - startTime) + " ms");
	}

	/**
	 * Applies...
	 *
	 * @param suggestion
	 *            the suggestion.
	 * @param exceptionModel
	 *            the exception model for ...
	 * @param trackedMethodNameAnnotationMap
	 *            tracked method name annotation map.
	 * @param trackedExceptionNameAnnotationMap
	 *            tracked exception name annotation map.
	 * @return the changed text
	 *
	 * @throws BadLocationException
	 *             is thrown if one of the edits in the tree can't be executed.
	 *             The state of the document is undefined if this exception is
	 *             thrown.
	 * @throws CoreException
	 *             the exception is thrown if the rewrite fails.
	 */
	public String changeAST(final String suggestion, final ExceptionDataModel exceptionModel,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedMethodNameAnnotationMap,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedExceptionNameAnnotationMap) throws BadLocationException,
			CoreException {
		fLogger.info("CatchExceptionContext#changeAST");
		String changedText = null;
		if (CatchExceptionMode.UNICATCH_EXCEPTIONS.equals(suggestion)) {
			changedText = changeASTforUniCatch(exceptionModel, trackedMethodNameAnnotationMap,
					trackedExceptionNameAnnotationMap);
		} else if (CatchExceptionMode.MULTICATCH_EXCEPTIONS.equals(suggestion)) {
			changedText = changeASTforMultiCatch(exceptionModel, trackedMethodNameAnnotationMap,
					trackedExceptionNameAnnotationMap);
		} else if (CatchExceptionMode.THROW_EXCEPTIONS.equals(suggestion)) {
			changedText = changeASTforThrow(exceptionModel, trackedMethodNameAnnotationMap, trackedExceptionNameAnnotationMap);
		} else { // CUSTOMIZED by DND
			changedText = changeASTforMultiCatch(exceptionModel, trackedMethodNameAnnotationMap,
					trackedExceptionNameAnnotationMap);
		}
		return changedText;
	}

	private String changeASTforMultiCatch(final ExceptionDataModel exceptionModel,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedMethodNameAnnotationMap,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedExceptionNameAnnotationMap) throws BadLocationException,
			CoreException {
		fLogger.info("CatchExceptionContext#changeASTforMultiCatch");

		// get DOM/AST from a ICompilationUnit
		CompilationUnit root = getContext().getASTRoot();

		// creation of ASTRewrite
		AST ast = root.getAST();
		ASTRewrite rewrite = ASTRewrite.create(ast);
		ICompilationUnit cu = getContext().getCompilationUnit();

		// ----------------------------- Change -----------------------------//

		ASTNode[] annotationNodes = getAnnotationNodes();
		TryStatement surroundingTry = ASTResolving.findParentTryStatement(annotationNodes[0]);
		MethodDeclaration methodDecl = (MethodDeclaration) ASTResolving.findParentBodyDeclaration(annotationNodes[0]);

		// track the linked nodes and annotations for method name
		int i = 0;
		for (ASTNode annotationNode : annotationNodes) {
			SimpleName methodNode = (SimpleName) annotationNode;
			String exceptions = exceptionModel.getExceptionsByMethod(methodNode.getFullyQualifiedName());
			addMethodNameInTryBlock(ast, rewrite, methodNode, trackedMethodNameAnnotationMap, i++, exceptions);
		}

		deleteCatchClause(rewrite, surroundingTry);

		// change (catch clause)
		fLogger.info("Catch: \n");
		if (surroundingTry != null) {
			ImportRewrite importRewrite = StubUtility.createImportRewrite(root, true);
			ImportRewriteContext importRewriteContext = new ContextSensitiveImportRewriteContext(methodDecl, importRewrite);

			int index = 0;
			for (MethodsExceptionsPair pair : exceptionModel.getMethodsExceptionsToCatch()) {
				String varName = "e"; // use default name 'e';
										// //StubUtility.getExceptionVariableName(cu.getJavaProject());
				SingleVariableDeclaration svd = ast.newSingleVariableDeclaration();
				svd.setName(ast.newSimpleName(varName));

				List<String> exceptionNames = pair.getExceptions();
				if (exceptionNames.size() == 1) {
					String exceptionName = exceptionNames.get(0);
					ITypeBinding typeBinding = findTypeBinding(annotationNodes, exceptionName);
					SimpleType exceptionType = (SimpleType) importRewrite.addImport(typeBinding, ast, importRewriteContext);

					// [track]
					String methods = exceptionModel.getMethodsByException(exceptionNames.get(0));
					addExceptionToCatch(ast, rewrite, exceptionType, trackedExceptionNameAnnotationMap, index, true, false,
							methods);
					svd.setType(exceptionType);
				} else if (exceptionNames.size() > 1) {
					UnionType unionType = ast.newUnionType();
					List typeList = unionType.types(); // now, it is empty
					for (String exceptionName : exceptionNames) {
						ITypeBinding typeBinding = findTypeBinding(annotationNodes, exceptionName);
						SimpleType exceptionType = (SimpleType) importRewrite.addImport(typeBinding, ast, importRewriteContext);

						// [track]
						String methods = exceptionModel.getMethodsByException(exceptionName);
						addExceptionToCatch(ast, rewrite, exceptionType, trackedExceptionNameAnnotationMap, index, true, true,
								methods);
						typeList.add(exceptionType);
					}
					svd.setType(unionType);
				}

				CatchClause newCatchClause = ast.newCatchClause();
				newCatchClause.setException(svd);
				String catchBody = StubUtility.getCatchBodyContent(cu, exceptionNames.get(0), varName, annotationNodes[0],
						String.valueOf('\n'));
				if (catchBody != null) {
					ASTNode node = rewrite.createStringPlaceholder(catchBody, ASTNode.RETURN_STATEMENT);
					newCatchClause.getBody().statements().add(node);
				}
				ListRewrite clauseRewrite = rewrite.getListRewrite(surroundingTry, TryStatement.CATCH_CLAUSES_PROPERTY);
				clauseRewrite.insertLast(newCatchClause, null);
				index++;
			}

			//
			if (JavaModelUtil.is17OrHigher(cu.getJavaProject())) {
				// YS do something
			}
		} else {
			fLogger.warning("Something wrong: x");
		}

		// change (throws)
		fLogger.info("Throws: \n");
		if (surroundingTry != null) {
			for (ExceptionsMethodsPair pair : exceptionModel.getExceptionsMethodsToThrow()) {
				ITypeBinding typeBinding = findTypeBinding(annotationNodes, pair.getExceptions().get(0));
				SimpleName exceptionName = ast.newSimpleName(typeBinding.getName());
				SimpleType exceptionType = ast.newSimpleType(exceptionName);
				String methods = exceptionModel.getMethodsByException(exceptionName.getIdentifier());
				addExceptionToThrows(ast, methodDecl, rewrite, exceptionType, trackedExceptionNameAnnotationMap, methods);
			}
		}
		// System.out.println("CU:\n" + rewriteCU(rewrite));
		fRewrite = rewrite; // (important for quickfix)
		return rewriteMethodDecl(rewrite);
	}

	private String changeASTforUniCatch(final ExceptionDataModel exceptionModel,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedMethodNameAnnotationMap,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedExceptionNameAnnotationMap) throws BadLocationException,
			CoreException {
		fLogger.info("CatchExceptionContext#changeASTUniCatch");

		// get DOM/AST from a ICompilationUnit
		CompilationUnit root = getContext().getASTRoot();

		// creation of ASTRewrite
		AST ast = root.getAST();
		ASTRewrite rewrite = ASTRewrite.create(root.getAST());
		ICompilationUnit cu = getContext().getCompilationUnit();

		System.out.println("ExceptionDataModel (UniCatchException):\n" + exceptionModel);
		// ----------------------------- Change -----------------------------//

		ASTNode[] annotationNodes = getAnnotationNodes();
		TryStatement surroundingTry = ASTResolving.findParentTryStatement(annotationNodes[0]);
		MethodDeclaration methodDecl = (MethodDeclaration) ASTResolving.findParentBodyDeclaration(annotationNodes[0]);

		// track the linked nodes and annotations for method name
		int i = 0;
		for (ASTNode annotationNode : annotationNodes) {
			SimpleName methodNode = (SimpleName) annotationNode;
			String exceptions = exceptionModel.getExceptionsByMethod(methodNode.getFullyQualifiedName());
			addMethodNameInTryBlock(ast, rewrite, methodNode, trackedMethodNameAnnotationMap, i++, exceptions);
		}

		Assert.isNotNull(rewrite);
		deleteCatchClause(rewrite, surroundingTry);

		// change (catch clause)
		fLogger.info("Catch: \n");
		if (surroundingTry != null) {
			ImportRewrite importRewrite = StubUtility.createImportRewrite(root, true);
			ImportRewriteContext importRewriteContext = new ContextSensitiveImportRewriteContext(methodDecl, importRewrite);

			int index = 0;
			List<String> exceptionNames = getUniCatchExceptions(exceptionModel);
			for (String exceptionName : exceptionNames) {
				String varName = "e"; // use default name 'e';
				// //StubUtility.getExceptionVariableName(cu.getJavaProject());
				SingleVariableDeclaration svd = ast.newSingleVariableDeclaration();
				svd.setName(ast.newSimpleName(varName));

				ITypeBinding typeBinding = findTypeBinding(annotationNodes, exceptionName);
				SimpleType exceptionType = (SimpleType) importRewrite.addImport(typeBinding, ast, importRewriteContext);

				// [track]
				String methods = exceptionModel.getMethodsByException(exceptionNames.get(0));
				addExceptionToCatch(ast, rewrite, exceptionType, trackedExceptionNameAnnotationMap, index, true, false, methods);

				svd.setType(exceptionType);

				CatchClause newCatchClause = ast.newCatchClause();
				newCatchClause.setException(svd);
				String catchBody = null;
				try {
					catchBody = StubUtility.getCatchBodyContent(cu, exceptionNames.get(0), varName, annotationNodes[0],
							String.valueOf('\n'));
				} catch (CoreException e) {
					e.printStackTrace();
				}
				if (catchBody != null) {
					ASTNode node = rewrite.createStringPlaceholder(catchBody, ASTNode.RETURN_STATEMENT);
					newCatchClause.getBody().statements().add(node);
				}
				ListRewrite clauseRewrite = rewrite.getListRewrite(surroundingTry, TryStatement.CATCH_CLAUSES_PROPERTY);
				clauseRewrite.insertLast(newCatchClause, null);
				index++;
			}

			//
			if (JavaModelUtil.is17OrHigher(cu.getJavaProject())) {
				// YS do something
			}
		} else {
			fLogger.warning("Something wrong: x");
		}

		fRewrite = rewrite; // (important for quickfix)
		return rewriteMethodDecl(rewrite);
	}

	private String changeASTforThrow(final ExceptionDataModel exceptionModel,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedMethodNameAnnotationMap,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedExceptionNameAnnotationMap) throws BadLocationException,
			CoreException {
		fLogger.info("CatchExceptionContext#changeASTforThrow");

		// get DOM/AST from a ICompilationUnit
		CompilationUnit root = getContext().getASTRoot();

		// creation of ASTRewrite
		AST ast = root.getAST();
		ASTRewrite rewrite = ASTRewrite.create(root.getAST());

		System.out.println("ExceptionDataModel (ThrowExceptionsResolution):\n" + exceptionModel);

		// --------------------- Pre-conditions -----------------------//

		int offset = getContext().getSelectionOffset();
		int length = getContext().getSelectionLength();
		int selectionEnd = getContext().getSelectionOffset() + getContext().getSelectionLength();
		if (selectionEnd > offset + length) {
			// extend the selection if more than one statement is selected (bug
			// 72149)
			length = selectionEnd - offset;
		}

		ASTNode[] annotationNodes = getAnnotationNodes();
		TryStatement surroundingTry = ASTResolving.findParentTryStatement(annotationNodes[0]);
		// MethodDeclaration methodDecl = (MethodDeclaration)
		// ASTResolving.findParentBodyDeclaration(annotationNodes[0]);
		BodyDeclaration decl = ASTResolving.findParentBodyDeclaration(annotationNodes[0]);
		if (decl == null || !(decl instanceof MethodDeclaration)) {
			return "<N/A>";
		}
		MethodDeclaration methodDecl = (MethodDeclaration) decl;

		// ITypeBinding[] uncaughtExceptions =
		// org.eclipse.jdt.internal.corext.refactoring.surround.ExceptionAnalyzer
		// .perform(decl, Selection.createFromStartLength(offset, length));
		// if (uncaughtExceptions.length == 0) {
		// return "<N/A>";
		// }
		// for (ITypeBinding b : uncaughtExceptions) {
		// System.out.println("TB: " + b.getName());
		// }

		// ----------------------------- Change -----------------------------//

		// track the linked nodes and annotations for method name
		List<String> exceptions = exceptionModel.getAllExceptions();
		for (String ex : exceptions) {
			System.out.println("Exceptions to throw: " + ex);
		}

		// change (throws)
		fLogger.info("Throws: \n");
		if (surroundingTry != null) {
			for (String ex : exceptions) {
				SimpleName exceptionName = ast.newSimpleName(ex);
				SimpleType exceptionType = ast.newSimpleType(exceptionName);
				String methods = exceptionModel.getMethodsByException(exceptionName.getIdentifier());
				addExceptionToThrows(ast, methodDecl, rewrite, exceptionType, trackedExceptionNameAnnotationMap, methods);
			}
		}

		// leave the try-block and remove exception handlers
		Block parentBlock = (Block) surroundingTry.getParent();
		ListRewrite blockRewrite = rewrite.getListRewrite(parentBlock, Block.STATEMENTS_PROPERTY);

		Block tryBlock = surroundingTry.getBody();
		for (Object statement : tryBlock.statements()) {
			Statement copiedStatement = (Statement) ASTNode.copySubtree(ast, (Statement) statement);
			blockRewrite.insertBefore(copiedStatement, surroundingTry, null);
		}

		// Remove try statement
		rewrite.remove(surroundingTry, null);

		/********************************** Writing ****************************** */

		fRewrite = rewrite; // (important for quickfix)
		return rewriteMethodDecl(rewrite);
	}

	//
	private void addExceptionToThrows(final AST ast, final MethodDeclaration methodDecl, final ASTRewrite rewrite,
			final SimpleType type, String methods) {
		ITypeBinding binding = type.resolveBinding();
		if (binding == null || isNotYetThrown(binding, methodDecl.thrownExceptions())) {
			Name name = type.getName();
			Name newName = (Name) ASTNode.copySubtree(ast, name);
			ListRewrite listRewriter = rewrite.getListRewrite(methodDecl, MethodDeclaration.THROWN_EXCEPTIONS_PROPERTY);
			listRewriter.insertLast(newName, null);
		}
	}

	private List<String> getUniCatchExceptions(ExceptionDataModel exceptionModel) {
		List<String> exceptions = new ArrayList<String>();
		for (MethodsExceptionsPair pair : exceptionModel.getMethodsExceptionsToCatch()) {
			List<String> exceptionNames = pair.getExceptions();
			for (String ex : exceptionNames) {
				exceptions.add(ex);
			}
		}
		return exceptions;
	}

	private void addMethodNameInTryBlock(final AST ast, final ASTRewrite rewrite, final SimpleName name,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedMethodNameAnnotationMap, final int index,
			String exceptions) {
		ITrackedNodePosition exceptionPosition = rewrite.track(name);
		String methodName = name.getFullyQualifiedName();
		int i = ExceptionColorRegistry.getInstance().getIndex(methodName);
		String annotationType = findMethodAnnotationType(i, index);
		MethodCallAnnotation annotation = new MethodCallAnnotation(annotationType, null); // initial
		annotation.setText(Messages.format(Messages.methodThrowsExceptions, new String[] { methodName, exceptions }));
		trackedMethodNameAnnotationMap.put(exceptionPosition, annotation);
		ExceptionColorRegistry.getInstance().register(methodName, false, index % TOTAL_ANNOTATIONS);
	}

	private String findMethodAnnotationType(final int i, final int index) {
		// single method may throw the exception.
		if (i == -1) {
			return fMethodAnnotations.get(index % TOTAL_ANNOTATIONS);
		}
		return fMethodAnnotations.get(i % TOTAL_ANNOTATIONS);
	}

	public TryStatement findTryStatement() {
		FixBugsAnnotation annotation = findFirstAnnotationInTryStatement();
		ASTNode node = findASTNode(annotation.getPosition());
		return ASTResolving.findParentTryStatement(node);
	}

	private void addExceptionToCatch(final AST ast, final ASTRewrite rewrite, final SimpleType type,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedExceptionNameAnnotationMap, final int index,
			final boolean inCatch, final boolean inMultiCatch, final String methods) {
		ITrackedNodePosition exceptionPosition = rewrite.track(type.getName());

		String exceptionName = type.getName().getFullyQualifiedName();
		int i = ExceptionColorRegistry.getInstance().getIndex(exceptionName);
		String annotationType = findExceptionAnnotationType(i, index, exceptionName);
		ExceptionAnnotation annotation = new ExceptionAnnotation(annotationType, null); // initial
		annotation.setText(Messages.format(Messages.exceptionMayThrownByMethod, new String[] { exceptionName, methods }));
		annotation.setName(exceptionName); // don't miss
		annotation.setCaught(inCatch);
		annotation.setMultiCaught(inMultiCatch);
		trackedExceptionNameAnnotationMap.put(exceptionPosition, annotation);
		ExceptionColorRegistry.getInstance().register(exceptionName, hasMultiCallers(exceptionName), index % TOTAL_ANNOTATIONS);
	}

	private String findExceptionAnnotationType(final int i, final int index, final String exceptionName) {
		if (hasMultiCallers(exceptionName)) {
			return fMultiCallerAnnotation;
		}
		// single method may throw the exception.
		if (i == -1) {
			return fExceptionAnnotations.get(index % TOTAL_ANNOTATIONS);
		}
		return fExceptionAnnotations.get(i % TOTAL_ANNOTATIONS);
	}

	private void addExceptionToThrows(final AST ast, final MethodDeclaration methodDeclaration, final ASTRewrite rewrite,
			final SimpleType type, final Map<ITrackedNodePosition, FixBugsAnnotation> trackedExceptionNameAnnotationMap,
			final String methods) {
		ITypeBinding binding = type.resolveBinding();
		if (binding == null || isNotYetThrown(binding, methodDeclaration.thrownExceptions())) {
			Name name = type.getName();
			Name newName = (Name) ASTNode.copySubtree(ast, name);

			ListRewrite listRewriter = rewrite.getListRewrite(methodDeclaration, MethodDeclaration.THROWN_EXCEPTIONS_PROPERTY);
			listRewriter.insertLast(newName, null);

			// for tracking node
			ITrackedNodePosition exceptionPosition = rewrite.track(newName);
			String exceptionName = type.getName().getFullyQualifiedName();
			int i = ExceptionColorRegistry.getInstance().getIndex(exceptionName);
			String annotationTtype = findExceptionAnnotationType(i, 1, exceptionName);
			ExceptionAnnotation annotation = new ExceptionAnnotation(annotationTtype, null);
			annotation.setText(Messages.format(Messages.exceptionMayThrownByMethod, new String[] { exceptionName, methods }));
			annotation.setName(newName.getFullyQualifiedName());
			annotation.setCaught(false);
			annotation.setMultiCaught(false);
			trackedExceptionNameAnnotationMap.put(exceptionPosition, annotation);
		}
	}

	// CatchClause handling "Exception"
	private static void deleteCatchClause(final ASTRewrite rewrite, final TryStatement tryStatement) {
		List<ASTNode> catches = tryStatement.catchClauses();
		for (ASTNode node : catches) {
			if (node instanceof CatchClause) {
				CatchClause cc = (CatchClause) node;
				SingleVariableDeclaration svd = cc.getException();
				if (svd.getType() instanceof SimpleType) {
					SimpleType st = (SimpleType) svd.getType();
					fLogger.fine("#deleteCatchClause: " + st.getName().getFullyQualifiedName());
					if ("Exception".equals(st.getName().getFullyQualifiedName())) {
						rewrite.remove(cc, null);
					}
				}
			}
		}
	}

	private static IMethodBinding findMethodBinding(ASTNode node) {
		MethodInvocation mi = findMethodInvocation(node);
		ClassInstanceCreation cic = findClassInstanceCreation(node);
		IMethodBinding mb = null;
		if (mi != null) {
			mb = mi.resolveMethodBinding();
		} else if (cic != null) {
			mb = cic.resolveConstructorBinding();
		}
		return mb;
	}

	private static MethodInvocation findMethodInvocation(ASTNode node) {
		while ((node != null) && (!(node instanceof MethodInvocation))) {
			node = node.getParent();
			if (node instanceof BodyDeclaration) {
				return null;
			}
		}
		return (MethodInvocation) node;
	}

	private static ClassInstanceCreation findClassInstanceCreation(ASTNode node) {
		while ((node != null) && (!(node instanceof ClassInstanceCreation))) {
			node = node.getParent();
			if (node instanceof BodyDeclaration) {
				return null;
			}
		}
		return (ClassInstanceCreation) node;
	}

	private static ITypeBinding findTypeBinding(final ASTNode[] nodes, final String exceptionName) {
		fLogger.info("CatchExceptionAssistant#findTypeBinding");
		for (ASTNode node : nodes) {
			MethodInvocation mi = findMethodInvocation(node);
			if (mi != null) {
				fLogger.info("\tMI: " + mi.getName());
				IMethodBinding mb = mi.resolveMethodBinding();
				ITypeBinding[] tbs = mb.getExceptionTypes();
				for (ITypeBinding tb : tbs) {
					if (tb.getName().contains(exceptionName)) {
						fLogger.info("\t\tEx: " + tb.getName());
						return tb;
					}
				}
			}
			ClassInstanceCreation cic = findClassInstanceCreation(node);
			if (cic != null) {
				// fLogger.info("\tCIC: " + cic..getName());
				IMethodBinding mb = cic.resolveConstructorBinding();
				ITypeBinding[] tbs = mb.getExceptionTypes();
				for (ITypeBinding tb : tbs) {
					if (tb.getName().contains(exceptionName)) {
						fLogger.info("\t\tEx: " + tb.getName());
						return tb;
					}
				}
			}
		}
		return null;
	}

	// Find already thrown exceptions
	private static List<String> findAlreadyThrownExceptions(final ASTNode node) {
		List<String> exceptions = new ArrayList<String>();
		BodyDeclaration bodyDeclaration = ASTResolving.findParentBodyDeclaration(node);
		if (bodyDeclaration instanceof MethodDeclaration) {
			MethodDeclaration md = (MethodDeclaration) bodyDeclaration;
			for (Object name : md.thrownExceptions()) {
				// Remember that it stores FQN, not simple name!!
				exceptions.add(((Name) name).getFullyQualifiedName());
			}
		}
		return exceptions;
	}

	private static boolean isNotYetThrown(final ITypeBinding binding, final List<Name> thrownExceptions) {
		for (int i = 0; i < thrownExceptions.size(); i++) {
			Name name = thrownExceptions.get(i);
			ITypeBinding elem = (ITypeBinding) name.resolveBinding();
			if (elem != null) {
				// existing exception is base class of new
				if (Bindings.isSuperType(elem, binding)) {
					return false;
				}
			}
		}
		return true;
	}

	// private static boolean isAlreadyEnclosed(final List<String>
	// thrownExceptions, final String exceptionName) {
	// for (String exceptionFQN : thrownExceptions) {
	// // System.out.println(exceptionFQN + " vs " + exceptionName);
	// if (exceptionFQN.contains(exceptionName) ||
	// exceptionName.contains(exceptionFQN)) {
	// return true;
	// }
	// }
	// return false;
	// }

	// private static boolean isDeclarationImported(ASTNode node) {
	// while ((node != null) && (!(node instanceof ImportDeclaration))) {
	// node = node.getParent();
	// if (node instanceof ImportDeclaration) {
	// return true;
	// }
	// }
	// return false;
	// }
	// private static boolean isSubtype(ITypeBinding curr, ITypeBinding[]
	// addedExceptions) {
	// while (curr != null) {
	// for (int i = 0; i < addedExceptions.length; i++) {
	// if (curr == addedExceptions[i]) {
	// return true;
	// }
	// }
	// curr = curr.getSuperclass();
	// }
	// return false;
	// }
}
