//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.ParameterizedType;
import org.eclipse.jdt.core.dom.SimpleType;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.internal.corext.dom.ASTNodes;

/**
 */
public final class ExceptionCollector extends ASTVisitor {

	/** */
	private final List<ASTNode> fMethods;

	/** */
	public ExceptionCollector() {
		fMethods = new ArrayList<ASTNode>();
	}

	/**
	 * Returns the ASTNode array of the methods that throw exceptions.
	 * 
	 * @param statements
	 *            the statements
	 * @return the methods that throw exceptions.
	 */
	public static ASTNode[] perform(final List<Statement> statements) {
		ExceptionCollector collector = new ExceptionCollector();
		for (Statement node : statements) {
			node.accept(collector);
		}
		return collector.getMethods();
	}

	@Override
	public boolean visit(final MethodInvocation node) {
		IMethodBinding methodBinding = node.resolveMethodBinding();
		ITypeBinding[] tbs = methodBinding.getExceptionTypes();
		if (tbs.length > 0) {
			// fMethodExceptionsMap.put(node.getName(), tbs);
			// fMap.put(node.getName(), tbs);
			fMethods.add(node.getName());
		}
		return true;
	}

	@Override
	public boolean visit(final ClassInstanceCreation node) {
		IMethodBinding methodBinding = node.resolveConstructorBinding();
		ITypeBinding[] tbs = methodBinding.getExceptionTypes();
		if (tbs.length > 0) {
			// fMethodExceptionsMap.put(node.getName(), tbs);
			// fMap.put(node.getName(), tbs);
			Type type = node.getType();
			if (type instanceof ParameterizedType) {

			}
			if (type instanceof SimpleType) {
				Name name = ((SimpleType) type).getName();
				fMethods.add(ASTNodes.getLeftMostSimpleName(name));
			}
		}
		return true;
	}

	private ASTNode[] getMethods() {
		return fMethods.toArray(new ASTNode[0]);
	}
}
