//
// Copyright (c) 2007-2012 Yoonki Song.
// All rights reserved. This program and the accompanying materials
// are made available under the permission of me.
//
// Contributors:
// Yoonki Song - initial API and implementation
//
package edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException;

import java.util.List;

import edu.ncsu.csc.emh.fixbugs.core.util.PrintUtil;

/**
 * 
 * <pre>
 * [e1], [m1]
 * </pre>
 * 
 * <pre>
 * [e1, e2], [m1]
 * </pre>
 * 
 * <pre>
 * [e1], [m1, m2]
 * </pre>
 * 
 * @author Yoonki Song
 */
public final class ExceptionsMethodsPair {

	/** K: list of exception name. */
	private final List<String> fExceptions;

	/** V: list of methods may throw the exception. */
	private List<String> fMethods;

	/**
	 * Constructor.
	 * 
	 * @param exceptions
	 *            exception list may thrown by method list.
	 * @param methods
	 *            method list may throw exceptions.
	 */
	public ExceptionsMethodsPair(final List<String> exceptions, final List<String> methods) {
		fExceptions = exceptions;
		fMethods = methods;
	}

	/**
	 * Returns the method list.
	 * 
	 * @return the method list.
	 */
	public List<String> getMethods() {
		return fMethods;
	}

	/**
	 * Set the method list.
	 * 
	 * @param methods
	 *            method list.
	 */
	public void setMethodList(final List<String> methods) {
		fMethods = methods;
	}

	/**
	 * Returns the sting representation of the exception list.
	 * 
	 * @return the sting representation of the exception list.
	 */
	public String getMethodsAsString() {
		return PrintUtil.toListString(fMethods);
	}

	/**
	 * Returns the exception list.
	 * 
	 * @return the exception list.
	 */
	public List<String> getExceptions() {
		return fExceptions;
	}

	/**
	 * Checks if the method list contains the given method.
	 * 
	 * @param method
	 *            the method.
	 * @return true if the method list contains the given method; false
	 *         otherwise.
	 */
	public boolean containsMethod(final String method) {
		for (String me : fMethods) {
			if (me.equals(method)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Checks if the exception list contains the given exception.
	 * 
	 * @param exception
	 *            the exception.
	 * @return true if the exception list contains the given exception; false
	 *         otherwise.
	 */
	public boolean containsException(final String exception) {
		for (String ex : fExceptions) {
			if (ex.equals(exception)) {
				return true;
			}
		}
		return false;

	}

	@Override
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append('[');
		buf.append(PrintUtil.toListString(fExceptions));
		buf.append(", ");
		buf.append(PrintUtil.toListString(fMethods));
		buf.append(']');
		return buf.toString();
	}
}
