//
// Copyright (c) 2007-2012 Yoonki Song.
// All rights reserved. This program and the accompanying materials
// are made available under the permission of me.
//
// Contributors:
// Yoonki Song - initial API and implementation
//
package edu.ncsu.csc.emh.fixbugs.core.util;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jface.text.Position;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;

/**
 * @author Yoonki Song
 */
public final class PositionTracker {

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(PositionTracker.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/**
	 * No instance.
	 */
	private PositionTracker() {
	}

	// update the positions of annotations
	public static List<FixBugsAnnotation> perform(final Map<ITrackedNodePosition, FixBugsAnnotation> trackedNodeMap) {
		List<FixBugsAnnotation> annotations = new ArrayList<FixBugsAnnotation>();
		for (Entry<ITrackedNodePosition, FixBugsAnnotation> entry : trackedNodeMap.entrySet()) {
			Position p = perform(entry.getKey());
			FixBugsAnnotation annotation = entry.getValue();
			annotation.setPosition(p);
			fLogger.info("[Annotation] " + annotation);
			annotations.add(annotation);
		}
		return annotations;
	}

	public static List<Position> perform(final List<ITrackedNodePosition> links) {
		List<Position> positions = new ArrayList<Position>();
		for (ITrackedNodePosition tracked : links) {
			positions.add(perform(tracked));
		}
		return positions;
	}

	public static FixBugsAnnotation perform(final Pair<ITrackedNodePosition, FixBugsAnnotation> trackedNode) {
		Position p = perform(trackedNode.getLeft());
		FixBugsAnnotation annotation = trackedNode.getRight();
		annotation.setPosition(p);
		return annotation;
	}

	public static Position perform(final ITrackedNodePosition p) {
		return new Position(p.getStartPosition(), p.getLength());
	}
}
