//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.util;

import java.util.List;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.internal.corext.dom.ASTNodes;
import org.eclipse.jface.text.link.LinkedPosition;
import org.eclipse.jface.text.link.LinkedPositionGroup;
import org.eclipse.swt.widgets.Shell;

/**
 * @author Yoonki
 * 
 */
public final class PrintUtil {

	/**
	 * no instance.
	 */
	private PrintUtil() {
	}

	/**
	 * 
	 * @param list
	 *            input
	 * @return the string representation.
	 */
	public static String toListString(final List<String> list) {
		StringBuilder builder = new StringBuilder();
		int methodSize = list.size();
		for (int i = 0; i < methodSize; i++) {
			builder.append(list.get(i));
			if (i < methodSize - 1) {
				builder.append(", ");
			}
		}
		return builder.toString();
	}

	public static String toString(final List<LinkedPositionGroup> linkedPositionGroupList) {
		StringBuilder builder = new StringBuilder();
		for (LinkedPositionGroup lpGroup : linkedPositionGroupList) {
			builder.append(toString(lpGroup)).append('\n');
		}
		return builder.toString();
	}

	public static String toString(final LinkedPositionGroup linkedPositionGroup) {
		StringBuilder builder = new StringBuilder();
		for (LinkedPosition lp : linkedPositionGroup.getPositions()) {
			builder.append('[').append(lp.getOffset()).append(',').append(lp.getLength()).append(']');
		}
		return builder.toString();
	}

	public static String toString(final String message, final ASTNode node) {
		StringBuilder builder = new StringBuilder();
		if (node != null) {
			builder.append(message + " at [" + node.getStartPosition() + ", " + node.getLength() + "] ");
			builder.append(ASTNodes.asString(node));
		}
		return builder.toString();
	}

	public static String toString(final String message, final SimpleName node) {
		StringBuilder builder = new StringBuilder();
		builder.append(message + " at [" + node.getStartPosition() + ", " + node.getLength() + "] ");
		builder.append('\"' + node.getIdentifier() + '\"');
		return builder.toString();
	}

	public static String toString(final String message, final ASTNode node, final ITrackedNodePosition position) {
		StringBuilder builder = new StringBuilder();
		builder.append(message + " from [" + node.getStartPosition() + ", " + node.getLength() + ']');
		builder.append(" to [" + position.getStartPosition() + ", " + position.getLength() + ']');
		return builder.toString();
	}

	public static String toString(final String message, final ASTNode node, final String text) {
		StringBuilder builder = new StringBuilder();
		builder.append(message + " from [" + node.getStartPosition() + ", " + node.getLength() + ']');
		builder.append(" to [" + node.getStartPosition() + ", " + text.length() + ']');
		return builder.toString();
	}

	public static String toString(final String message, final ITrackedNodePosition position, final String type) {
		StringBuilder builder = new StringBuilder();
		builder.append(message + " [" + position.getStartPosition() + ", " + position.getLength() + "] ");
		builder.append(type);
		return builder.toString();
	}

	public static String toString(final ASTNode coveringBD, final ITrackedNodePosition position, final String newText) {
		StringBuilder builder = new StringBuilder();
		builder.append("[Current]" + " [" + position.getStartPosition() + ", " + position.getLength() + "]\n");
		builder.append(ASTNodes.asString(coveringBD)).append('\n');
		builder.append("[New] ").append(newText.length()).append(('\n'));
		builder.append(newText);
		return builder.toString();
	}

	public static String toString(final ASTNode coveringBD, final String newText) {
		StringBuilder builder = new StringBuilder();
		builder.append("[Current] " + " [" + coveringBD.getStartPosition() + ", " + coveringBD.getLength() + "]\n");
		builder.append(ASTNodes.asString(coveringBD)).append('\n');
		builder.append("[New] ").append(newText.length()).append(('\n'));
		builder.append(newText);
		return builder.toString();
	}

	public static String toString(final Shell shell) {
		StringBuffer buf = new StringBuffer();
		buf.append("Shell Location: " + shell.getLocation() + ", Size: " + shell.getSize() + ", Shell Bounds: "
				+ shell.getBounds());
		return buf.toString();
	}

}
