/**
 * 
 */
package edu.ncsu.csc.emh.fixbugs.core.util;

import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @author Yoonki
 *
 */
public final class SignatureUtil {

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(SignatureUtil.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}
	
	/**
	 * No instance.
	 */
	private SignatureUtil() {
	}
	
	//
	public static final String PRIMITIVE_BOOLEAN = "boolean";
	public static final String PRIMITIVE_CHAR = "char";
	public static final String PRIMITIVE_BYTE = "byte";
	public static final String PRIMITIVE_SHORT = "short";
	public static final String PRIMITIVE_INT = "int";
	public static final String PRIMITIVE_LONG = "long";
	public static final String PRIMITIVE_DOUBLE = "double";
	public static final String PRIMITIVE_FLOAT = "float";
	
	//
	public static final String WRAPPER_BOOLEAN = "Boolean";
	public static final String WRAPPER_CHARACTER = "Character";
	public static final String WRAPPER_BYTE = "Byte";
	public static final String WRAPPER_SHORT = "Short";
	public static final String WRAPPER_INTEGER = "Integer";
	public static final String WRAPPER_STRING = "String";
	public static final String WRAPPER_LONG = "Long";
	public static final String WRAPPER_DOUBLE = "Double";
	public static final String WRAPPER_FLOAT = "Float";

	/**
	 * Finds the array of possible types.
	 * 
	 * @param formatString
	 * @return
	 * @see <a
	 *      href="http://docs.oracle.com/javase/7/docs/api/java/util/Formatter.html">Formatter</a>
	 */	
	public static String[] findPossibleTypes(final String formatString) {
		String lastCharacter = formatString.substring(formatString.length() - 1);
		if ("b".equalsIgnoreCase(lastCharacter)) { //$NON-NLS-1
			return new String[] { PRIMITIVE_BOOLEAN, WRAPPER_BOOLEAN };
		} else if ("h".equalsIgnoreCase(lastCharacter)) { //$NON-NLS-1
			return new String[] { PRIMITIVE_INT, WRAPPER_INTEGER };
		} else if ("s".equalsIgnoreCase(lastCharacter)) {//$NON-NLS-1
			return new String[] { WRAPPER_STRING };
		} else if ("c".equalsIgnoreCase(lastCharacter)) {//$NON-NLS-1
			return new String[] { PRIMITIVE_CHAR, WRAPPER_CHARACTER };
		} else if ("d".equalsIgnoreCase(lastCharacter) || "o".equalsIgnoreCase(lastCharacter)
				|| "x".equalsIgnoreCase(lastCharacter)) {//$NON-NLS-1
			return new String[] { PRIMITIVE_BYTE, PRIMITIVE_SHORT, PRIMITIVE_INT, PRIMITIVE_LONG,
					WRAPPER_BYTE, WRAPPER_SHORT, WRAPPER_INTEGER, WRAPPER_LONG };
		} else if ("f".equalsIgnoreCase(lastCharacter) || "e".equalsIgnoreCase(lastCharacter)
				|| "g".equalsIgnoreCase(lastCharacter) || "a".equalsIgnoreCase(lastCharacter)) {//$NON-NLS-1
			return new String[] { PRIMITIVE_DOUBLE, WRAPPER_DOUBLE, PRIMITIVE_FLOAT, WRAPPER_FLOAT };
		} else if ("n".equalsIgnoreCase(lastCharacter)) {//$NON-NLS-1
			// Ignored!
		}
		return new String[] { "" };
	}
	
}
