//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.util;

import java.util.AbstractList;
import java.util.Collection;

import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.NamingConventions;

/**
 * 
 * @see Stub-Utility (remove '-')
 * @author Yoonki Song
 */
public final class SuggestionUtil {

	private SuggestionUtil() {
		// no instantiation.
	}

	/**
	 * Returns suggestions for a variable name.
	 * 
	 * @param variableKind
	 *            the variable kind.
	 * @param project
	 *            the project.
	 * @param baseName
	 *            the base name.
	 * @param dimensions
	 *            the dimensions.
	 * @param excluded
	 *            if it is in the excluded list.
	 * @param evaluateDefault
	 *            whether ...
	 * @return variable name suggestions.
	 */
	public static String[] getVariableNameSuggestions(final int variableKind, final IJavaProject project,
			final String baseName, final int dimensions, final Collection<String> excluded, final boolean evaluateDefault) {
		return NamingConventions.suggestVariableNames(variableKind, NamingConventions.BK_TYPE_NAME,
				removeTypeArguments(baseName), project, dimensions, getExcludedArray(excluded), evaluateDefault);
	}

	private static String removeTypeArguments(final String baseName) {
		int idx = baseName.indexOf('<');
		if (idx != -1) {
			return baseName.substring(0, idx);
		}
		return baseName;
	}

	private static String[] getExcludedArray(final Collection<String> excluded) {
		if (excluded == null) {
			return null;
		} else if (excluded instanceof ExcludedCollection) {
			return ((ExcludedCollection) excluded).getExcludedArray();
		}
		return excluded.toArray(new String[excluded.size()]);
	}

	private static class ExcludedCollection extends AbstractList<String> {
		private final String[] fExcluded;

		public ExcludedCollection(final String[] excluded) {
			fExcluded = excluded;
		}

		public String[] getExcludedArray() {
			return fExcluded;
		}

		@Override
		public int size() {
			return fExcluded.length;
		}

		@Override
		public String get(final int index) {
			return fExcluded[index];
		}

		@Override
		public int indexOf(final Object o) {
			if (o instanceof String) {
				for (int i = 0; i < fExcluded.length; i++) {
					if (o.equals(fExcluded[i])) {
						return i;
					}
				}
			}
			return -1;
		}

		@Override
		public boolean contains(final Object o) {
			return indexOf(o) != -1;
		}
	}
}
