//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.actions;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.internal.ui.javaeditor.EditorUtility;
import org.eclipse.jdt.internal.ui.javaeditor.JavaEditor;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.swt.widgets.Shell;

/**
 * @see org.eclipse.jdt.internal.ui.refactoring.actions.RenameJavaElementAction
 * @author Yoonki
 */
@SuppressWarnings("restriction")
public abstract class FixBugsAction extends Action {

	/** */
	private final IMarker fMarker;

	/** */
	private final JavaEditor fEditor;

	/**
	 * Constructor for FixBugsAction.
	 * 
	 * @param marker
	 *            the marker.
	 */
	public FixBugsAction(final IMarker marker) {
		fMarker = marker;
		fEditor = (JavaEditor) EditorUtility.isOpenInEditor(getCompilationUnit(marker));
	}

	protected final JavaEditor getEditor() {
		return fEditor;
	}

	protected final Shell getShell() {
		return fEditor.getEditorSite().getShell();
	}

	protected final ICompilationUnit getCompilationUnit(final IMarker marker) {
		IResource res = marker.getResource();
		if (res instanceof IFile && res.isAccessible()) {
			IJavaElement element = JavaCore.create((IFile) res);
			if (element instanceof ICompilationUnit) {
				return (ICompilationUnit) element;
			}
		}
		return null;
	}

	protected final ITextSelection getTextSelection() {
		ISelectionProvider selectionProvider = fEditor.getSelectionProvider();
		if (selectionProvider == null) {
			return null;
		}

		ISelection selection = selectionProvider.getSelection();
		if (!(selection instanceof ITextSelection)) {
			return null;
		}

		return (ITextSelection) selection;
	}
}
