//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.actions;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.resources.IMarker;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.ITextSelection;

import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.resolution.correctness.missingArgument.MissingArgumentAssistant;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.correctness.missingArgument.MissingArgumentMode;

/**
 * 
 * @author Yoonki
 */
public final class MissingArgumentAction extends FixBugsAction {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(MissingArgumentAction.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	public MissingArgumentAction(final IMarker marker) {
		super(marker);
	}

	@Override
	public void run() {
		fLogger.info("MissingArgumentAction#run()");
		if (!ActionUtil.isEditable(getEditor())) {
			return;
		}

		ICompilationUnit cu = JavaUI.getWorkingCopyManager().getWorkingCopy(getEditor().getEditorInput());
		if (cu == null) {
			return;
		}

		ITextSelection textSelection = getTextSelection();
		FixBugsContext context = new FixBugsContext(cu, textSelection.getOffset(), textSelection.getLength());
		MissingArgumentAssistant assistant = new MissingArgumentAssistant(context);

		if (getEditor() instanceof CompilationUnitEditor) {
			new MissingArgumentMode((CompilationUnitEditor) getEditor(), assistant).start(true);
		}
	}
}
