//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.actions;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.resources.IMarker;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.internal.ui.actions.ActionUtil;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.ITextSelection;

import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.resolution.SearchHint;
import edu.ncsu.csc.emh.fixbugs.core.resolution.performance.stringBufferConcetenation.StringBufferConcatenationAssistant;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.performance.stringBufferConcatenation.StringBufferConcatenationMode;

/**
 * @author Yoonki
 */
public final class StringBufferConcatenationAction extends FixBugsAction {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(StringBufferConcatenationAction.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/**
	 * 
	 * @param marker
	 */
	public StringBufferConcatenationAction(final IMarker marker) {
		super(marker);
	}

	@Override
	public void run() {
		fLogger.info("StringBufferConcatenationAction#run()");
		if (!ActionUtil.isEditable(getEditor())) {
			return;
		}

		ICompilationUnit cu = JavaUI.getWorkingCopyManager().getWorkingCopy(getEditor().getEditorInput());
		if (cu == null) {
			return;
		}

		ITextSelection textSelection = getTextSelection();

		FixBugsContext context = new FixBugsContext(cu, textSelection.getOffset(), textSelection.getLength());
		SearchHint hint = new SearchHint();
		hint.add("String");
		StringBufferConcatenationAssistant assistant = new StringBufferConcatenationAssistant(context, hint);

		if (getEditor() instanceof CompilationUnitEditor) {
			new StringBufferConcatenationMode((CompilationUnitEditor) getEditor(), assistant).start(true);
		}
	}
}
