//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.commands.operations.IUndoableOperation;
import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.internal.ui.text.correction.proposals.LinkedNamesAssistProposal.DeleteBlockingExitPolicy;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.DocumentEvent;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IEditingSupport;
import org.eclipse.jface.text.IEditingSupportRegistry;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.link.ILinkedModeListener;
import org.eclipse.jface.text.link.LinkedModeModel;
import org.eclipse.jface.text.link.LinkedModeUI;
import org.eclipse.jface.text.link.LinkedModeUI.ExitFlags;
import org.eclipse.jface.text.link.LinkedPosition;
import org.eclipse.jface.text.link.LinkedPositionGroup;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.VerifyEvent;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.texteditor.link.EditorLinkedModeUI;

import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsLinkedAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsStatus;
import edu.ncsu.csc.emh.fixbugs.ui.handlers.TogglePopupHandler;
import edu.ncsu.csc.emh.fixbugs.ui.services.CommandState;

/**
 * 
 * XXX Need to revise refresh* methods!
 * 
 * @see <a href="http://en.wikipedia.org/wiki/Strategy_pattern">Strategy
 *      pattern</a>
 * 
 * 
 * 
 * @author Yoonki Song
 */
@SuppressWarnings("restriction")
public abstract class FixBugsLinkedMode extends FixBugsMode {
	private class EditorSynchronizer implements ILinkedModeListener {
		@Override
		public void left(final LinkedModeModel model, final int flags) {
			fLogger.info("FixBugsLinkedMode$EditorSynchronizer#linkedModeLeft");
			if (fPopup != null) {
				fLogger.info("\t do nothing!! let's keep holding the popup");
				return; // do nothing
			}
			stop();
			if ((flags & ILinkedModeListener.UPDATE_CARET) != 0) {
				// YS doRename(fShowPreview);
			}
		}

		@Override
		public void resume(final LinkedModeModel model, final int flags) {
			// do nothing
		}

		@Override
		public void suspend(final LinkedModeModel model) {
			// do nothing
		}
	}

	private class ExitPolicy extends DeleteBlockingExitPolicy {
		public ExitPolicy(final IDocument document) {
			super(document);
		}

		@Override
		public ExitFlags doExit(final LinkedModeModel model, final VerifyEvent event, final int offset, final int length) {
			return super.doExit(model, event, offset, length);
		}
	}

	private class FocusEditingSupport implements IEditingSupport {
		@Override
		public boolean isOriginator(final DocumentEvent event, final IRegion subjectRegion) {
			return false; // leave on external modification outside positions
		}

		@Override
		public boolean ownsFocusShell() {
			if (fPopup == null) {
				return false;
			}
			if (fPopup.ownsFocusShell()) {
				return true;
			}

			Shell editorShell = fEditor.getSite().getShell();
			Shell activeShell = editorShell.getDisplay().getActiveShell();
			if (editorShell == activeShell) {
				return true;
			}
			return false;
		}
	}

	private class EscKeyListener implements KeyListener {

		private final StyledText fTextWidget;

		public EscKeyListener() {
			final ISourceViewer viewer = fEditor.getViewer();
			fTextWidget = viewer.getTextWidget();
		}

		@Override
		public void keyPressed(final KeyEvent e) {
			if (e.keyCode == SWT.ESC) {
				fTextWidget.removeKeyListener(this);
				stop();
			}
		}

		@Override
		public void keyReleased(final KeyEvent e) {
			// do nothing
		}

		public void start() {
			fTextWidget.addKeyListener(this);
		}
	}

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(FixBugsLinkedMode.class.getName());

	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	protected static FixBugsLinkedMode fgActiveMode;

	public static FixBugsLinkedMode getActiveMode() {
		if (fgActiveMode != null) {
			ISourceViewer viewer = fgActiveMode.fEditor.getViewer();
			if (viewer != null) {
				StyledText textWidget = viewer.getTextWidget();
				if (textWidget != null && !textWidget.isDisposed()) {
					return fgActiveMode;
				}
			}
			// make sure we don't hold onto the active linked mode if anything
			// went wrong with canceling:
			fgActiveMode = null;
		}
		return null;
	}

	/** */
	private final FocusEditingSupport fFocusEditingSupport;

	/**
	 * The operation on top of the undo stack when the rename is
	 * {@link #start()}ed, or <code>null</code> if rename has not been started
	 * or the undo stack was empty.
	 * 
	 * @since 3.5
	 */
	private IUndoableOperation fStartingUndoOperation;

	/** */
	protected FixBugsLinkedPopup fPopup;

	/** */
	protected FixBugsLinkedAssistant fAssistant;

	/** Linked Mode Model. */
	private LinkedModeModel fLinkedModeModel;

	/** */
	private LinkedPosition fFirstLinkedPosition;

	/** */
	private final List<LinkedPositionGroup> fLinkedPositionGroupList = new ArrayList<LinkedPositionGroup>();

	public FixBugsLinkedPopup getPopup() {
		return fPopup;
	}

	/**
	 * Constructor for FixBugsLinkedMode.
	 * 
	 * @param editor
	 *            the editor.
	 * @param assistant
	 *            the invocation context.
	 */
	public FixBugsLinkedMode(final CompilationUnitEditor editor, final FixBugsLinkedAssistant assistant) {
		super(editor);
		Assert.isNotNull(editor, "Editor may not be null");
		fAssistant = assistant;
		fFocusEditingSupport = new FocusEditingSupport();
	}

	@Override
	public void applyOriginal() {
		fLogger.info("FixBugsLinkedMode#applyOriginal");
		long startTime = System.currentTimeMillis();
		try {
			// 1. read changes from the original context
			String originalText = fAssistant.getCoveringBodyDeclAsString();
			writeOriginal(originalText, fAssistant.getAdjustedSelection());

			// 2. update linked modes and annotations
			fAssistant.init();
			refresh(fAssistant);
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		}
		fLogger.warning("#applyOriginal: " + (System.currentTimeMillis() - startTime) + " ms");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void cancel() {
		fLogger.info("FixBugsLinkedMode#cancel");
		if (fLinkedModeModel != null) {
			fLinkedModeModel.exit(ILinkedModeListener.NONE);
		}
		stop();
	}

	public LinkedPosition getCurrentLinkedPosition() {
		Point selection = fEditor.getViewer().getSelectedRange();
		int start = selection.x;
		int end = start + selection.y;

		for (LinkedPositionGroup group : fLinkedPositionGroupList) {
			for (LinkedPosition position : group.getPositions()) {
				if (position.includes(start) && position.includes(end)) {
					return position;
				}
			}
		}

		return null;
	}

	/**
	 * Returns the first linked position.
	 * 
	 * @return the linked position.
	 */
	public final LinkedPosition getFirstLinkedPosition() {
		return fFirstLinkedPosition;
	}

	/**
	 * 
	 * @return
	 */
	protected LinkedModeModel getLinkedModeModel() {
		return fLinkedModeModel;
	}

	/**
	 * Returns the original context.
	 * 
	 * @return the original context.
	 */
	@Override
	public FixBugsLinkedAssistant getOriginalAssistant() {
		return fAssistant;
	}

	/**
	 * Checks if the cursor is in linked positions.
	 * 
	 * @return true if the cursor is in linked positions; false otherwise.
	 */
	public final boolean isCaretInLinkedPosition() {
		fLogger.info("FixBugsLinkedMode#isCaretInLinkedPosition");
		Point selection = fEditor.getViewer().getSelectedRange();
		fLogger.info("selection: " + selection);
		int start = selection.x;
		return fLinkedModeModel.anyPositionContains(start);
	}

	public final boolean isInLinkedPosition(final int x) {
		return fLinkedModeModel.anyPositionContains(x);
	}

	/**
	 * Refreshes annotations (and linked nodes) by given the context.
	 * 
	 * @param assistant
	 *            the assistant
	 * @return FixBugsStatus the FixBugs Status
	 */
	public FixBugsStatus refresh(final FixBugsLinkedAssistant assistant) {
		// System.out.println("FixBugsLinkedMode#refresh");
		FixBugsStatus result = new FixBugsStatus();

		// update linked mode (Model and UI)
		refreshLinkedNodesList(assistant.getLinkedNodesList());

		// update annotations
		refreshAnnotations(assistant.getAnnotations());

		// YS support animation startAnimation();

		return result;
	}

	public FixBugsStatus refresh(final FixBugsLinkedAssistant assistant, final List<List<Position>> linkedNodesList) {
		// System.out.println("FixBugsLinkedMode#refresh");
		FixBugsStatus result = new FixBugsStatus();

		// update linked mode (Model and UI)
		refreshLinkedNodesList2(linkedNodesList);

		// update annotations
		refreshAnnotations(assistant.getAnnotations());

		// YS support animation startAnimation();

		return result;
	}

	/**
	 * Sets linked positions.
	 * 
	 * @param nodes
	 * @return the linked position.
	 */
	private void refreshLinkedNodesList(final List<ASTNode[]> linkedNodesList) {
		Assert.isNotNull(linkedNodesList);
		fLogger.info("FixBugsLinkedMode#refreshLinkedNodesList: " + linkedNodesList.size());
		ISourceViewer viewer = fEditor.getViewer();
		IDocument document = viewer.getDocument();

		try {
			fLinkedModeModel = new LinkedModeModel();
			for (ASTNode[] nodes : linkedNodesList) {
				LinkedPositionGroup linkedPositionGroup = new LinkedPositionGroup();
				for (int i = 0; i < nodes.length; i++) {
					ASTNode node = nodes[i];
					LinkedPosition position = new LinkedPosition(document, node.getStartPosition(), node.getLength(), i);
					if (i == 0) {
						fFirstLinkedPosition = position;
					}
					linkedPositionGroup.addPosition(position);
					fLinkedPositionGroupList.add(linkedPositionGroup);
				}

				fLinkedModeModel.addGroup(linkedPositionGroup);
			}
			fLinkedModeModel.forceInstall();
			fLinkedModeModel.addLinkingListener(new EditorSynchronizer());

			LinkedModeUI ui = new EditorLinkedModeUI(fLinkedModeModel, viewer);
			ui.setExitPosition(viewer, fFirstLinkedPosition.getOffset(), 0, Integer.MAX_VALUE);
			ui.setExitPolicy(new ExitPolicy(document));
			ui.enter();

			if (viewer instanceof IEditingSupportRegistry) {
				IEditingSupportRegistry registry = (IEditingSupportRegistry) viewer;
				registry.register(fFocusEditingSupport);
			}
		} catch (BadLocationException e) {
			e.printStackTrace();
			// stop();
		}

	}

	/**
	 * Refreshes linked positions.
	 * 
	 * @param varLinks
	 *            the variable links..
	 */
	protected final void refreshLinkedNodesList2(final List<List<Position>> linkedNodesList) {
		ISourceViewer viewer = fEditor.getViewer();
		IDocument document = viewer.getDocument();

		try {
			fLinkedModeModel = new LinkedModeModel();
			int i = 0;
			for (List<Position> linkedNodes : linkedNodesList) {
				LinkedPositionGroup linkedPositionGroup1 = new LinkedPositionGroup();
				for (Position p : linkedNodes) {
					LinkedPosition position = new LinkedPosition(document, p.getOffset(), p.getLength(), i);
					if (i == 0) {
						fFirstLinkedPosition = position;
					}
					linkedPositionGroup1.addPosition(position);
				}
				fLinkedPositionGroupList.add(linkedPositionGroup1);
				fLinkedModeModel.addGroup(linkedPositionGroup1);
			}
			fLinkedModeModel.forceInstall();
			fLinkedModeModel.addLinkingListener(new EditorSynchronizer());

			LinkedModeUI ui = new EditorLinkedModeUI(fLinkedModeModel, viewer);
			ui.setExitPosition(viewer, fFirstLinkedPosition.getOffset(), 0, Integer.MAX_VALUE);
			ui.setExitPolicy(new ExitPolicy(document));
			ui.enter();

			if (viewer instanceof IEditingSupportRegistry) {
				IEditingSupportRegistry registry = (IEditingSupportRegistry) viewer;
				registry.register(fFocusEditingSupport);
			}
		} catch (BadLocationException e) {
			e.printStackTrace();
			// stop();
		}

	}

	@Override
	public void start(final boolean applyDefault) {
		fLogger.info("FixBugsLinkedMode#start");

		if (getActiveMode() != null) {
			// for safety; should already be handled in RenameJavaElementAction
			// YS fgActiveLinkedMode.startFullDialog();
			fLogger.warning("Linked Mode is still activated.");
			fgActiveMode = null; // It's tricky
			return;
		}

		CommandState.toggle();
		TogglePopupHandler.showOrHide(Boolean.TRUE);

		handleUndo();

		FixBugsStatus result = null;
		try {
			FixBugsLinkedAssistant assistant = getOriginalAssistant();
			assistant.init();
			result = refresh(assistant);
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		} finally {
			fgActiveMode = this;

			if (!result.hasFatalError()) {
				// System.out.println("Ok! Let's open the popup.");
				openPopup();
			} else {
				fLogger.warning(";-(");
			}
		}

		fEscKeyListener = new EscKeyListener();
		fEscKeyListener.start();
	}

	private EscKeyListener fEscKeyListener = null;

	@Override
	public void stop() {
		fLogger.info("FixBugsLinkedMode#stop");
		fgActiveMode = null;

		ISourceViewer viewer = fEditor.getViewer();
		if (viewer instanceof IEditingSupportRegistry) {
			IEditingSupportRegistry registry = (IEditingSupportRegistry) viewer;
			registry.unregister(fFocusEditingSupport);
		}

		if (fEscKeyListener != null) {
			fEscKeyListener = null;
		}

		closePopup();
		removeAnnotations();

		CommandState.toggle();
		TogglePopupHandler.showOrHide(Boolean.FALSE);
	}

	public void closePopup() {
		if (fPopup != null) {
			fPopup.close();
		}
	}

	public void hide() {
		getPopup().hide();
		removeAnnotations();
		List<List<Position>> lofl = new ArrayList<List<Position>>(); // empty
		refreshLinkedNodesList2(lofl);
	}

	public void show() {
		// openPopup();
		getPopup().show();
		refresh((FixBugsLinkedAssistant) getCurrentAssistant());
	}

}
