//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jdt.internal.ui.JavaPlugin;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.internal.ui.preferences.JavaBasePreferencePage;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IMenuListener2;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.bindings.keys.IKeyLookup;
import org.eclipse.jface.bindings.keys.KeyLookupFactory;
import org.eclipse.jface.bindings.keys.KeyStroke;
import org.eclipse.jface.dialogs.IDialogSettings;
import org.eclipse.jface.text.ITextListener;
import org.eclipse.jface.text.ITextViewerExtension5;
import org.eclipse.jface.text.IViewportListener;
import org.eclipse.jface.text.IWidgetTokenKeeper;
import org.eclipse.jface.text.IWidgetTokenKeeperExtension;
import org.eclipse.jface.text.IWidgetTokenOwner;
import org.eclipse.jface.text.IWidgetTokenOwnerExtension;
import org.eclipse.jface.text.TextEvent;
import org.eclipse.jface.text.link.LinkedPosition;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.util.Geometry;
import org.eclipse.jface.util.Util;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.graphics.Region;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.ToolBar;
import org.eclipse.swt.widgets.ToolItem;
import org.eclipse.swt.widgets.Tracker;
import org.eclipse.ui.IPartListener2;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchPartReference;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.dialogs.PreferencesUtil;
import org.eclipse.ui.keys.IBindingService;
import org.eclipse.ui.progress.UIJob;

import edu.ncsu.csc.emh.fixbugs.ui.Images;
import edu.ncsu.csc.emh.fixbugs.ui.Messages;

/**
 * 
 * @author Yoonki Song
 */
@SuppressWarnings("restriction")
public abstract class FixBugsLinkedPopup implements IFixBugsPopup, IWidgetTokenKeeper, IWidgetTokenKeeperExtension {

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(FixBugsLinkedPopup.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	private class VisibilityManager implements ITextListener, IPartListener2, ControlListener, MouseListener, KeyListener,
			IViewportListener {

		public void start() {
			fLogger.info("VisibilityManager#start");
			fEditor.getSite().getWorkbenchWindow().getPartService().addPartListener(this);
			final ISourceViewer viewer = fEditor.getViewer();
			final StyledText textWidget = viewer.getTextWidget();
			textWidget.addControlListener(this);
			textWidget.addMouseListener(this);
			textWidget.addKeyListener(this);
			fEditor.getSite().getShell().addControlListener(this);
			viewer.addTextListener(this);
			viewer.addViewportListener(this);
			fShell.addDisposeListener(new DisposeListener() {
				@Override
				public void widgetDisposed(final DisposeEvent e) {
					fLogger.info("FixBugsLinkedPopup$VisibilityManager#widgetDisposed");
					fEditor.getSite().getWorkbenchWindow().getPartService().removePartListener(VisibilityManager.this);
					if (!textWidget.isDisposed()) {
						textWidget.removeControlListener(VisibilityManager.this);
						textWidget.removeMouseListener(VisibilityManager.this);
						textWidget.removeKeyListener(VisibilityManager.this);
					}
					fEditor.getSite().getShell().removeControlListener(VisibilityManager.this);
					viewer.removeTextListener(VisibilityManager.this);
					viewer.removeViewportListener(VisibilityManager.this);
					if (fMenuImage != null) {
						fMenuImage.dispose();
						fMenuImage = null;
					}
					if (fMenuManager != null) {
						fMenuManager.dispose();
						fMenuManager = null;
					}
					fLinkedMode.cancel();
				}
			});
		}

		@Override
		public void partActivated(final IWorkbenchPartReference partRef) {
			fLogger.info("FixBugsLinkedPopup$VisibilityManager#partActivated");
			IWorkbenchPart fPart = fEditor.getEditorSite().getPart();
			if (partRef.getPart(false) == fPart) {
				updateVisibility();
			}
		}

		@Override
		public void partBroughtToTop(final IWorkbenchPartReference partRef) {
			// nothing to do
		}

		@Override
		public void partClosed(final IWorkbenchPartReference partRef) {
			// nothing to do
		}

		@Override
		public void partDeactivated(final IWorkbenchPartReference partRef) {
			fLogger.info("FixBugsLinkedPopup$VisibilityManager#partDeactivated");
			IWorkbenchPart fPart = fEditor.getEditorSite().getPart();
			if (fShell != null && !fShell.isDisposed() && partRef.getPart(false) == fPart) {
				fShell.setVisible(false);
			}
		}

		@Override
		public void partHidden(final IWorkbenchPartReference partRef) {
			// nothing to do
		}

		@Override
		public void partInputChanged(final IWorkbenchPartReference partRef) {
			// nothing to do
		}

		@Override
		public void partOpened(final IWorkbenchPartReference partRef) {
			// nothing to do
		}

		@Override
		public void partVisible(final IWorkbenchPartReference partRef) {
			// nothing to do
		}

		@Override
		public void controlMoved(final ControlEvent e) {
			fLogger.info("FixBugsLinkedPopup$VisibilityManager#controlMoved");
			updateLocation(true);
			// updateVisibility(); // only for hiding outside editor area
		}

		@Override
		public void controlResized(final ControlEvent e) {
			updateLocation(true);
			// updateVisibility(); // only for hiding outside editor area
		}

		@Override
		public void mouseDoubleClick(final MouseEvent e) {
			// nothing to do
		}

		@Override
		public void mouseDown(final MouseEvent e) {
			// nothing to do
		}

		@Override
		public void mouseUp(final MouseEvent e) {
			fLogger.info("FixBugsLinkedPopup$VisibilityManager#mouseUp");
			updateLocation(false);
			updateVisibility();
		}

		@Override
		public void keyPressed(final KeyEvent e) {
			fLogger.info("FixBugsLinkedPopup$VisibilityManager#keyPressed");
			updateLocation(false);
			updateVisibility();
		}

		@Override
		public void keyReleased(final KeyEvent e) {
			// nothing to do
		}

		@Override
		public void textChanged(final TextEvent event) {
			fLogger.info("FixBugsLinkedPopup$VisibilityManager#textChanged");
			if (!event.getViewerRedrawState()) {
				return;
			}
			linkedNodeSelectionChanged(event);
			updateLocation(false); // YS
			updateVisibility(); // only for hiding outside editor area
		}

		@Override
		public void viewportChanged(final int verticalOffset) {
			fLogger.info("FixBugsLinkedPopup$VisibilityManager#viewportChanged: " + verticalOffset);
			updateLocation(true);
			updateVisibility(); // only for hiding outside editor area
		}
	}

	public void linkedNodeSelectionChanged(TextEvent event) {
		// System.out.println("FixBugsPopup#linkedNodeSelectionChanged");
	}

	/**
	 * Cached platform flag for dealing with platform-specific issue:
	 * https://bugs.eclipse.org/bugs/show_bug.cgi?id=219326 : Shell with custom
	 * region and SWT.NO_TRIM still has border.
	 */
	private static boolean MAC = Util.isMac();

	/** Y.S Changed "RenameInformationPopup" to "FixBugsPopup". */
	private static final String DIALOG_SETTINGS_SECTION = "FixBugsLinkedPopup"; //$NON-NLS-1$

	/** */
	private static final String SNAP_POSITION_KEY = "linked_popup_snap_position"; //$NON-NLS-1$

	/** */
	private final CompilationUnitEditor fEditor;

	/** */
	protected final FixBugsLinkedMode fLinkedMode;

	/** */
	private int fSnapPosition;

	/** */
	private boolean fSnapPositionChanged;

	/** */
	protected Shell fShell;

	/** */
	private final boolean fNeedToHold = true;

	/** */
	private GridLayout fPopupLayout;

	/** */
	private Region fRegion;

	/** */
	private Image fMenuImage;

	/** */
	private MenuManager fMenuManager;

	/** */
	private ToolBar fToolBar;

	/** */
	private String fOpenDialogBinding = ""; //$NON-NLS-1$

	/** */
	private boolean fIsMenuUp = false;

	/** */
	private boolean fDelayJobFinished = false;

	/** */
	private Button fOriginalButton;

	/** */
	private final boolean fIsOriginalButtonSelected = false;

	/**
	 * Creates the popup.
	 * 
	 * @param editor
	 *            the Editor
	 * @param linkedMode
	 *            the linked mode
	 */
	public FixBugsLinkedPopup(final CompilationUnitEditor editor, final FixBugsLinkedMode linkedMode) {
		fEditor = editor;
		fLinkedMode = linkedMode;
		restoreSnapPosition();
	}

	private void restoreSnapPosition() {
		IDialogSettings settings = getDialogSettings();
		try {
			fSnapPosition = settings.getInt(SNAP_POSITION_KEY);
		} catch (NumberFormatException e) {
			fSnapPosition = SNAP_POSITION_TOP_RIGHT;
			// YS default: SNAP_POSITION_UNDER_LEFT_FIELD;
		}
		fSnapPositionChanged = true;
	}

	private IDialogSettings getDialogSettings() {
		return JavaPlugin.getDefault().getDialogSettingsSection(DIALOG_SETTINGS_SECTION);
	}

	/**
	 * 
	 * @see IWidgetTokenKeeper
	 */
	@Override
	public final boolean requestWidgetToken(final IWidgetTokenOwner owner) {
		return false;
	}

	/**
	 * 
	 * @see IWidgetTokenKeeperExtension
	 */
	@Override
	public final boolean requestWidgetToken(final IWidgetTokenOwner owner, final int priority) {
		if (priority > WIDGET_PRIORITY) {
			if (fShell != null && !fShell.isDisposed()) {
				fShell.setVisible(false);
			}
			return true;
		}
		return false;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final boolean setFocus(final IWidgetTokenOwner owner) {
		if (fToolBar != null && !fToolBar.isDisposed()) {
			showMenu(fToolBar);
		}
		return true;
	}

	private void releaseWidgetToken() {
		ISourceViewer viewer = fEditor.getViewer();
		if (viewer instanceof IWidgetTokenOwner) {
			IWidgetTokenOwner widgetTokenOwner = (IWidgetTokenOwner) viewer;
			widgetTokenOwner.releaseWidgetToken(this);
		}
	}

	/**
	 * Creates the SWT controls for the popup.
	 * 
	 * @param parent
	 */
	private void createContent(final Composite parent) {
		Display display = parent.getDisplay();
		// Color foreground = display.getSystemColor(SWT.COLOR_INFO_FOREGROUND);
		Color background = display.getSystemColor(SWT.COLOR_INFO_BACKGROUND);
		addMoveSupport(fShell, parent);

		Composite mainComposite = new Composite(fShell, SWT.NONE/* | SWT.BORDER */);
		GridLayout layout = new GridLayout();
		layout.numColumns = 3; // vertical
		mainComposite.setLayout(layout);

		// +-----------------+---+---+
		// | Message ********| v | x | (close)
		// |-----------------+---+---|
		// | ()Op#1 ()Op#2 ()Op#3 ***| (context)
		// +-------------------------+

		// 1st row
		createMessage(mainComposite); // depends on the context
		hookContextMenu(mainComposite); // common
		createCloseMenu(mainComposite); // common

		// 2nd row
		createSuggestions(mainComposite); // depends on the context

		addMoveSupport(fShell, mainComposite);

		recursiveSetBackgroundColor(parent, background);
	}

	protected static void recursiveSetBackgroundColor(final Control control, final Color color) {
		control.setBackground(color);
		if (control instanceof Composite) {
			Control[] children = ((Composite) control).getChildren();
			for (int i = 0; i < children.length; i++) {
				recursiveSetBackgroundColor(children[i], color);
			}
		}
	}

	private boolean stickToTheFirstLinkedNode = false;

	/**
	 * 
	 * @param stick
	 *            update popup location
	 */
	@Override
	public final void open(final boolean stick) {
		fLogger.info("FixBugsLinkedPopup#open");
		stickToTheFirstLinkedNode = stick;
		// Must cache here, since editor context is not available in menu from
		// popup shell:
		fOpenDialogBinding = getOpenDialogBinding();

		Shell workbenchShell = fEditor.getSite().getShell();
		final Display display = workbenchShell.getDisplay();

		fShell = new Shell(workbenchShell, SWT.ON_TOP | SWT.NO_TRIM | SWT.TOOL);
		fPopupLayout = new GridLayout(2, false);
		fPopupLayout.marginWidth = 1;
		fPopupLayout.marginHeight = 1;
		fPopupLayout.marginLeft = 2;
		fPopupLayout.marginRight = 2;
		fPopupLayout.horizontalSpacing = 0;
		fShell.setLayout(fPopupLayout);

		createContent(fShell);

		updateLocation(true);
		new VisibilityManager().start();

		// Leave linked mode when popup loses focus
		// (except when focus goes back to workbench window or menu is open):
		fShell.addShellListener(new ShellAdapter() {
			@Override
			public void shellDeactivated(final ShellEvent e) {
				if (fIsMenuUp) {
					return;
				}

				final Shell editorShell = fEditor.getSite().getShell();
				display.asyncExec(new Runnable() {
					// post to UI thread since editor shell only gets activated
					// after popup has lost focus
					@Override
					public void run() {
						Shell activeShell = display.getActiveShell();
						if (activeShell != editorShell) {
							// System.out.println("here???");
							fLinkedMode.cancel();
						}
					}
				});
			}
		});

		if (!MAC) { // carbon and cocoa draw their own border...
			fShell.addPaintListener(new PaintListener() {
				@Override
				public void paintControl(final PaintEvent pe) {
					pe.gc.drawPolygon(getPolygon(true));
				}
			});
		}

		// fPopup.moveBelow(null); // make sure hovers are on top of the info
		// popup
		// YS workaround for
		// https://bugs.eclipse.org/bugs/show_bug.cgi?id=170774
		// fPopup.moveBelow(workbenchShell.getShells()[0]);

		UIJob delayJob = new UIJob(display, Messages.popupDelayJobName) {
			@Override
			public IStatus runInUIThread(final IProgressMonitor monitor) {
				fDelayJobFinished = true;
				if (fShell != null && !fShell.isDisposed()) {
					updateVisibility();
				}
				return Status.OK_STATUS;
			}
		};
		delayJob.setSystem(true);
		delayJob.setPriority(Job.INTERACTIVE);
		delayJob.schedule(POPUP_VISIBILITY_DELAY);
	}

	/**
    *
    */
	@Override
	public final void close() {
		fLogger.info("FixBugsLinkedPopup#close");
		if (fShell != null) {
			if (!fShell.isDisposed()) {
				fShell.close();
				fShell.dispose();
			}
			fShell = null;
		}
		releaseWidgetToken();
		if (fRegion != null) {
			if (!fRegion.isDisposed()) {
				fRegion.dispose();
			}
		}
	}

	public boolean isDisposed() {
		if (fShell == null || fShell.isDisposed()) {
			return true;
		}
		return false;
	}

	public final Shell getShell() {
		return fShell;
	}

	public FixBugsLinkedMode getMode() {
		return fLinkedMode;
	}

	public final boolean ownsFocusShell() {
		if (fIsMenuUp) {
			return true;
		}
		if (fShell == null || fShell.isDisposed()) {
			return false;
		}
		Shell activeShell = fShell.getDisplay().getActiveShell();
		if (fShell == activeShell) {
			return true;
		}
		return false;
	}

	/**
	 * Updates popup location.
	 * 
	 * @param force
	 *            true if ...
	 */
	public final void updateLocation(final boolean force) {
		if (!force
				&& (fSnapPosition == SNAP_POSITION_TOP_LEFT || fSnapPosition == SNAP_POSITION_TOP_RIGHT
						|| fSnapPosition == SNAP_POSITION_BOTTOM_LEFT || fSnapPosition == SNAP_POSITION_BOTTOM_RIGHT)) {
			return;
		}

		packPopup();
		Point loc = computeLocation(fSnapPosition);
		// System.out.println("\t" + loc + ", " + fPopup.getLocation());
		if (loc != null && !loc.equals(fShell.getLocation())) {
			fShell.setLocation(loc);
			// YS workaround for
			// https://bugs.eclipse.org/bugs/show_bug.cgi?id=170774
			// fPopup.moveBelow(fEditor.getSite().getShell().getShells()[0]);
		}
	}

	private void updateVisibility() {
		fLogger.info("InteractivePopup#updateVisibility");
		if (fShell != null && !fShell.isDisposed() && fDelayJobFinished) {
			boolean visible = false;
			// YS: Check for visibility of linked position, not whether popup
			// is outside of editor?
			if (fLinkedMode.isCaretInLinkedPosition() || fNeedToHold) {
				StyledText textWidget = fEditor.getViewer().getTextWidget();
				Rectangle eArea = Geometry.toDisplay(textWidget, textWidget.getClientArea());
				Rectangle pBounds = fShell.getBounds();
				pBounds.x -= GAP;
				pBounds.y -= GAP;
				pBounds.width += 2 * GAP;
				pBounds.height += 2 * GAP;
				if (eArea.intersects(pBounds)) {
					visible = true;
				}
			}
			visible = true; // YS forcefully set the visibility true
			if (visible && !fShell.isVisible()) {
				ISourceViewer viewer = fEditor.getViewer();
				if (viewer instanceof IWidgetTokenOwnerExtension) {
					IWidgetTokenOwnerExtension widgetTokenOwnerExtension = (IWidgetTokenOwnerExtension) viewer;
					visible = widgetTokenOwnerExtension.requestWidgetToken(this, WIDGET_PRIORITY);
				}
			} else if (!visible && fShell.isVisible()) {
				releaseWidgetToken();
			}
			fShell.setVisible(visible);
		}
	}

	/** */
	private static final int FIXED_GAP = 2;

	/**
	 * Computes popup location.
	 * 
	 * @param snapPosition
	 *            one of the SNAP_POSITION_* constants
	 * @return the location in display coordinates or <code>null</code> iff not
	 *         visible
	 */
	private Point computeLocation(final int snapPosition) {
		// System.out.println("FixBugsPopup#computePopupLocation");
		if (fShell == null || fShell.isDisposed()) {
			return null;
		}

		switch (snapPosition) {
		case SNAP_POSITION_TOP_LEFT: {
			StyledText eWidget = fEditor.getViewer().getTextWidget();
			Rectangle eBounds = eWidget.getClientArea();
			Point eUpperRight = eWidget.toDisplay(eBounds.x, eBounds.y);
			return new Point(eUpperRight.x + FIXED_GAP, eUpperRight.y + FIXED_GAP);
		}
		case SNAP_POSITION_TOP_RIGHT: {
			StyledText eWidget = fEditor.getViewer().getTextWidget();
			Rectangle eBounds = eWidget.getClientArea();
			Point eUpperRight = eWidget.toDisplay(eBounds.x + eBounds.width, eBounds.y);
			Point pSize = getExtent();
			return new Point(eUpperRight.x - pSize.x - FIXED_GAP, eUpperRight.y + FIXED_GAP);
		}
		case SNAP_POSITION_BOTTOM_LEFT: {
			StyledText eWidget = fEditor.getViewer().getTextWidget();
			Rectangle eBounds = eWidget.getClientArea();
			Point eLowerRight = eWidget.toDisplay(eBounds.x, eBounds.y + eBounds.height);
			Point pSize = getExtent();
			return new Point(eLowerRight.x + FIXED_GAP, eLowerRight.y - pSize.y - FIXED_GAP);
		}
		case SNAP_POSITION_BOTTOM_RIGHT: {
			StyledText eWidget = fEditor.getViewer().getTextWidget();
			Rectangle eBounds = eWidget.getClientArea();
			Point eLowerRight = eWidget.toDisplay(eBounds.x + eBounds.width, eBounds.y + eBounds.height);
			Point pSize = getExtent();
			return new Point(eLowerRight.x - pSize.x - FIXED_GAP, eLowerRight.y - pSize.y - FIXED_GAP);
		}

		case SNAP_POSITION_UNDER_RIGHT_FIELD:
		case SNAP_POSITION_OVER_RIGHT_FIELD: {
			// YS
			LinkedPosition position = null;
			if (stickToTheFirstLinkedNode) {
				position = fLinkedMode.getFirstLinkedPosition();
			} else {
				position = fLinkedMode.getCurrentLinkedPosition();
			}

			if (position == null) {
				return null;
			}
			ISourceViewer viewer = fEditor.getViewer();
			ITextViewerExtension5 viewer5 = (ITextViewerExtension5) viewer;
			int widgetOffset = viewer5.modelOffset2WidgetOffset(position.offset + position.length);

			StyledText textWidget = viewer.getTextWidget();
			Point pos = textWidget.getLocationAtOffset(widgetOffset);
			Point pSize = getExtent();
			if (snapPosition == SNAP_POSITION_OVER_RIGHT_FIELD) {
				pos.y -= pSize.y + GAP;
			} else {
				pos.y += textWidget.getLineHeight(widgetOffset) + GAP;
			}
			pos.x += GAP;
			Point dPos = textWidget.toDisplay(pos);
			Rectangle displayBounds = textWidget.getDisplay().getClientArea();
			Rectangle dPopupRect = Geometry.createRectangle(dPos, pSize);
			Geometry.moveInside(dPopupRect, displayBounds);
			return new Point(dPopupRect.x, dPopupRect.y);
		}

		case SNAP_POSITION_UNDER_LEFT_FIELD:
		case SNAP_POSITION_OVER_LEFT_FIELD:
		default: { // same as SNAP_POSITION_UNDER_LEFT_FIELD
			// YS
			LinkedPosition position = null;
			if (stickToTheFirstLinkedNode) {
				position = fLinkedMode.getFirstLinkedPosition();
			} else {
				position = fLinkedMode.getCurrentLinkedPosition();
			}

			if (position == null) {
				return null;
			}
			ISourceViewer viewer = fEditor.getViewer();
			ITextViewerExtension5 viewer5 = (ITextViewerExtension5) viewer;
			int widgetOffset = viewer5.modelOffset2WidgetOffset(position.offset);

			StyledText textWidget = viewer.getTextWidget();
			Point pos = textWidget.getLocationAtOffset(widgetOffset);
			Point pSize = getExtent();
			pSize.y += HAH + 1;
			pos.x -= HAO;
			if (snapPosition == SNAP_POSITION_OVER_LEFT_FIELD) {
				pos.y -= pSize.y;
			} else {
				pos.y += textWidget.getLineHeight(widgetOffset);
			}
			Point dPos = textWidget.toDisplay(pos);
			Rectangle displayBounds = textWidget.getDisplay().getClientArea();
			Rectangle dPopupRect = Geometry.createRectangle(dPos, pSize);
			Geometry.moveInside(dPopupRect, displayBounds);
			return new Point(dPopupRect.x, dPopupRect.y);
		}

		}
	}

	protected final void addMoveSupport(final Shell popupShell, final Control movedControl) {
		movedControl.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseDown(final MouseEvent downEvent) {
				// System.out.println("MouseAdapter#mouseDown");
				if (downEvent.button != 1) {
					return;
				}

				final Point POPUP_SOURCE = popupShell.getLocation();
				final StyledText textWidget = fEditor.getViewer().getTextWidget();
				Point pSize = getExtent();
				int originalSnapPosition = fSnapPosition;

				/*
				 * Feature in Tracker: it is not possible to directly control
				 * the feedback, see
				 * https://bugs.eclipse.org/bugs/show_bug.cgi?id=121300 and
				 * https://bugs.eclipse.org/bugs/show_bug.cgi?id=121298#c1 .
				 * 
				 * Workaround is to have an offscreen rectangle for tracking
				 * mouse movement and a manually updated rectangle for the
				 * actual drop target.
				 */
				final Tracker tracker = new Tracker(textWidget, SWT.NONE);

				final Point[] LOCATIONS = { //
				textWidget.toControl(computeLocation(SNAP_POSITION_UNDER_RIGHT_FIELD)),
						textWidget.toControl(computeLocation(SNAP_POSITION_OVER_RIGHT_FIELD)),
						textWidget.toControl(computeLocation(SNAP_POSITION_UNDER_LEFT_FIELD)),
						textWidget.toControl(computeLocation(SNAP_POSITION_OVER_LEFT_FIELD)),
						textWidget.toControl(computeLocation(SNAP_POSITION_BOTTOM_RIGHT)),
						textWidget.toControl(computeLocation(SNAP_POSITION_BOTTOM_LEFT)),
						textWidget.toControl(computeLocation(SNAP_POSITION_TOP_LEFT)),
						textWidget.toControl(computeLocation(SNAP_POSITION_TOP_RIGHT)), //
				};

				final Rectangle[] DROP_TARGETS = { //
				Geometry.createRectangle(LOCATIONS[0], pSize), //
						Geometry.createRectangle(LOCATIONS[1], pSize), //
						new Rectangle(LOCATIONS[2].x, LOCATIONS[2].y + HAH, pSize.x, pSize.y), //
						Geometry.createRectangle(LOCATIONS[3], pSize), //
						Geometry.createRectangle(LOCATIONS[4], pSize), //
						Geometry.createRectangle(LOCATIONS[5], pSize), //
						Geometry.createRectangle(LOCATIONS[6], pSize), //
						Geometry.createRectangle(LOCATIONS[7], pSize), //
				};
				final Rectangle MOUSE_MOVE_SOURCE = new Rectangle(1000000, 0, 0, 0);
				tracker.setRectangles(new Rectangle[] { MOUSE_MOVE_SOURCE, DROP_TARGETS[fSnapPosition] });
				tracker.setStippled(true);

				ControlListener moveListener = new ControlAdapter() {
					/*
					 * @see
					 * org.eclipse.swt.events.ControlAdapter#controlMoved(org
					 * .eclipse.swt.events.ControlEvent)
					 */
					@Override
					public void controlMoved(final ControlEvent moveEvent) {
						Rectangle[] currentRects = tracker.getRectangles();
						final Rectangle mouseMoveCurrent = currentRects[0];
						Point popupLoc = new Point(POPUP_SOURCE.x + mouseMoveCurrent.x - MOUSE_MOVE_SOURCE.x, POPUP_SOURCE.y
								+ mouseMoveCurrent.y - MOUSE_MOVE_SOURCE.y);

						popupShell.setLocation(popupLoc);

						Point ePopupLoc = textWidget.toControl(popupLoc);
						int minDist = Integer.MAX_VALUE;
						for (int snapPos = 0; snapPos < DROP_TARGETS.length; snapPos++) {
							int dist = Geometry.distanceSquared(ePopupLoc, LOCATIONS[snapPos]);
							if (dist < minDist) {
								minDist = dist;
								fSnapPosition = snapPos;
								fSnapPositionChanged = true;
								currentRects[1] = DROP_TARGETS[snapPos];
							}
						}
						tracker.setRectangles(currentRects);
					}
				};
				tracker.addControlListener(moveListener);
				boolean committed = tracker.open();
				tracker.close();
				tracker.dispose();
				if (committed) {
					getDialogSettings().put(SNAP_POSITION_KEY, fSnapPosition);
				} else {
					fSnapPosition = originalSnapPosition;
					fSnapPositionChanged = true;
				}
				updateLocation(true);
				activateEditor();
			}
		});
	}

	public void packPopup() {
		if (!fSnapPositionChanged) {
			return;
		}
		fSnapPositionChanged = false;

		boolean isUnderLeft = fSnapPosition == SNAP_POSITION_UNDER_LEFT_FIELD;
		boolean isOverLeft = fSnapPosition == SNAP_POSITION_OVER_LEFT_FIELD;
		fPopupLayout.marginTop = isUnderLeft ? HAH : 0;
		fPopupLayout.marginBottom = isOverLeft ? HAH + 1 : 0;
		fShell.pack();

		Region oldRegion = fRegion;
		if (isUnderLeft || isOverLeft) {
			fRegion = new Region();
			fRegion.add(getPolygon(false));
			fShell.setRegion(fRegion);
			Rectangle bounds = fRegion.getBounds();
			fShell.setSize(bounds.width, bounds.height + 1);
		} else {
			fRegion = null;
			fShell.setRegion(null);
		}

		if (oldRegion != null) {
			oldRegion.dispose();
		}
	}

	private Point getExtent() {
		Point e = fShell.getSize();
		switch (fSnapPosition) {
		case SNAP_POSITION_UNDER_LEFT_FIELD:
			e.y -= HAH;
			break;
		case SNAP_POSITION_OVER_LEFT_FIELD:
			e.y -= HAH + 1;
			break;
		}
		return e;
	}

	private int[] getPolygon(final boolean border) {
		Point e = getExtent();
		int b = border ? 1 : 0;
		boolean isRTL = (fShell.getStyle() & SWT.RIGHT_TO_LEFT) != 0;
		int ha1 = isRTL ? e.x - HAO : HAO + HAW;
		int ha2 = isRTL ? e.x - HAO - HAW / 2 : HAO + HAW / 2;
		int ha3 = isRTL ? e.x - HAO - HAW : HAO;
		int[] poly;
		switch (fSnapPosition) {
		case SNAP_POSITION_OVER_LEFT_FIELD:
			poly = new int[] { 0, 0, //
					e.x - b, 0, //
					e.x - b, e.y - b, //
					ha1, e.y - b, //
					ha2, e.y + HAH - b, //
					ha3, e.y - b, //
					0, e.y - b, //
					0, 0 };
			break;

		case SNAP_POSITION_UNDER_LEFT_FIELD:
			poly = new int[] { 0, HAH, //
					ha3 + b, HAH, //
					ha2, b, ha1 - b, //
					HAH, e.x - b, //
					HAH, e.x - b, //
					e.y + HAH - b, 0, //
					e.y + HAH - b, //
					0, HAH };
			break;

		default:
			poly = new int[] { 0, 0, //
					e.x - b, 0, //
					e.x - b, e.y - b, //
					0, e.y - b, //
					0, 0 };
			break;
		}
		return poly;
	}

	// YS support link on popup for more details.

	private StyledText fMessage = null;

	public void createMessage(final Composite parent) {
		fMessage = new StyledText(parent, SWT.READ_ONLY | SWT.SINGLE);
		setDefaultMessageText();
		fMessage.setEnabled(false);
	}

	public abstract void setDefaultMessageText();

	public void setMessage(String text) {
		fMessage.setText(text);
	}

	private void createCloseMenu(final Composite parent) {
		ToolBar toolBar = new ToolBar(parent, SWT.FLAT);
		final ToolItem menuButton = new ToolItem(toolBar, SWT.PUSH, 0);
		menuButton.setImage(Images.CLOSE.createImage());
		menuButton.setHotImage(Images.CLOSE_ACTIVE.createImage());
		menuButton.setToolTipText(Messages.popupClose);
		toolBar.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(final MouseEvent e) {
				close();
			}
		});
	}

	protected final ToolBar hookContextMenu(final Composite parent) {
		fToolBar = new ToolBar(parent, SWT.FLAT);

		GridData gridData = new GridData();
		gridData.horizontalAlignment = SWT.RIGHT;
		gridData.grabExcessHorizontalSpace = true;
		fToolBar.setLayoutData(gridData);

		final ToolItem menuButton = new ToolItem(fToolBar, SWT.PUSH, 0);
		fMenuImage = Images.VIEW_MENU.createImage();
		menuButton.setImage(fMenuImage);
		menuButton.setToolTipText(Messages.popupMenu);
		fToolBar.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(final MouseEvent e) {
				showMenu(fToolBar);
			}
		});
		menuButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(final SelectionEvent e) {
				showMenu(fToolBar);
			}
		});
		fToolBar.pack();

		return fToolBar;
	}

	private void showMenu(final ToolBar toolBar) {
		Menu menu = getMenuManager().createContextMenu(toolBar);
		menu.setLocation(toolBar.toDisplay(0, toolBar.getSize().y));
		fIsMenuUp = true;
		menu.setVisible(true);
	}

	private MenuManager getMenuManager() {
		if (fMenuManager != null) {
			return fMenuManager;
		}

		fMenuManager = new MenuManager();
		fMenuManager.setRemoveAllWhenShown(true);

		fMenuManager.addMenuListener(new IMenuListener2() {
			@Override
			public void menuAboutToHide(final IMenuManager manager) {
				fIsMenuUp = false;
			}

			@Override
			public void menuAboutToShow(final IMenuManager manager) {
				MenuManager subMenuManager = new MenuManager(Messages.popupSnapTo);
				addMoveMenuItem(subMenuManager, SNAP_POSITION_TOP_LEFT, Messages.popupSnapTopLeft);
				addMoveMenuItem(subMenuManager, SNAP_POSITION_TOP_RIGHT, Messages.popupSnapTopRight);
				addMoveMenuItem(subMenuManager, SNAP_POSITION_BOTTOM_LEFT, Messages.popupSnapBottomLeft);
				addMoveMenuItem(subMenuManager, SNAP_POSITION_BOTTOM_RIGHT, Messages.popupSnapBottomRight);
				addMoveMenuItem(subMenuManager, SNAP_POSITION_OVER_LEFT_FIELD, Messages.popupSnapOverLeft);
				addMoveMenuItem(subMenuManager, SNAP_POSITION_OVER_RIGHT_FIELD, Messages.popupSnapOverRight);
				addMoveMenuItem(subMenuManager, SNAP_POSITION_UNDER_LEFT_FIELD, Messages.popupSnapUnderLeft);
				addMoveMenuItem(subMenuManager, SNAP_POSITION_UNDER_RIGHT_FIELD, Messages.popupSnapUnderRight);
				manager.add(subMenuManager);

				IAction prefsAction = new Action(Messages.popupPreferences) {
					@Override
					public void run() {
						fLinkedMode.cancel();
						String linkedModePrefPageID = "org.eclipse.ui.editors.preferencePages.LinkedModePreferencePage"; //$NON-NLS-1$
						String refactoringPrefPageID = JavaBasePreferencePage.JAVA_BASE_PREF_PAGE_ID;
						PreferencesUtil.createPreferenceDialogOn(fEditor.getSite().getShell(), refactoringPrefPageID,
								new String[] { linkedModePrefPageID, refactoringPrefPageID }, null).open();
					}
				};
				manager.add(prefsAction);
			}
		});
		return fMenuManager;
	}

	private void addMoveMenuItem(final IMenuManager manager, final int snapPosition, final String text) {
		IAction action = new Action(text, IAction.AS_RADIO_BUTTON) {
			@Override
			public void run() {
				fSnapPosition = snapPosition;
				fSnapPositionChanged = true;
				getDialogSettings().put(SNAP_POSITION_KEY, fSnapPosition);
				updateLocation(true);
				activateEditor();
			}
		};
		action.setChecked(fSnapPosition == snapPosition);
		manager.add(action);
	}

	private static String getEnterBinding() {
		return KeyStroke.getInstance(KeyLookupFactory.getDefault().formalKeyLookup(IKeyLookup.CR_NAME)).format();
	}

	/**
	 * WARNING: only works in workbench window context!
	 * 
	 * @return the keybinding for Refactor &gt; Rename
	 */
	private static String getOpenDialogBinding() {
		IBindingService bindingService = (IBindingService) PlatformUI.getWorkbench().getAdapter(IBindingService.class);
		if (bindingService == null) {
			return ""; //$NON-NLS-1$
		}
		String binding = bindingService.getBestActiveBindingFormattedFor(FIX_BUGS);
		return binding == null ? "" : binding; //$NON-NLS-1$
	}

	/** */
	public static final String FIX_BUGS = "fixbugs";

	private void activateEditor() {
		fEditor.getSite().getShell().setActive();
	}

	/*****************************************************/

	/**
	 * Creates suggestions.
	 * 
	 * @param parent
	 *            the parent Composite.
	 */
	public abstract void createSuggestions(final Composite parent);

	/**
	 * Creates an (original) button.
	 * 
	 * @param parent
	 *            the parent Composite.
	 */
	protected final void createOriginalButton(final Composite parent) {
		// 1st option: restore
		fOriginalButton = new Button(parent, SWT.RADIO);
		fOriginalButton.setText(Messages.originalOption);
		// NOTE When the SWT.RADIO style bit is set, the widgetSelected method
		// is also called when the receiver loses, because
		// another item in the same radio group was selected by the user
		// fOriginalButton.addSelectionListener(new SelectionAdapter() {
		// @Override
		// public void widgetSelected(final SelectionEvent e) {
		// fLogger.warning("fOriginalButton#widgetSelected: " +
		// fOriginalButton.getSelection());
		// if (fOriginalButton.getSelection()) {
		// getMode().applyOriginal();
		// fMessage.setText(UIMessages.popupMessage);
		// } else {
		// fLogger.warning("Already Selected. Do nothing!");
		// }
		// }
		// });
		fOriginalButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(MouseEvent e) {
				fLogger.info("fOriginalButton#mouseDown: " + fOriginalButton.getSelection());
				if (!fOriginalButton.getSelection()) {
					getMode().applyOriginal();
					fMessage.setText(Messages.popupMessage);
				} else {
					fLogger.info("Already Selected. Do nothing!");
				}
			}
		});
	}

	protected Button getOriginalButton() {
		return fOriginalButton;
	}

	/**
	 * Sets the selection state of the original button.
	 * 
	 * @param selected
	 *            the new selection state.
	 */
	public final void setOriginalButtonSelected(final boolean selected) {
		fOriginalButton.setSelection(selected);
	}

	public void hide() {
		fShell.setVisible(false);
	 }
	
	public void show() {
		fShell.setVisible(true);
	}
}
