//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions.badPractice.openStreamExceptionPath;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Position;
import org.eclipse.text.edits.MalformedTreeException;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsLinkedAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.SearchHint;
import edu.ncsu.csc.emh.fixbugs.core.resolution.badPractice.openStreamExceptionPath.OpenStreamExceptionPathAssistant;
import edu.ncsu.csc.emh.fixbugs.core.util.PositionTracker;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.FixBugsLinkedMode;

/**
 * 
 * @author Yoonki
 */
public final class OpenStreamExceptionPathMode extends FixBugsLinkedMode {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(OpenStreamExceptionPathMode.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	public static final String SPLIT_AND_MOVE_CLOSE_TO_FINALLY_BLOCK = "SplitAndMoveCloseToFinallyBlock"; //$NON-NLS-1$

	/** */
	public static final String USE_TRY_WITH_RESOURCES = "UseTryWithResources"; //$NON-NLS-1$

	/**
	 * Constructor for FixBugsLinkedMode.
	 * 
	 * @param editor
	 *            the editor.
	 * @param assistant
	 *            the assistant.
	 */
	public OpenStreamExceptionPathMode(final CompilationUnitEditor editor, final FixBugsLinkedAssistant assistant) {
		super(editor, assistant);
	}

	@Override
	public void start(final boolean applyDefault) {
		fLogger.info("OpenStreamExceptionPathMode#start");
		super.start(applyDefault);

		if (applyDefault) {
			applyDefaultSuggestion();
			if (getPopup() != null) {
				getPopup().setCloseInFinallyBlockSelected(true);
			}
		}
	}

	@Override
	protected void openPopup() {
		fPopup = new OpenStreamExceptionPathPopup(getEditor(), this);
		fPopup.open(false);
	}

	/**
	 * 
	 * There is a user operation (writing latency)!!
	 * 
	 * @param baseName
	 *            the base name
	 */
	private void reopenPopup(final String baseName) {
		fLogger.info("OpenStreamExceptionPathMode#reopenPopup()");
		Assert.isNotNull(fPopup);
		fPopup = new OpenStreamExceptionPathPopup(getEditor(), this);
		fPopup.open(false);
		if (SPLIT_AND_MOVE_CLOSE_TO_FINALLY_BLOCK.equals(baseName)) {
			((OpenStreamExceptionPathPopup) fPopup).setCloseInFinallyBlockSelected(true);
		} else {
			((OpenStreamExceptionPathPopup) fPopup).setTryWithResourcesButtonSelected(true);
		}
	}

	@Override
	public OpenStreamExceptionPathPopup getPopup() {
		return (OpenStreamExceptionPathPopup) fPopup;
	}

	@Override
	public OpenStreamExceptionPathAssistant getOriginalAssistant() {
		return (OpenStreamExceptionPathAssistant) fAssistant;
	}

	@Override
	public OpenStreamExceptionPathAssistant getCurrentAssistant() {
		Position p = getFirstLinkedPosition();
		return getCurrentAssistant(p);
	}

	@Override
	public OpenStreamExceptionPathAssistant getCurrentAssistant(final Position selection) {
		fLogger.info("OpenStreamExceptionPathMode#getCurrentAssistant: " + selection);
		ICompilationUnit cu = JavaUI.getWorkingCopyManager().getWorkingCopy(getEditor().getEditorInput());
		FixBugsContext context = new FixBugsContext(cu, selection.getOffset(), selection.getLength());
		SearchHint hint = new SearchHint();
		// TODO We should use hierarchy instead of post-fixes of sub classes!
		hint.add("Stream");
		hint.add("Reader");
		hint.add("JarFile");
		return new OpenStreamExceptionPathAssistant(context, hint);
	}

	@Override
	public FixBugsAssistant applyDefaultSuggestion() {
		fLogger.info("OpenStreamExceptionPathMode#applyDefaultSuggestion");
		return applySuggestion(SPLIT_AND_MOVE_CLOSE_TO_FINALLY_BLOCK, true);
	}

	/**
	 * 
	 * @param suggestion
	 *            string representation of suggestion.
	 * @param fromOriginal
	 * @return
	 */
	public OpenStreamExceptionPathAssistant applySuggestion(final String suggestion, final boolean fromOriginal) {
		fLogger.info("OpenStreamExceptionPathMode#applyCloseInFinallyBlock: " + fromOriginal);
		long startTime = System.currentTimeMillis();
		OpenStreamExceptionPathAssistant currentAssistant = null;
		try {
			// 1. read changes from the original context
			OpenStreamExceptionPathAssistant originalAssistant = getOriginalAssistant(); // original
																							// context
			Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap = new HashMap<ITrackedNodePosition, FixBugsAnnotation>();
			List<ITrackedNodePosition> trackedVariables = new ArrayList<ITrackedNodePosition>();
			String changeText = originalAssistant.changeAST(suggestion, trackedAnnotationMap, trackedVariables);

			// 2. write the changes to the current context
			writeSuggestion(changeText, originalAssistant.getAdjustedSelection());

			// 3. update linked modes and annotations
			List<Position> varLinks = PositionTracker.perform(trackedVariables);
			currentAssistant = getCurrentAssistant(varLinks.get(0));
			currentAssistant.update(PositionTracker.perform(trackedAnnotationMap)); // update
			List<List<Position>> linkedNodesList = new ArrayList<List<Position>>();
			linkedNodesList.add(varLinks);
			refresh(currentAssistant, linkedNodesList);

			if (fPopup != null && fPopup.isDisposed()) {
				reopenPopup(suggestion);
			}
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (MalformedTreeException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		} catch (CoreException e) {
			e.printStackTrace();
		}
		fLogger.warning("#applyCloseInFinallyBlock: " + (System.currentTimeMillis() - startTime) + " ms");
		return currentAssistant;
	}
}
