//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions.badPractice.openStreamExceptionPath;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;

import edu.ncsu.csc.emh.fixbugs.ui.Messages;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.FixBugsLinkedMode;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.FixBugsLinkedPopup;

/**
 * @author Yoonki
 */
public final class OpenStreamExceptionPathPopup extends FixBugsLinkedPopup {

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(OpenStreamExceptionPathPopup.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/**
	 * Constructor for MissingArgumentPopup.
	 * 
	 * @param editor
	 *            the editor.
	 * @param linkedMode
	 *            the linked mode.
	 */
	public OpenStreamExceptionPathPopup(final CompilationUnitEditor editor, final FixBugsLinkedMode linkedMode) {
		super(editor, linkedMode);
	}

	@Override
	public OpenStreamExceptionPathMode getMode() {
		return (OpenStreamExceptionPathMode) fLinkedMode;
	}

	@Override
	public void setDefaultMessageText() {
		setMessage(Messages.openStreamExceptionPathMessage);
	}

	/** */
	private Button fCloseInFinallyBlockButton;

	/** */
	private Button fTryWithResourcesButton;

	@Override
	public void createSuggestions(final Composite mainComposite) {
		fLogger.info("OpenStreamExceptionPathPopup#createSuggestions");
		Composite optionComposite = new Composite(mainComposite, SWT.NONE);

		// gridData
		GridData gridData = new GridData(GridData.FILL, GridData.BEGINNING, true, true);
		gridData.horizontalSpan = 3;
		optionComposite.setLayoutData(gridData);

		// gridLayout
		GridLayout gridLayout = new GridLayout();
		gridLayout.numColumns = 1;
		optionComposite.setLayout(gridLayout);

		// 1st option: original
		createOriginalButton(optionComposite);

		createCloseInFinallyBlockSuggestion(optionComposite);

		createTryWithResourcesSuggestion(optionComposite);
	}

	private void createCloseInFinallyBlockSuggestion(final Composite optionComposite) {
		// 2nd option: Close in new finally block
		// [] split variable declaration
		// [] new finally block
		// [] move close() / or create new close() in the finally block
		fCloseInFinallyBlockButton = new Button(optionComposite, SWT.RADIO);
		fCloseInFinallyBlockButton.setText(Messages.splitAndMoveCloseToFinallyBlock);
		fCloseInFinallyBlockButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(MouseEvent e) {
				fLogger.info("fCloseInFinallyBlockButton#mouseDown");
				if (!fCloseInFinallyBlockButton.getSelection()) {
					getMode().applySuggestion(OpenStreamExceptionPathMode.SPLIT_AND_MOVE_CLOSE_TO_FINALLY_BLOCK,
							getOriginalButton().getSelection());
				} else {
					fLogger.warning("do nothing!");
				}
			}
		});
		fCloseInFinallyBlockButton.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				fLogger.info("fCloseInFinallyBlockButton#mouseDown");
				if (!fCloseInFinallyBlockButton.getSelection()) {
					getMode().applySuggestion(OpenStreamExceptionPathMode.SPLIT_AND_MOVE_CLOSE_TO_FINALLY_BLOCK,
							getOriginalButton().getSelection());
				} else {
					fLogger.warning("do nothing!");
				}
			}
		});
	}

	private void createTryWithResourcesSuggestion(final Composite optionComposite) {
		// 3rd option: try-with-resources (jdk7)
		fTryWithResourcesButton = new Button(optionComposite, SWT.RADIO);
		fTryWithResourcesButton.setText(Messages.useTryWithResources);
		fTryWithResourcesButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(MouseEvent e) {
				fLogger.info("fTryWithResourcesButton#mouseDown");
				if (!fTryWithResourcesButton.getSelection()) {
					getMode().applySuggestion(OpenStreamExceptionPathMode.USE_TRY_WITH_RESOURCES,
							getOriginalButton().getSelection());
				} else {
					fLogger.warning("do nothing!");
				}
			}

		});

		fTryWithResourcesButton.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				fLogger.info("fTryWithResourcesButton#mouseDown");
				if (!fTryWithResourcesButton.getSelection()) {
					getMode().applySuggestion(OpenStreamExceptionPathMode.USE_TRY_WITH_RESOURCES,
							getOriginalButton().getSelection());
				} else {
					fLogger.warning("do nothing!");
				}
			}
		});
	}

	public void setCloseInFinallyBlockSelected(boolean selected) {
		// setOriginalButtonSelected(false);
		fCloseInFinallyBlockButton.setSelection(selected);
		// fTryWithResourcesButton.setSelection(false);
	}

	public void setTryWithResourcesButtonSelected(boolean selected) {
		// setOriginalButtonSelected(false);
		// fCloseInFinallyBlockButton.setSelection(false);
		fTryWithResourcesButton.setSelection(selected);
	}

}
