//
// Copyright (c) 2007-2012 Yoonki Song.
// All rights reserved. This program and the accompanying materials
// are made available under the permission of me.
//
// Contributors:
// Yoonki Song - initial API and implementation
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions.correctness.missingArgument;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.widgets.Display;

/**
 * <code>VariableColorRegistry</code>.
 * 
 * @author Yoonki Song
 * 
 */
public final class VariableColorRegistry {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(VariableColorRegistry.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}
	/** */
	private static VariableColorRegistry instance = null;

	private VariableColorRegistry() {
	}

	/**
	 * 
	 * @return
	 */
	public static synchronized VariableColorRegistry getInstance() {
		if (instance == null) {
			instance = new VariableColorRegistry();
		}
		return instance;
	}

	/**
	 * 
	 */
	public void init() {
		fLogger.info("VariableColorRegistry#init()");
		Display d = Display.getDefault();
		fColorList = new ArrayList<Color>();
		fColorList.add(new Color(d, 75, 254, 120)); // 0, 255, 0
		fColorList.add(new Color(d, 51, 153, 255));
		fColorList.add(new Color(d, 255, 153, 51));
		fColorList.add(new Color(d, 51, 255, 153));
		fColorList.add(new Color(d, 255, 0, 255));
		fColorList.add(new Color(d, 0, 0, 255));
		fColorList.add(new Color(d, 0, 255, 255));
		fColorList.add(new Color(d, 153, 51, 255));
		fColorList.add(new Color(d, 153, 255, 51));
		fColorList.add(new Color(d, 255, 51, 153));
		fColorList.add(new Color(d, 255, 0, 0));
		fVariableColorMap = new TreeMap<String, Integer>();
	}

	/** */
	private Map<String, Integer> fVariableColorMap;

	/** */
	private List<Color> fColorList = null;

	/**
	 * Registers new exception and its color.
	 * 
	 * @param exceptionName
	 *            exception name.
	 * @param colorIndex
	 *            color index.
	 */
	public void register(final String exceptionName, final Integer colorIndex) {
		// let's keep the previous entry
		if (!fVariableColorMap.containsKey(exceptionName)) {
			fVariableColorMap.put(exceptionName, colorIndex);
		}
	}

	/**
	 * Returns the color of given exception.
	 * 
	 * @param exceptionName
	 *            exception name.
	 * @return the color
	 */
	@SuppressWarnings("boxing")
	public Color getColor(final String exceptionName) {
		Integer colorIndex = fVariableColorMap.get(exceptionName);
		if (colorIndex == null) { // UNKNOWN
			return fColorList.get(0);
		}
		return fColorList.get(colorIndex);
	}

	/**
	 * Returns the index.
	 * 
	 * @param exceptionName
	 *            exception name.
	 * @return the index.
	 */
	public int getIndex(final String exceptionName) {
		Integer colorIndex = fVariableColorMap.get(exceptionName);
		if (colorIndex == null) { // Not found
			return -1;
		}
		return colorIndex.intValue();
	}

	public void dispose() {
		if (fColorList != null) {
			for (Color c : fColorList) {
				if (!c.isDisposed()) {
					c.dispose();
				}
			}
			fColorList = null;
		}
		fLogger.info("fExceptionColorMap = null;");
		fVariableColorMap = null;
	}

	public void clear() {
		fColorList.clear();
		fVariableColorMap.clear();
	}

	public void refresh() {
		dispose();
		init();
	}
}
