//
// Copyright (c) 2007-2012 Yoonki Song.
// All rights reserved. This program and the accompanying materials
// are made available under the permission of me.
//
// Contributors:
// Yoonki Song - initial API and implementation
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions.correctness.missingArgument;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.ui.PreferenceConstants;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.text.ITextViewerExtension5;
import org.eclipse.jface.text.JFaceTextUtil;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.TextViewer;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.util.Geometry;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.MouseTrackListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.LineAttributes;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Tracker;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.annotation.VariableAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.correctness.missingArgument.MissingArgumentAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.correctness.missingArgument.VariableDropPosition;
import edu.ncsu.csc.emh.fixbugs.core.resolution.correctness.missingArgument.VariableDropPositionFinder;

/**
 * This class is the View class of MVC pattern.
 * 
 * @author Yoonki Song
 */
public final class VariableDNDManager {

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(VariableDNDManager.class.getName());

	// ------------------- Mouse Handling -----------------------------//

	static {
		fLogger.setLevel(Level.WARNING);
	}

	private class MoustEventListener extends MouseAdapter implements MouseMoveListener, MouseTrackListener {

		public void start() {
			fLogger.fine("VariableDNDManager$EventListener#start");
			final ISourceViewer viewer = fEditor.getViewer();
			final StyledText textWidget = viewer.getTextWidget();
			textWidget.addMouseMoveListener(MoustEventListener.this); // important
			textWidget.addMouseListener(MoustEventListener.this); // important
			textWidget.addMouseTrackListener(MoustEventListener.this); // important
		}

		public void stop() {
			fLogger.fine("VariableDNDManager$EventListener#stop");
			final ISourceViewer viewer = fEditor.getViewer();
			final StyledText textWidget = viewer.getTextWidget();
			if (!textWidget.isDisposed()) {
				textWidget.removeMouseMoveListener(MoustEventListener.this); // remove
				textWidget.removeMouseListener(MoustEventListener.this); // remove
				textWidget.removeMouseTrackListener(MoustEventListener.this); // remove
			}
		}

		/**
		 * Computes the document offset underlying the given text widget
		 * coordinates. This method uses a linear search as it cannot make any
		 * assumption about how the document is actually presented in the
		 * widget. (Covers cases such as bidirectional text.)
		 * 
		 * @param x
		 *            the horizontal coordinate inside the text widget
		 * @param y
		 *            the vertical coordinate inside the text widget
		 * @return the document offset corresponding to the given point
		 */
		private int computeOffsetAtLocation(final TextViewer textViewer, final int x, final int y) {
			try {
				StyledText textWidget = textViewer.getTextWidget();
				int widgetOffset = textWidget.getOffsetAtLocation(new Point(x, y));
				Point p = textWidget.getLocationAtOffset(widgetOffset);
				if (p.x > x) {
					widgetOffset--;
				}

				if (textViewer instanceof ITextViewerExtension5) {
					ITextViewerExtension5 extension = (ITextViewerExtension5) textViewer;
					return extension.widgetOffset2ModelOffset(widgetOffset);
				}

				return widgetOffset; // YS
										// fTextViewer._getVisibleRegionOffset();

			} catch (IllegalArgumentException e) {
				return -1;
			}
		}

		@Override
		public void mouseDown(MouseEvent e) {
			// mouseMove(e);
		}

		@Override
		public void mouseEnter(MouseEvent e) {
		}

		@Override
		public void mouseExit(MouseEvent e) {
		}

		// @Override
		// public void mouseDoubleClick(MouseEvent e) {
		// // do nothing
		// }

		@Override
		public void mouseHover(MouseEvent e) {
			// System.out.println("#MouseHover");
			if (fDragAnnotation != null) {
				System.out.println("fDragAnnotation != null");
			}
		}

		// ---- MouseTrackListener

		@Override
		public void mouseMove(MouseEvent e) {
			fLogger.fine("VariableDNDManager$EventListener#mouseMove: " + e.x + ", " + e.y);
			ISourceViewer viewer = fEditor.getViewer();
			TextViewer textViewer = ((TextViewer) viewer);
			int offset = computeOffsetAtLocation(textViewer, e.x, e.y);
			FixBugsAnnotation annotation = fAssistant.findAnnotation(offset);

			if (annotation != null && annotation instanceof VariableAnnotation) {
				// the annotation is already presented!!
				if (fAssistant.isPresentedVariableName(annotation.getName())) {
					// do nothing!
				} else if (fDragAnnotation != annotation) {
					fLogger.info("Selected (Hover): " + annotation);
					closeShell(fDragShell); // for double-check
					Position pos = new Position(annotation.getPosition().getOffset(), annotation.getPosition().getLength());
					String name = annotation.getName();
					fDragShell = openDragShell(pos, name);
					fDragAnnotation = annotation;
					fMode.removeAnnotation(annotation); // important
					fLogger.info("Removed: " + annotation);

					// FixBugsAnnotation presentedVarAnno =
					// fAssistant.getPresentedVariableAnnotation();
					// if (presentedVarAnno != null) {
					// fMode.removeAnnotation(presentedVarAnno);
					// // System.out.println("removed!!!!!");
					// }
				}
			} else {
				fLogger.finer("No annotation");
				if (fDragAnnotation != null) {
					fMode.addAnnotation(fDragAnnotation); // important
					fLogger.info("Added: " + annotation);
					closeShell(fDragShell);
					fDragAnnotation = null;
				} else {
					// System.out.println("??");
				}
			}
		}

		// @Override
		// public void mouseUp(MouseEvent e) {
		// // do nothing
		// }

	}

	/** */
	private final MissingArgumentMode fMode;

	/** */
	private final MissingArgumentAssistant fAssistant;

	/** */
	private final CompilationUnitEditor fEditor;

	/** */
	private StyledText fEditorWidget;

	/** do not dispose. */
	private Font fFont = null;

	/** */
	private int fAvgCharWidth = 0;

	/** */
	private int fLineHeight = 0;

	/** shell for currently selected exception to move. */
	private Shell fDragShell = null;

	/** annotation for currently selected exception to move. */
	private FixBugsAnnotation fDragAnnotation = null;

	/** possible drop positions for given exception being dragged. */
	private List<Shell> fDropShells = null;

	/** currently selected drop position. */
	private int fSnapPosition = 0;

	/** listener for mouse tracker. */
	private MoustEventListener fEventListener = null;

	/** */
	boolean DROPZONE_REMOVED = false;

	/**
	 * 
	 * @param editor
	 *            the Editor
	 * @param mode
	 *            the annotation mode
	 * @param assistant
	 *            the assistant.
	 */
	public VariableDNDManager(final CompilationUnitEditor editor, final MissingArgumentMode mode,
			final MissingArgumentAssistant assistant) {
		fEditor = editor;
		fMode = mode;
		fAssistant = assistant;
	}

	/**
	 * 
	 * @param parent
	 *            the parent shell
	 * @param label
	 *            control to move
	 * @param dragPosition
	 * @param name
	 */
	private void addMoveSupport(final Shell parent, final Control label, final Position dragPosition, final String name) {
		label.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(final MouseEvent me) {
				if (me.button != 1) {
					return;
				}
				fLogger.info("MouseAdapter#MouseEvent: " + me.x + ", " + me.y);

				final VariableDropPosition[] dropPositions = openDropShells(dragPosition, name);
				final Rectangle[] dropRectangles = calculateAllBounds(dropPositions);

				final Point POPUP_SOURCE = parent.getLocation(); // shell
				final StyledText textWidget = fEditor.getViewer().getTextWidget();
				final Tracker tracker = new Tracker(textWidget, SWT.NONE);
				int originalSnapPosition = fSnapPosition;

				final List<Point> LOCATIONS = new ArrayList<Point>();
				final List<Rectangle> DROP_TARGETS = new ArrayList<Rectangle>();
				for (Rectangle dropRectangle : dropRectangles) {
					// System.out.println("DropRectangle: " + dropRectangle);
					Point location = textWidget.toControl(Geometry.getLocation(dropRectangle));
					LOCATIONS.add(location);
					Point size = Geometry.getSize(dropRectangle);
					// location.x = location.x + size.x - CURSOR_WIDTH;
					// size.x = CURSOR_WIDTH;
					DROP_TARGETS.add(Geometry.createRectangle(location, size));
				}

				final Rectangle MOUSE_MOVE_SOURCE = new Rectangle(1000000, 0, 0, 0);
				tracker.setRectangles(new Rectangle[] { MOUSE_MOVE_SOURCE, DROP_TARGETS.get(fSnapPosition) });
				tracker.setStippled(true);

				ControlListener moveListener = new ControlAdapter() {
					@Override
					public void controlMoved(final ControlEvent e) {
						// System.out.println("ControlListener#controlMoved");

						if (!DROPZONE_REMOVED) {
							FixBugsAnnotation presentedVarAnno = fAssistant.getPresentedVariableAnnotation();
							if (presentedVarAnno != null) {
								// System.out.println("PresentedVarAnno: " +
								// presentedVarAnno);
								fMode.removeAnnotation(presentedVarAnno);
								fAssistant.setPresentedVariableName(null);
								DROPZONE_REMOVED = true;
							} else {
								// System.out.println("! DROPZONE_REMOVED failed");
							}
						}

						Rectangle[] currentRects = tracker.getRectangles();
						final Rectangle currentMouseMove = currentRects[0];
						Point popupLoc = new Point(POPUP_SOURCE.x + currentMouseMove.x - MOUSE_MOVE_SOURCE.x, POPUP_SOURCE.y
								+ currentMouseMove.y - MOUSE_MOVE_SOURCE.y);
						parent.setLocation(popupLoc); // shell
						// NOTE distance calculation is based on the mouse
						// cursor.
						// if the calculation is based on the offset of the
						// selected popup,
						// it need to delete the two following lines of code
						popupLoc.x += me.x;
						popupLoc.y += me.y;
						Point mouseLocInEditor = textWidget.toControl(popupLoc);
						int minDist = Integer.MAX_VALUE;
						int snapPos = 0;
						for (Rectangle rect : DROP_TARGETS) {
							int dist = Geometry.distanceSquared(mouseLocInEditor, LOCATIONS.get(snapPos));
							if (dist < minDist) {
								minDist = dist;
								fSnapPosition = snapPos;
								// fSnapPositionChanged = true;
								currentRects[1] = DROP_TARGETS.get(snapPos); // rect
							}
							snapPos++;
						}
						tracker.setRectangles(currentRects);
					}
				};
				tracker.addListener(SWT.Selection, new Listener() {

					@Override
					public void handleEvent(Event event) {
						System.out.println("Event: " + event);

					}
				});
				tracker.addControlListener(moveListener);
				boolean committed = tracker.open();
				tracker.close();
				tracker.dispose();

				if (committed) {
					String variableName = name; // label.getText();
					fLogger.info("Committed: @" + LOCATIONS.get(fSnapPosition) + ", fDropPosition: " + fSnapPosition);
					VariableDropPosition dropPosition = dropPositions[fSnapPosition];
					closeDropShells();
					closeShell(fDragShell);
					if (dragPosition.includes(dropPosition.getOffset())) {
						System.out.println("Original; do nothing");
						fMode.addAnnotation(fDragAnnotation); // important

						if (DROPZONE_REMOVED) {
							System.out.println("Restore the presented variable annotation");
							FixBugsAnnotation presentedVarAnno = fAssistant.getPresentedVariableAnnotation();
							if (presentedVarAnno != null) {
								fMode.addAnnotation(presentedVarAnno);
							}
						}
					} else {
						fMode.performDND(variableName, dropPosition);
					}
				} else {
					fSnapPosition = originalSnapPosition;
					// fSnapPositionChanged = true;
				}
			}

		});
		// label.addDragDetectListener(new DragDetectListener() {
		//
		// @Override
		// public void dragDetected(DragDetectEvent e) {
		// System.out.println("DragDetectEvent");
		// }
		//
		// });
	}

	/**
	 * Converts the array of drop positions into the array of rectangles in
	 * display coordinates.
	 * 
	 * @param positions
	 *            the array of drop positions
	 * @return
	 */
	private Rectangle[] calculateAllBounds(final Position[] positions) {
		List<Rectangle> dropRectangles = new ArrayList<Rectangle>();
		for (Position position : positions) {
			dropRectangles.add(calculateBounds(position));
		}
		return dropRectangles.toArray(new Rectangle[0]);
	}

	/**
	 * Calculates the model position of the given object into display
	 * coordinates.
	 * 
	 * @param position
	 *            the model position in editor
	 * @return the display coordinates of the local model position as Rectangle.
	 */
	private Rectangle calculateBounds(final Position position) {
		ITextViewerExtension5 viewer5 = (ITextViewerExtension5) fEditor.getViewer();

		int widgetOffsetStart = viewer5.modelOffset2WidgetOffset(position.offset);
		Point documentOffsetStart = fEditorWidget.getLocationAtOffset(widgetOffsetStart);

		int widgetOffsetEnd = viewer5.modelOffset2WidgetOffset(position.offset + position.length);
		Point documentOffsetEnd = fEditorWidget.getLocationAtOffset(widgetOffsetEnd);

		int modelWidth = documentOffsetEnd.x - documentOffsetStart.x;
		if (modelWidth <= 0) { // important, at least the width of 1 character;
			modelWidth = fAvgCharWidth / 2;
		}

		Point displayOffset = fEditorWidget.toDisplay(documentOffsetStart);
		return new Rectangle(displayOffset.x, displayOffset.y, modelWidth, fLineHeight);
	}

	private void closeDropShells() {
		fLogger.info("VariableDNDManager#closeDropShells");
		if (fDropShells != null) {
			for (Shell shell : fDropShells) {
				closeShell(shell);
			}
			fDropShells = null;
		}
	}

	private void closeShell(Shell shell) {
		if (shell != null) {
			if (!shell.isDisposed()) {
				shell.close();
				shell.dispose(); // .close();
			}
			shell = null;
		}
	}

	private void createDragLabel(final Shell parent, final Position dragPosition, final String name) {
		final Label label = new Label(parent, SWT.NONE);
		label.setText(name);
		label.setFont(fFont); // text
		label.setLocation(0, 0); // relative location of parent
		label.setSize(parent.getSize());
		final Color labelBg = VariableColorRegistry.getInstance().getColor(name);
		label.setBackground(labelBg);

		final Color borderFg = fEditorWidget.getDisplay().getSystemColor(SWT.COLOR_BLACK);
		label.addPaintListener(new PaintListener() {
			@Override
			public void paintControl(final PaintEvent e) {
				Rectangle rect = label.getBounds();
				rect.width -= 1;
				rect.height -= 1;
				e.gc.setForeground(borderFg);
				e.gc.setLineAttributes(new LineAttributes(3, SWT.CAP_FLAT, SWT.JOIN_MITER, SWT.LINE_DOT, null, 5, 10));
				e.gc.drawRectangle(rect);
				e.gc.dispose();
			}
		});

		addMoveSupport(parent, label, dragPosition, name);
	}

	/**
	 * 
	 * @param parent
	 *            the shell containing exception
	 * @param dropPosition
	 *            the model position in editor
	 */
	private void createDropLabel(final Composite parent, final VariableDropPosition dropPosition, Color color) {
		final Label label = new Label(parent, SWT.NONE);
		label.setFont(fFont);
		label.setLocation(0, 0);
		label.setSize(parent.getSize());
		label.setBackground(color);
	}

	/**
	 * gathers graphic (font, position, ...) information. It should be updated
	 * by user actions/preferences.
	 */
	private void gatherGraphicInformation() {
		fFont = JFaceResources.getFont(PreferenceConstants.EDITOR_TEXT_FONT); // Important!
		fEditorWidget = fEditor.getViewer().getTextWidget();
		fLineHeight = fEditorWidget.getLineHeight();
		fAvgCharWidth = JFaceTextUtil.getAverageCharWidth(fEditorWidget);
	}

	private Shell openDragShell(final Position position, final String name) {
		Shell workbenchShell = fEditor.getSite().getShell();
		final Shell shell = new Shell(workbenchShell, SWT.NO_TRIM);
		shell.setBounds(calculateBounds(position));
		shell.setVisible(true);
		shell.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(final MouseEvent me) {
				super.mouseDown(me);
				fLogger.info("shell. mouse down");
			}
		});
		createDragLabel(shell, position, name);
		return shell;
	}

	/**
	 * 
	 * @param dropPosition
	 *            the model position in document
	 * @return
	 */
	private Shell openDropShell(final VariableDropPosition dropPosition, Color color) {
		Shell workbenchShell = fEditor.getSite().getShell();
		final Shell shell = new Shell(workbenchShell, SWT.NO_TRIM);
		Rectangle rect = calculateBounds(dropPosition);
		shell.setBounds(rect);
		shell.setVisible(true);
		shell.setAlpha(128);
		createDropLabel(shell, dropPosition, color);
		return shell;
	}

	/**
	 * 
	 * @param annotation
	 *            the selected annotation
	 * @return
	 */
	private VariableDropPosition[] openDropShells(final Position position, final String name) {
		VariableDropPosition[] dropPositions = VariableDropPositionFinder.perform(fAssistant, position, name);
		fDropShells = new ArrayList<Shell>();
		Color color = fEditorWidget.getDisplay().getSystemColor(SWT.COLOR_GRAY); // default
		for (VariableDropPosition varpos : dropPositions) {
			// System.out.println("[DropPosition] " + position);
			fDropShells.add(openDropShell(varpos, color));
		}
		return dropPositions;
	}

	/**
	 *
	 */
	public void start() {
		fLogger.info("VariableDNDManager#start");
		gatherGraphicInformation(); // re-gather whenever called.
		fEventListener = new MoustEventListener();
		fEventListener.start();
	}

	/**
	 *
	 */
	public void stop() {
		fLogger.info("VariableDNDManager#stop");
		closeShell(fDragShell);
		if (fEventListener != null) {
			fEventListener.stop();
			fEventListener = null;
		}
		if (fDragAnnotation != null) {
			fDragAnnotation = null;
		}
	}

}
