//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions.dodgyCode.catchException;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;

import edu.ncsu.csc.emh.fixbugs.ui.Messages;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.FixBugsAnnotationMode;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.FixBugsAnnotationPopup;

/**
 * @author Yoonki
 * 
 */
public final class CatchExceptionPopup extends FixBugsAnnotationPopup {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(CatchExceptionPopup.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	private Button fDefaultSuggestionButton;

	/** */
	private Button fLastChangeButton;

	/**
	 * Constructor for MissingArgumentPopup.
	 * 
	 * @param editor
	 *            the editor.
	 * @param annotationMode
	 *            the annotation mode.
	 */
	public CatchExceptionPopup(final CompilationUnitEditor editor, final FixBugsAnnotationMode annotationMode) {
		super(editor, annotationMode);
	}

	@Override
	public CatchExceptionMode getMode() {
		return (CatchExceptionMode) fAnnotationMode;
	}

	@Override
	public void createSuggestions(final Composite parent) {
		fLogger.info("CatchExceptionPopup#createSuggestions");
		Composite optionComposite = new Composite(parent, SWT.NONE);
		// optionComposite.setLayout(new RowLayout());

		// gridData
		GridData gridData = new GridData(GridData.FILL, GridData.BEGINNING, true, true);
		gridData.horizontalSpan = 3;
		optionComposite.setLayoutData(gridData);

		// gridLayout
		GridLayout gridLayout = new GridLayout();
		gridLayout.numColumns = 1;
		optionComposite.setLayout(gridLayout);

		// 1st option: original
		createOriginalButton(optionComposite);

		// 2nd option:
		createDefaultSuggestion(optionComposite);

		// last action
		createLastChange(optionComposite);
	}

	@Override
	public void setDefaultMessageText() {
		setMessage(Messages.catchExceptionMessage);
	}

	private void createDefaultSuggestion(final Composite parent) {
		fDefaultSuggestionButton = new Button(parent, SWT.RADIO);
		fDefaultSuggestionButton.setText(Messages.defaultSuggestion);
		fDefaultSuggestionButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(MouseEvent e) {
				fLogger.info("fDefaultSuggestion#mouseDown: " + fDefaultSuggestionButton.getSelection());
				if (!fDefaultSuggestionButton.getSelection()) {
					setMessage(Messages.missingArgumentMessage);
					getMode().applyDefaultSuggestion();
				} else {
					fLogger.info("do nothing!");
				}
			}
		});
	}

	private void createLastChange(final Composite parent) {
		fLastChangeButton = new Button(parent, SWT.RADIO);
		fLastChangeButton.setText(Messages.myLastChange);
		fLastChangeButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(MouseEvent e) {
				fLogger.info("fLastActionButton#mouseDown: " + fLastChangeButton.getSelection());
				if (!fLastChangeButton.getSelection()) {
					setMessage(Messages.missingArgumentMessage);
					getMode().applyLastChange();
				} else {
					fLogger.info("do nothing!");
				}
			}
		});
		fLastChangeButton.setEnabled(false);
	}

	public void setDefaultSuggestion(final boolean selected) {
		fDefaultSuggestionButton.setSelection(selected);
	}

	public void setLastChange(boolean set) {
		fLastChangeButton.setSelection(set);
		fLastChangeButton.setEnabled(set);

		if (set) {
			super.setOriginalButton(false);
			setDefaultSuggestion(false);
		}
	}
}
