//
// Copyright (c) 2007-2012 Yoonki Song.
// All rights reserved. This program and the accompanying materials
// are made available under the permission of me.
//
// Contributors:
// Yoonki Song - initial API and implementation
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions.dodgyCode.catchException;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.ui.PreferenceConstants;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.text.ITextViewerExtension5;
import org.eclipse.jface.text.JFaceTextUtil;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.TextViewer;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.util.Geometry;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.LineAttributes;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Tracker;

import edu.ncsu.csc.emh.fixbugs.core.annotation.ExceptionAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException.CatchExceptionAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException.ExceptionDataModel;
import edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException.ExceptionDropPosition;
import edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException.ExceptionDropPositionFinder;

/**
 * This class is the View class of MVC pattern.
 * 
 * @see ExceptionDataModel (Model)
 * @see CatchExceptionMode (Controller)
 * 
 * @author Yoonki Song
 */
public final class ExceptionDNDManager {

	private class MouseEventListener implements MouseMoveListener {

		/**
		 * Computes the document offset underlying the given text widget
		 * coordinates. This method uses a linear search as it cannot make any
		 * assumption about how the document is actually presented in the
		 * widget. (Covers cases such as bidirectional text.)
		 * 
		 * @param x
		 *            the horizontal coordinate inside the text widget
		 * @param y
		 *            the vertical coordinate inside the text widget
		 * @return the document offset corresponding to the given point
		 */
		private int computeOffsetAtLocation(final TextViewer textViewer, final int x, final int y) {
			try {
				StyledText textWidget = textViewer.getTextWidget();
				int widgetOffset = textWidget.getOffsetAtLocation(new Point(x, y));
				Point p = textWidget.getLocationAtOffset(widgetOffset);
				if (p.x > x) {
					widgetOffset--;
				}

				if (textViewer instanceof ITextViewerExtension5) {
					ITextViewerExtension5 extension = (ITextViewerExtension5) textViewer;
					return extension.widgetOffset2ModelOffset(widgetOffset);
				}
				// replaced return fTextViewer._getVisibleRegionOffset();
				return widgetOffset;

			} catch (IllegalArgumentException e) {
				return -1;
			}
		}

		@Override
		public void mouseMove(final MouseEvent e) {
			fLogger.info("ExceptionDNDManager$EventListener#mousemove: " + e.x + ", " + e.y);
			ISourceViewer viewer = fEditor.getViewer();
			TextViewer textViewer = ((TextViewer) viewer);
			int offset = computeOffsetAtLocation(textViewer, e.x, e.y);
			FixBugsAnnotation annotation = fAssistant.findAnnotation(offset);

			if (annotation != null) {
				if (fDragAnnotation != annotation) {
					// System.out.println("Selected (Hover): " + annotation);
					closeShell(fDragShell); // important!
					if (fDragAnnotation != null) {
						fMode.addAnnotation(fDragAnnotation); // important
					}
					Position pos = new Position(annotation.getPosition().getOffset(), annotation.getPosition().getLength());
					String name = annotation.getName();
					fDragShell = openDragShell(pos, name, ((ExceptionAnnotation) annotation).isThrown());
					fDragAnnotation = annotation;
					fMode.removeAnnotation(annotation); // important
				} else {
					// System.out.println("??");
				}
			} else {
				// System.out.println("No annotation");
				if (fDragAnnotation != null) {
					fMode.addAnnotation(fDragAnnotation); // important
					fLogger.info("Added: " + annotation);
					closeShell(fDragShell);
					fDragAnnotation = null;
				} else if (annotation != null) {
					// System.out.println("!!");
				} else {
					// System.out.println("?2323!!");
				}
			}
		}

		public void start() {
			fLogger.info("ExceptionDNDManager$EventListener#start");
			final ISourceViewer viewer = fEditor.getViewer();
			final StyledText textWidget = viewer.getTextWidget();
			textWidget.addMouseMoveListener(MouseEventListener.this); // important
		}

		public void stop() {
			fLogger.info("ExceptionDNDManager$EventListener#stop");
			final ISourceViewer viewer = fEditor.getViewer();
			final StyledText textWidget = viewer.getTextWidget();
			if (!textWidget.isDisposed()) {
				textWidget.removeMouseMoveListener(MouseEventListener.this); // remove
			}
		}
	}

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(ExceptionDNDManager.class.getName());

	// ------------------- Mouse Handling -----------------------------//

	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	private final CatchExceptionMode fMode;

	/** */
	private final CatchExceptionAssistant fAssistant;

	/** */
	private final CompilationUnitEditor fEditor;

	/** */
	private StyledText fEditorWidget;

	/** do not dispose. */
	private Font fFont = null;

	/** */
	private int fAvgCharWidth = 0;

	/** */
	private int fLineHeight = 0;

	/** shell for currently selected exception to move. */
	private Shell fDragShell = null;

	/** annotation for currently selected exception to move. */
	private FixBugsAnnotation fDragAnnotation = null;

	/** possible drop positions for given exception being dragged. */
	private List<Shell> fDropShells = null;

	/** currently selected drop position. */
	private int fSnapPosition = 0;

	/** listener for mouse tracker. */
	private MouseEventListener fEventListener = null;

	/**
	 * 
	 * @param editor
	 *            the Editor
	 * @param mode
	 *            the annotation mode
	 * @param assistant
	 *            the assistant.
	 */
	public ExceptionDNDManager(final CompilationUnitEditor editor, final CatchExceptionMode mode,
			final CatchExceptionAssistant assistant) {
		fEditor = editor;
		fMode = mode;
		fAssistant = assistant;
	}

	/**
	 * Converts the array of drop positions into the array of rectangles in
	 * display coordinates.
	 * 
	 * @param positions
	 *            the array of drop positions
	 * @return
	 */
	private Rectangle[] calculateAllBounds(final Position[] positions) {
		List<Rectangle> dropRectangles = new ArrayList<Rectangle>();
		for (Position position : positions) {
			dropRectangles.add(calculateBounds(position));
		}
		return dropRectangles.toArray(new Rectangle[0]);
	}

	/**
	 * Calculates the model position of the given object into display
	 * coordinates.
	 * 
	 * @param position
	 *            the model position in editor
	 * @return the display coordinates of the local model position as Rectangle.
	 */
	private Rectangle calculateBounds(final Position position) {
		ITextViewerExtension5 viewer5 = (ITextViewerExtension5) fEditor.getViewer();

		int widgetOffsetStart = viewer5.modelOffset2WidgetOffset(position.offset);
		Point documentOffsetStart = fEditorWidget.getLocationAtOffset(widgetOffsetStart);

		int widgetOffsetEnd = viewer5.modelOffset2WidgetOffset(position.offset + position.length);
		Point documentOffsetEnd = fEditorWidget.getLocationAtOffset(widgetOffsetEnd);

		int modelWidth = documentOffsetEnd.x - documentOffsetStart.x;
		if (modelWidth <= 0) { // important, at least the width of 1 character;
			modelWidth = fAvgCharWidth / 2;
		}

		Point displayOffset = fEditorWidget.toDisplay(documentOffsetStart);
		return new Rectangle(displayOffset.x, displayOffset.y, modelWidth, fLineHeight);
	}

	private void closeDropShells() {
		fLogger.info("ExceptionDNDManager#closeDropShells");
		if (fDropShells != null) {
			for (Shell shell : fDropShells) {
				closeShell(shell);
			}
			fDropShells = null;
		}
	}

	private void closeShell(Shell shell) {
		if (shell != null) {
			if (!shell.isDisposed()) {
				shell.close();
				shell.dispose(); // .close();
			}
			shell = null;
		}
	}

	private void createDragLabel(final Composite parent, final Position dragPosition, final String name, final boolean isThrown) {
		final Label label = new Label(parent, SWT.NONE);
		label.setText(name);
		label.setFont(fFont); // text
		label.setLocation(0, 0); // relative location of parent
		label.setSize(parent.getSize());
		final Color labelBg = ExceptionColorRegistry.getInstance().getColor(name);
		label.setBackground(labelBg);
		label.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(final MouseEvent me) {
				super.mouseDown(me);
				fLogger.info("Selected Annotation: ");
				fLogger.info("MouseEvent: " + me.x + ", " + me.y);
				final ExceptionDropPosition[] dropPositions = openDropShells(dragPosition, name, isThrown);
				final Rectangle[] dropRectangles = calculateAllBounds(dropPositions);

				final Point popupSource = parent.getLocation(); // shell
				final StyledText textWidget = fEditor.getViewer().getTextWidget();
				final Tracker tracker = new Tracker(textWidget, SWT.NONE);
				int originalSnapPosition = fSnapPosition;

				final List<Point> locations = new ArrayList<Point>();
				final List<Rectangle> dropTargets = new ArrayList<Rectangle>();
				for (Rectangle dropRectangle : dropRectangles) {
					// System.out.println("DropRectangle: " + dropRectangle);
					Point location = textWidget.toControl(Geometry.getLocation(dropRectangle));
					locations.add(location);
					Point size = Geometry.getSize(dropRectangle);
					// location.x = location.x + size.x - CURSOR_WIDTH;
					// size.x = CURSOR_WIDTH;
					dropTargets.add(Geometry.createRectangle(location, size));
				}

				final Rectangle mouseMoveSource = new Rectangle(1000000, 0, 0, 0);
				tracker.setRectangles(new Rectangle[] { mouseMoveSource, dropTargets.get(fSnapPosition) });
				tracker.setStippled(true);

				ControlListener moveListener = new ControlAdapter() {
					@Override
					public void controlMoved(final ControlEvent e) {
						// System.out.println("ControlListener#controlMoved");
						Rectangle[] currentRects = tracker.getRectangles();
						final Rectangle currentMouseMove = currentRects[0];
						Point popupLoc = new Point(popupSource.x + currentMouseMove.x - mouseMoveSource.x, popupSource.y
								+ currentMouseMove.y - mouseMoveSource.y);
						parent.setLocation(popupLoc); // shell
						// distance calculation is based on the mouse cursor.
						// if the calculation is based on the offset of the
						// selected popup,
						// it need to delete the two following lines of code
						popupLoc.x += me.x;
						popupLoc.y += me.y;
						Point mouseLocInEditor = textWidget.toControl(popupLoc);
						int minDist = Integer.MAX_VALUE;
						int snapPos = 0;
						for (Rectangle rect : dropTargets) {
							int dist = Geometry.distanceSquared(mouseLocInEditor, locations.get(snapPos));
							if (dist < minDist) {
								minDist = dist;
								fSnapPosition = snapPos;
								// fSnapPositionChanged = true;
								currentRects[1] = dropTargets.get(snapPos); // rect
							}
							snapPos++;
						}
						tracker.setRectangles(currentRects);
					}
				};
				tracker.addControlListener(moveListener);
				boolean committed = tracker.open();
				tracker.close();
				tracker.dispose();

				if (committed) {
					String exceptionName = label.getText();
					fLogger.info("Committed: @" + locations.get(fSnapPosition) + ", fDropPosition: " + fSnapPosition);
					ExceptionDropPosition dropPosition = dropPositions[fSnapPosition];
					closeDropShells(); // drop zones
					closeShell(fDragShell);
					fMode.addAnnotation(fDragAnnotation); // important
					if (dragPosition.includes(dropPosition.getOffset())) {
						System.out.println("Original; do nothing");
						fMode.addAnnotation(fDragAnnotation); // important

						// if (DROPZONE_REMOVED) {
						// System.out.println("Restore the presented variable annotation");
						// FixBugsAnnotation presentedVarAnno =
						// fAssistant.getPresentedVariableAnnotation();
						// if (presentedVarAnno != null) {
						// fMode.addAnnotation(presentedVarAnno);
						// }
						// }
					} else {
						fMode.performDND(!isThrown, exceptionName, dropPosition);
					}
				} else {
					fSnapPosition = originalSnapPosition;
					// fSnapPositionChanged = true;
				}
			}
		});
		final Color borderFg = fEditorWidget.getDisplay().getSystemColor(SWT.COLOR_BLACK);
		label.addPaintListener(new PaintListener() {
			@Override
			public void paintControl(final PaintEvent e) {
				Rectangle rect = label.getBounds();
				rect.width -= 1;
				rect.height -= 1;
				e.gc.setForeground(borderFg);
				e.gc.setLineAttributes(new LineAttributes(3, SWT.CAP_FLAT, SWT.JOIN_MITER, SWT.LINE_DOT, null, 5, 10));
				e.gc.drawRectangle(rect);
				e.gc.dispose();
			}
		});
	}

	/**
	 * 
	 * @param parent
	 *            the shell containing exception
	 * @param dropPosition
	 *            the model position in editor
	 * @param color
	 *            the color.
	 */
	private void createDropLabel(final Composite parent, final ExceptionDropPosition dropPosition, final Color color) {
		final Label label = new Label(parent, SWT.NONE);
		label.setFont(fFont);
		label.setLocation(0, 0);
		label.setSize(parent.getSize());
		label.setBackground(color);
	}

	/**
	 * gathers graphic (font, position, ...) information. It should be updated
	 * by user actions/preferences
	 */
	private void gatherGraphicInformation() {
		fFont = JFaceResources.getFont(PreferenceConstants.EDITOR_TEXT_FONT); // Important!
		fEditorWidget = fEditor.getViewer().getTextWidget();
		fLineHeight = fEditorWidget.getLineHeight();
		fAvgCharWidth = JFaceTextUtil.getAverageCharWidth(fEditorWidget);
	}

	private Shell openDragShell(final Position position, final String name, final boolean isThrown) {
		Shell workbenchShell = fEditor.getSite().getShell();
		final Shell shell = new Shell(workbenchShell, SWT.NO_TRIM);
		shell.setBounds(calculateBounds(position));
		shell.setVisible(true);
		createDragLabel(shell, position, name, isThrown);
		return shell;
	}

	/**
	 * 
	 * @param dropPosition
	 *            the model position in document
	 * @return
	 */
	private Shell openDropShell(final ExceptionDropPosition dropPosition, Color color) {
		Shell workbenchShell = fEditor.getSite().getShell();
		final Shell shell = new Shell(workbenchShell, SWT.NO_TRIM);
		Rectangle rect = calculateBounds(dropPosition);
		shell.setBounds(rect);
		shell.setVisible(true);
		shell.setAlpha(128);
		createDropLabel(shell, dropPosition, color);
		return shell;
	}

	/**
	 * 
	 * @param position
	 * @param name
	 * @param isThrown
	 * @return
	 */
	private ExceptionDropPosition[] openDropShells(final Position position, final String name, final boolean isThrown) {
		ExceptionDropPosition[] dropPositions = ExceptionDropPositionFinder.perform(fAssistant, position, name, isThrown);
		fDropShells = new ArrayList<Shell>();
		Color color = fEditorWidget.getDisplay().getSystemColor(SWT.COLOR_GRAY); // default
		// NOTE Use the same color as the exception being dragged
		// String exceptionName = name;
		// if (exceptionName != null && !exceptionName.isEmpty()) {
		// color = ExceptionColorRegistry.getInstance().getColor(exceptionName);
		// }
		for (ExceptionDropPosition expos : dropPositions) {
			// System.out.println("[DropPosition] " + position);
			fDropShells.add(openDropShell(expos, color));
		}
		return dropPositions;
	}

	/**
     *
     */
	public void start() {
		fLogger.info("ExceptionDNDManager#start");
		gatherGraphicInformation(); // re-gather whenever called.
		fEventListener = new MouseEventListener();
		fEventListener.start();
	}

	/**
     *
     */
	public void stop() {
		fLogger.info("ExceptionDNDManager#stop");
		closeShell(fDragShell);
		if (fEventListener != null) {
			fEventListener.stop();
			fEventListener = null;
		}
		if (fDragAnnotation != null) {
			fDragAnnotation = null;
		}
	}

}
