//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions.performance.stringBufferConcatenation;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;

import edu.ncsu.csc.emh.fixbugs.ui.Messages;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.FixBugsLinkedMode;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.FixBugsLinkedPopup;

/**
 * 
 * @author Yoonki
 */
public final class StringBufferConcatenationPopup extends FixBugsLinkedPopup {

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(StringBufferConcatenationPopup.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	private Button fStringBuilderButton;

	/** */
	private Button fStringBufferButton;

	/**
	 * Constructor for MissingArgumentPopup.
	 * 
	 * @param editor
	 *            the editor.
	 * @param linkedMode
	 *            the linked mode.
	 */
	public StringBufferConcatenationPopup(final CompilationUnitEditor editor, final FixBugsLinkedMode linkedMode) {
		super(editor, linkedMode);
	}

	@Override
	public StringBufferConcatenationMode getMode() {
		return (StringBufferConcatenationMode) fLinkedMode;
	}

	@Override
	public void setDefaultMessageText() {
		setMessage(Messages.stringBufferConcatenationMessage);
	}

	@Override
	public void createSuggestions(final Composite parent) {
		fLogger.info("StringBufferConcatenationPopup#createSuggestions");
		Composite optionComposite = new Composite(parent, SWT.NONE);
		// optionComposite.setLayout(new RowLayout());

		// gridData
		GridData gridData = new GridData(GridData.FILL, GridData.BEGINNING, true, true);
		gridData.horizontalSpan = 3;
		optionComposite.setLayoutData(gridData);

		// gridLayout
		GridLayout gridLayout = new GridLayout();
		gridLayout.numColumns = 1;
		optionComposite.setLayout(gridLayout);

		// 1st option: original
		createOriginalButton(optionComposite);

		// 2nd option: Use StringBuilder
		createStringBuilderSuggestion(optionComposite);

		// 3rd option: Use StringBuffer
		createStringBufferSuggestion(optionComposite);
	}

	private void createStringBuilderSuggestion(final Composite parent) {
		fStringBuilderButton = new Button(parent, SWT.RADIO);
		fStringBuilderButton.setText(Messages.useStringBuilder);
		fStringBuilderButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(MouseEvent e) {
				fLogger.info("fStringBuilderButton#mouseDown: " + fStringBuilderButton.getSelection());
				if (!fStringBuilderButton.getSelection()) {
					getMode().applySuggestion(StringBufferConcatenationMode.USE_STRING_BUILDER, getOriginalButton().getSelection());
				} else {
					fLogger.warning("do nothing!");
				}
			}
		});
	}

	private void createStringBufferSuggestion(final Composite parent) {
		fStringBufferButton = new Button(parent, SWT.RADIO);
		fStringBufferButton.setText(Messages.useStringBuffer);
		fStringBufferButton.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseDown(MouseEvent e) {
				fLogger.info("fStringBufferButton#mouseDown: " + fStringBufferButton.getSelection());
				if (!fStringBufferButton.getSelection()) {
					getMode().applySuggestion(StringBufferConcatenationMode.USE_STRING_BUFFER, getOriginalButton().getSelection());
				} else {
					fLogger.info("do nothing!");
				}
			}
		});
	}

	public void setStringBuilderSelected(boolean selected) {
		// setOriginalButtonSelected(false);
		fStringBuilderButton.setSelection(selected);
		// fStringBufferButton.setSelection(false);
	}

	public void setStringBufferSelected(boolean selected) {
		// setOriginalButtonSelected(false);
		// fStringBuilderButton.setSelection(false);
		fStringBufferButton.setSelection(selected);
	}
}
