//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.resolutions;

import static edu.umd.cs.findbugs.plugin.eclipse.quickfix.util.ConditionCheck.checkForNull;

import java.util.Comparator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.internal.ui.javaeditor.EditorUtility;
import org.eclipse.jdt.internal.ui.javaeditor.JavaEditor;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.swt.graphics.Image;
import org.eclipse.text.edits.TextEdit;
import org.eclipse.ui.IMarkerResolution2;

import de.tobject.findbugs.FindbugsPlugin;
import de.tobject.findbugs.reporter.MarkerUtil;
import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.umd.cs.findbugs.BugInstance;
import edu.umd.cs.findbugs.plugin.eclipse.quickfix.BugResolution;
import edu.umd.cs.findbugs.plugin.eclipse.quickfix.exception.BugResolutionException;

/**
 * @author Yoonki
 */
public abstract class PreviewBugResolution extends BugResolution implements IMarkerResolution2 {
	private static Logger fLogger = Logger.getLogger(PreviewBugResolution.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	@Override
	public final Image getImage() {
		return null;// Images.FIX.createImage();
	}

	private JavaEditor fEditor;

	protected IMarker fMarker;

	protected FixBugsContext fContext = null;

	protected ASTRewrite fRewrite = null;

	protected BugInstance fBugInstance = null;

	protected ICompilationUnit fCompilationUnit = null;

	protected Document fOriginalDoc = null;

	/** */
	protected String fCodePreview = null;

	/** */
	protected boolean fHasResolution = false;

	public PreviewBugResolution() {
		fContext = null;
		//System.out.println("PreviewBugResolution()");
	}

	public PreviewBugResolution(FixBugsContext context) {
		fContext = context;
		System.out.println("PreviewBugResolution(FixBugsContext)");
	}

	@Override
	public String getDescription() {
		if (fHasResolution) {
			return fCodePreview;
		}
		return "TBD";
	}

	@Override
	public void run(IMarker marker) {
		fLogger.info("PreviewBugResolution#run");
		checkForNull(marker, "marker");
		try {
			// do NOT inline this method invocation
			runInternal(marker);
		} catch (BugResolutionException e) {
			reportException(e);
		} catch (JavaModelException e) {
			reportException(e);
		} catch (BadLocationException e) {
			reportException(e);
		} catch (CoreException e) {
			reportException(e);
		}
	}

	/**
	 * This method is used by the test-framework, to catch the thrown exceptions and report it to the user.
	 *
	 * @see #run(IMarker)
	 */
	private void runInternal(IMarker marker) throws BugResolutionException, BadLocationException, CoreException {
		System.out.println("PreviewBugResolution#runInternal");
		fLogger.info("PreviewBugResolution#runInternal");
		Assert.isNotNull(marker);

		BugInstance bug = MarkerUtil.findBugInstanceForMarker(marker);
		if (bug == null) {
			throw new BugResolutionException(MISSING_BUG_INSTANCE);
		}

		IProject project = marker.getResource().getProject();
		ICompilationUnit originalUnit = getCompilationUnit(marker);
		if (originalUnit == null) {
			throw new BugResolutionException("No compilation unit found for marker " + marker.getType() + " (" + marker.getId()
					+ ")");
		}

		CompilationUnit workingUnit = createWorkingCopy(originalUnit, new NullProgressMonitor());

		ASTRewrite rewrite = ASTRewrite.create(workingUnit.getAST());

		try {
			if (fRewrite != null) {
				rewrite = fRewrite;
			}
			// rewrite = repairDefect(bug, marker);
			// repairDefect2(rewrite, workingUnit, bug);
			Document doc = new Document(originalUnit.getBuffer().getContents());
			rewriteCompilationUnit(rewrite, doc, originalUnit);

			FindbugsPlugin.getBugCollection(project, new NullProgressMonitor()).remove(bug);

			marker.delete();
		} finally {
			originalUnit.discardWorkingCopy();
		}
	}

	private void rewriteCompilationUnit(ASTRewrite rewrite, IDocument doc, ICompilationUnit originalUnit)
			throws JavaModelException, BadLocationException {
		TextEdit edits = rewrite.rewriteAST(doc, originalUnit.getJavaProject().getOptions(true));
		edits.apply(doc);

		originalUnit.getBuffer().setContents(doc.get());
		originalUnit.commitWorkingCopy(false, new NullProgressMonitor());
	}

	public void pseudoRun(final IMarker marker) {
		fLogger.info("PreviewBugResolution#pseudoRun");
		Assert.isNotNull(marker);
		try {
			// do NOT inline this method invocation
			fEditor = (JavaEditor) EditorUtility.isOpenInEditor(getCompilationUnit(marker));
			fMarker = marker;
			Assert.isNotNull(fEditor);
			pseudoRunInernal();
		} catch (BugResolutionException e) {
			reportException(e);
		} catch (JavaModelException e) {
			reportException(e);
		} catch (CoreException e) {
			reportException(e);
		}
	}

	static private final String MISSING_BUG_INSTANCE = "This bug is no longer in the system. "
			+ "The bugs somehow got out of sync with the memory representation. "
			+ "Try running FindBugs again. If that does not work, check the error log and remove the *.fbwarnings files.";

	private String pseudoRunInernal() throws BugResolutionException, CoreException {
		fLogger.info("PreviewBugResolution#pseudoRunInernal");
		fBugInstance = MarkerUtil.findBugInstanceForMarker(fMarker);
		if (fBugInstance == null) {
			throw new BugResolutionException(MISSING_BUG_INSTANCE);
		}
		fLogger.info("Line Number: " + fBugInstance.getProperty("lineNumber"));

		IProject project = fMarker.getResource().getProject();
		fCompilationUnit = getCompilationUnit(fMarker);
		if (fCompilationUnit == null) {
			throw new BugResolutionException("No compilation unit found for marker " + fMarker.getType() + " ("
					+ fMarker.getId() + ")");
		}

		fOriginalDoc = new Document(fCompilationUnit.getBuffer().getContents());
		CompilationUnit workingUnit = createWorkingCopy(fCompilationUnit, new NullProgressMonitor());

		ASTRewrite rewrite = ASTRewrite.create(workingUnit.getAST());

		String codePreview = "<none>";
		try {
			codePreview = repairDefectForPreview(fBugInstance, fMarker);
		} catch (Exception e) {
			reportException(e);
		}
		fLogger.info(codePreview);
		return codePreview;
	}

	// Why is the visibility of this method changed?
	private CompilationUnit createWorkingCopy(final ICompilationUnit unit, final IProgressMonitor monitor)
			throws JavaModelException {
		unit.becomeWorkingCopy(monitor);
		ASTParser parser = ASTParser.newParser(AST.JLS4);
		parser.setSource(unit);
		parser.setResolveBindings(true);
		parser.setStatementsRecovery(true);
		parser.setBindingsRecovery(true);
		return (CompilationUnit) parser.createAST(monitor);
	}

	protected abstract String repairDefectForPreview(BugInstance bug, IMarker marker) throws BugResolutionException,
			JavaModelException;

	public abstract String repairDefectForCommon(boolean previewMode);

	public ICompilationUnit getCompilationUnit() {
		return fCompilationUnit;
	}

	/***************************************************************/

	protected final ITextSelection getTextSelection() {
		ISelectionProvider selectionProvider = fEditor.getSelectionProvider();
		if (selectionProvider == null) {
			return null;
		}

		ISelection selection = selectionProvider.getSelection();
		if (!(selection instanceof ITextSelection)) {
			return null;
		}

		return (ITextSelection) selection;
	}

	/**
	 * Gets the invocation context for FixBugs.
	 *
	 * @return the invocation context.
	 */
	public final FixBugsContext getContext() {
		return fContext;
	}

	public void setContext(FixBugsContext context) {
		fContext = context;
	}

	protected void dispose() {
		fOriginalDoc = null;
		fContext = null;
	}

	protected String formatPreview(String text, int basePos, Map<ITrackedNodePosition, FixBugsAnnotation> annotationMap) {
		//System.out.println("BasePos: " + basePos + ", length: " + fCodePreview.length());
		Map<ITrackedNodePosition, FixBugsAnnotation> sortedAnnotationMap = new TreeMap<ITrackedNodePosition, FixBugsAnnotation>(
				new PositionComparator());

		for (Entry<ITrackedNodePosition, FixBugsAnnotation> entry : annotationMap.entrySet()) {
			sortedAnnotationMap.put(entry.getKey(), entry.getValue());
		}

		int delta = 0;
		for (Entry<ITrackedNodePosition, FixBugsAnnotation> entry : sortedAnnotationMap.entrySet()) {
			FixBugsAnnotation annotation = entry.getValue();
			if (annotation.isPreviewable()) {
				ITrackedNodePosition pos = entry.getKey();
				//System.out.println("Tracked: " + pos.getStartPosition() + ", " + pos.getLength());
				int newStartPos = pos.getStartPosition() - basePos + delta;
				// System.out.println("newStartPos: " + newStartPos + ", length: " + pos.getLength());
				String sub = text.substring(newStartPos, newStartPos + pos.getLength());
				// System.out.println("Sub: " + sub);
				text = text.replace(sub, "<b>" + sub + "</b>");
				delta += 7;
			}
		}

		// text = text.replace(">", "&lt;");
		// text = text.replace("<", "&gt;");
		text = text.replace("\n", "<br/>"); // Important!!!

		return text;
	}

	class PositionComparator implements Comparator<ITrackedNodePosition> {

		@Override
		public int compare(ITrackedNodePosition t1, ITrackedNodePosition t2) {
			return t1.getStartPosition() - t2.getStartPosition();
		}

	}

}
