//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.resolutions.badPractice.openStreamExceptionPath;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.ITextSelection;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.resolution.SearchHint;
import edu.ncsu.csc.emh.fixbugs.core.resolution.badPractice.openStreamExceptionPath.OpenStreamExceptionPathAssistant;
import edu.ncsu.csc.emh.fixbugs.resolutions.PreviewBugResolution;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.badPractice.openStreamExceptionPath.OpenStreamExceptionPathMode;
import edu.umd.cs.findbugs.BugInstance;
import edu.umd.cs.findbugs.plugin.eclipse.quickfix.exception.BugResolutionException;

/**
 *
 * @see edu.umd.cs.findbugs.detect.StringConcatenation
 * @version 1.0
 * @author Yoonki
 */
public final class CloseInFinallyBlockResolution extends PreviewBugResolution {
	private static Logger fLogger = Logger.getLogger(CloseInFinallyBlockResolution.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	@Override
	protected String repairDefectForPreview(BugInstance bug, IMarker marker) throws BugResolutionException {
		fLogger.info("CloseInFinallyBlockResolution#repairDefectForPreview");
		Assert.isNotNull(bug);
		Assert.isNotNull(marker);

		ITextSelection textSelection = getTextSelection();
		ICompilationUnit cu = getCompilationUnit(marker);
		dispose();// important
		fContext = new FixBugsContext(cu, textSelection.getOffset(), textSelection.getLength());

		return repairDefectForCommon(true);
	}

	@Override
	public String repairDefectForCommon(boolean previewMode) {
		fLogger.info("CloseInFinallyBlockResolution#repairDefectForCommon");
		Assert.isNotNull(fContext);

		// TODO We should use hierarchy instead of post-fixes of sub classes!
		SearchHint hint = new SearchHint();
		hint.add("Stream");
		hint.add("Reader");
		hint.add("JarFile");
		OpenStreamExceptionPathAssistant assistant = new OpenStreamExceptionPathAssistant(fContext, hint);
		String suggestion = OpenStreamExceptionPathMode.SPLIT_AND_MOVE_CLOSE_TO_FINALLY_BLOCK;

		try {
			assistant.init(); // required!

			Map<ITrackedNodePosition, FixBugsAnnotation> annotationMap = new HashMap<ITrackedNodePosition, FixBugsAnnotation>();
			List<ITrackedNodePosition> trackedVariables = new ArrayList<ITrackedNodePosition>();

			fCodePreview = assistant.changeAST(suggestion, annotationMap, trackedVariables);

			// System.out.println("Text: " + fCodePreview);

			// formatting code for preview!!
			if (previewMode) {
				// existingVariableLinks could be ignored!
				// newVariableLinks could be ignored!
				// annotationMap should not be ignore!
				int basePos = assistant.getCoveringBodyDecl().getStartPosition();
				fCodePreview = formatPreview(fCodePreview, basePos, annotationMap);
			}
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		} catch (CoreException e) {
			e.printStackTrace();
		}

		fRewrite = assistant.getASTRewrite();
		fHasResolution = true;

		return fCodePreview;
	}

	@Override
	protected boolean resolveBindings() {
		return true;
	}

	@Override
	protected void repairBug(ASTRewrite rewrite, CompilationUnit workingUnit, BugInstance bug) throws BugResolutionException {
		// do nothing

	}

}
