//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.resolutions.dodgyCode.catchException;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.ITextSelection;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException.CatchExceptionAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException.ExceptionDataModel;
import edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException.ExceptionsMethodsPair;
import edu.ncsu.csc.emh.fixbugs.resolutions.PreviewBugResolution;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.dodgyCode.catchException.CatchExceptionMode;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.dodgyCode.catchException.ExceptionColorRegistry;
import edu.umd.cs.findbugs.BugInstance;
import edu.umd.cs.findbugs.plugin.eclipse.quickfix.exception.BugResolutionException;

/**
 *
 * @see edu.umd.cs.findbugs.detect.UniCatchExceptionResolution
 * @version 1.0
 * @author Yoonki
 */
public final class ThrowExceptionResolution extends PreviewBugResolution {
	private static Logger fLogger = Logger.getLogger(ThrowExceptionResolution.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	@Override
	public String repairDefectForPreview(BugInstance bug, IMarker marker) throws BugResolutionException, JavaModelException {
		fLogger.info("ThrowExceptionResolution#repairDefectForPreview");
		Assert.isNotNull(bug);
		Assert.isNotNull(marker);

		ITextSelection textSelection = getTextSelection();
		ICompilationUnit cu = getCompilationUnit(marker);
		dispose();// important
		fContext = new FixBugsContext(cu, textSelection.getOffset(), textSelection.getLength());

		ExceptionColorRegistry.getInstance().init(); // important!
		return repairDefectForCommon(true);
	}

	@Override
	public String repairDefectForCommon(boolean previewMode) {
		fLogger.info("ThrowExceptionResolution#repairDefectForCommon");
		Assert.isNotNull(fContext);

		String suggestion = CatchExceptionMode.THROW_EXCEPTIONS;
		CatchExceptionAssistant assistant = new CatchExceptionAssistant(fContext);

		try {
			assistant.init(); // required!

			ExceptionDataModel dataModel = new ExceptionDataModel(assistant.getOriginalCaughtExceptions(),
					new ArrayList<ExceptionsMethodsPair>());

			Map<ITrackedNodePosition, FixBugsAnnotation> trackedMethodNameAnnotationMap = new HashMap<ITrackedNodePosition, FixBugsAnnotation>();
			Map<ITrackedNodePosition, FixBugsAnnotation> trackedExceptionNameAnnotationMap = new HashMap<ITrackedNodePosition, FixBugsAnnotation>();

			fCodePreview = assistant.changeAST(suggestion, dataModel, trackedMethodNameAnnotationMap,
					trackedExceptionNameAnnotationMap);

			System.out.println("Text: " + fCodePreview);

			// formatting code for preview!!
			if (previewMode) {
				int basePos = assistant.getCoveringBodyDecl().getStartPosition();
				fCodePreview = formatPreview(fCodePreview, basePos, trackedExceptionNameAnnotationMap);
			}
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		} catch (CoreException e) {
			e.printStackTrace();
		}

		fRewrite = assistant.getASTRewrite();
		fHasResolution = true;

		return fCodePreview;
	}


	@Override
	protected boolean resolveBindings() {
		return true;
	}


	@Override
	protected void repairBug(ASTRewrite rewrite, CompilationUnit workingUnit, BugInstance bug) throws BugResolutionException {
		// TODO Auto-generated method stub

	}


}
