//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.resolutions.performance.gc;

import static edu.umd.cs.findbugs.plugin.eclipse.quickfix.util.ASTUtil.getMethodDeclaration;
import static edu.umd.cs.findbugs.plugin.eclipse.quickfix.util.ASTUtil.getTypeDeclaration;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.text.edits.MalformedTreeException;
import org.eclipse.text.edits.TextEdit;

import edu.ncsu.csc.emh.fixbugs.resolutions.PreviewBugResolution;
import edu.umd.cs.findbugs.BugInstance;
import edu.umd.cs.findbugs.plugin.eclipse.quickfix.exception.BugResolutionException;
import edu.umd.cs.findbugs.plugin.eclipse.quickfix.util.ASTUtil;

/**
 * @author Yoonki
 * @version 1.0
 */
public final class ExplicitGarbageCollectionResolution extends PreviewBugResolution {
	private static Logger fLogger = Logger.getLogger(ExplicitGarbageCollectionResolution.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	@Override
	protected void repairBug(final ASTRewrite rewrite, final CompilationUnit workingUnit, final BugInstance bug)
			throws BugResolutionException {
		fLogger.info("ExplicitGarbageCollectionResolution#repairBug");
		Assert.isNotNull(rewrite);
		Assert.isNotNull(workingUnit);
		Assert.isNotNull(bug);

		Statement statement = findUselessStatement(ASTUtil.getASTNode(workingUnit, bug.getPrimarySourceLineAnnotation()));
		if (statement.toString().contains("System.gc")) {
			rewrite.remove(statement, null);
		}
	}

	private Statement findUselessStatement(final ASTNode node) {
		if (node instanceof Statement) {
			return (Statement) node;
		}
		return null;
	}

	protected String repairDefectForPreview(final ASTRewrite rewrite, final CompilationUnit workingUnit,
			final BugInstance bugInstance, final IMarker marker) throws BugResolutionException {
		fLogger.info("ExplicitGarbageCollectionResolution#repairDefectForPreview");
		Assert.isNotNull(rewrite);
		Assert.isNotNull(workingUnit);
		Assert.isNotNull(bugInstance);
		TypeDeclaration typeDecl = getTypeDeclaration(workingUnit, bugInstance.getPrimaryClass());
		fLogger.info("\tType: " + typeDecl.getName());
		MethodDeclaration methodDecl = getMethodDeclaration(typeDecl, bugInstance.getPrimaryMethod());
		fLogger.info("\tMethod: " + methodDecl.toString());
		Statement statement = findUselessStatement(ASTUtil
				.getASTNode(workingUnit, bugInstance.getPrimarySourceLineAnnotation()));
		ASTNode parent = statement.getParent();
		ASTNode grandParent = parent.getParent();

		AST baseAST = grandParent.getAST();
		ASTRewrite pseudoRewrite = ASTRewrite.create(baseAST);

		ITrackedNodePosition placeHolderPosition = pseudoRewrite.track(methodDecl);
		if (statement.toString().contains("System.gc")) {
			pseudoRewrite.remove(statement, null);
		}
		fLogger.info("\tpseudoRrewrite: " + pseudoRewrite.toString());
		TextEdit edits = null;
		try {
			Document doc = new Document(fCompilationUnit.getBuffer().getContents());

			edits = pseudoRewrite.rewriteAST();
			fLogger.info("\tPseudoRewrite: " + pseudoRewrite.toString());
			edits.apply(doc);
			fLogger.info("\tEdits: " + edits.toString());

			// YS it is important to call get() after changing documentation (by
			// calling TextEdit#apply(IDocument).
			fCodePreview = doc.get(placeHolderPosition.getStartPosition(), placeHolderPosition.getLength());
			fCodePreview = fCodePreview.replace("\n", "<br/>"); // YS Important!!!
			fHasResolution = true;
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (MalformedTreeException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		}
		fLogger.info("[Preview]\n" + fCodePreview);
		return fCodePreview;
	}

	@Override
	protected boolean resolveBindings() {
		return false;
	}

	@Override
	protected String repairDefectForPreview(BugInstance bug, IMarker marker) throws BugResolutionException, JavaModelException {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public String repairDefectForCommon(boolean previewMode) {
		// TODO Auto-generated method stub
		return null;
	}

}
