//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.resolutions.performance.stringBufferConcatenation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.ITextSelection;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.resolution.SearchHint;
import edu.ncsu.csc.emh.fixbugs.core.resolution.performance.stringBufferConcetenation.StringBufferConcatenationAssistant;
import edu.ncsu.csc.emh.fixbugs.resolutions.PreviewBugResolution;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.performance.stringBufferConcatenation.StringBufferConcatenationMode;
import edu.umd.cs.findbugs.BugInstance;
import edu.umd.cs.findbugs.plugin.eclipse.quickfix.exception.BugResolutionException;

/**
 *
 * @see edu.umd.cs.findbugs.detect.StringConcatenation
 * @version 1.0
 * @author Yoonki
 */
public final class StringBufferResolution extends PreviewBugResolution {
	private static Logger fLogger = Logger.getLogger(StringBufferResolution.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	@Override
	public String repairDefectForPreview(BugInstance bug, IMarker marker) throws BugResolutionException {
		fLogger.info("StringBufferResolution#repairDefectForPreview");
		Assert.isNotNull(bug);
		Assert.isNotNull(marker);

		ITextSelection textSelection = getTextSelection();
		ICompilationUnit cu = getCompilationUnit(marker);
		dispose();// important
		fContext = new FixBugsContext(cu, textSelection.getOffset(), textSelection.getLength());

		return repairDefectForCommon(true);
	}

	@Override
	public String repairDefectForCommon(boolean previewMode) {
		fLogger.info("StringBufferResolution#repairDefectForCommon");
		Assert.isNotNull(fContext);

		String suggestion = StringBufferConcatenationMode.USE_STRING_BUFFER;
		SearchHint hint = new SearchHint();
		hint.add("String");
		StringBufferConcatenationAssistant assistant = new StringBufferConcatenationAssistant(fContext, hint);

		try {
			assistant.init(); // required!

			Map<ITrackedNodePosition, FixBugsAnnotation> annotationMap = new HashMap<ITrackedNodePosition, FixBugsAnnotation>();
			List<ITrackedNodePosition> existingVariableLinks = new ArrayList<ITrackedNodePosition>();
			List<ITrackedNodePosition> newVariableLinks = new ArrayList<ITrackedNodePosition>();

			fCodePreview = assistant.changeAST(suggestion, annotationMap, existingVariableLinks, newVariableLinks);

			//System.out.println("Text: " + fCodePreview);

			// formatting code for preview!!
			if (previewMode) {
				// existingVariableLinks could be ignored!
				// newVariableLinks could be ignored!
				// annotationMap should not be ignore!
				int basePos = assistant.getCoveringBodyDecl().getStartPosition();
				fCodePreview = formatPreview(fCodePreview, basePos, annotationMap);
			}
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		}

		fRewrite = assistant.getASTRewrite();
		fHasResolution = true;

		return fCodePreview;
	}

	@Override
	protected boolean resolveBindings() {
		return true;
	}

	@Override
	protected void repairBug(ASTRewrite rewrite, CompilationUnit workingUnit, BugInstance bug) throws BugResolutionException {
		// do nothing
	}
}
